(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 7.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[   5521568,     161906]
NotebookOptionsPosition[   5474154,     160518]
NotebookOutlinePosition[   5474716,     160542]
CellTagsIndexPosition[   5474626,     160537]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell[TextData[StyleBox["The N.P. and G.H.P. formalisms. ",
 FontColor->RGBColor[1, 0, 0]]], "Title"],

Cell[TextData[{
 StyleBox["An application of ",
  FontColor->RGBColor[0, 0, 1]],
 StyleBox["Spinors",
  FontSlant->"Italic",
  FontColor->RGBColor[0, 0, 1]],
 StyleBox[" and ",
  FontColor->RGBColor[0, 0, 1]],
 StyleBox["xCoba.",
  FontSlant->"Italic",
  FontColor->RGBColor[0, 0, 1]]
}], "Subtitle"],

Cell["\<\
Alfonso Garc\[IAcute]a-Parrado G\[OAcute]mez-Lobo
Universidade do Minho, Portugal
September 2009\
\>", "Subsubtitle"],

Cell["\<\
In this notebook we derive all the equations of the Newman-Penrose formalism \
from scratch, by using spinor calculus. We use the equations obtained to \
introduce the essentials of the GHP formalism (GHP operators, weighted \
quantities, etc) and then we obtain the set of GHP equations. This notebook \
can be used as starting point for any computation involving the N.P. and \
G.H.P. formalisms.\
\>", "Text",
 TextJustification->1.],

Cell["\<\
For a quick start, execute sections 1, 2 and 3 and then proceed to section 4 \
where the essential explanations about how to use this notebook are provided. \
\>", "Text",
 TextJustification->1.],

Cell[CellGroupData[{

Cell["1. Spin structure on the spacetime manifold ", "Subsection"],

Cell[TextData[{
 "We start by loading  ",
 StyleBox["Spinors`", "Output"],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "xAct`Spinors`"}]], "Input"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.1\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2013", ",", "7", ",", "1"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.1", ", ", {2013, 7, 1}],
  Editable->False]], "Print"],

Cell[BoxData["\<\"CopyRight (C) 2003-2013, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.0\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2013", ",", "7", ",", "1"}], "}"}]}],
  SequenceForm["Package xAct`xTensor`  version ", "1.1.0", ", ", {2013, 7, 1}],
  Editable->False]], "Print"],

Cell[BoxData["\<\"CopyRight (C) 2002-2013, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`Spinors`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.0.5\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2013", ",", "7", ",", "1"}], "}"}]}],
  SequenceForm["Package xAct`Spinors`  version ", "1.0.5", ", ", {2013, 7, 1}],
  Editable->False]], "Print"],

Cell[BoxData["\<\"CopyRight (C) 2006-2012, Alfonso Garcia-Parrado Gomez-Lobo \
and Jose M. Martin-Garcia, under the General Public License.\"\>"], "Print"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"]
}, Open  ]],

Cell[TextData[{
 "Definition of a four dimensional  manifold ",
 StyleBox["M4",
  FontFamily->"Courier"],
 ":"
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M4", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "f", ",", "p", ",", "q", ",", "r",
      ",", "m", ",", "l", ",", "h", ",", "j", ",", "n", ",", "t", ",", "s"}], 
    "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M4", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M4, ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM4", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM4, ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
 "Definition of a Lorentzian metric on the manifold ",
 StyleBox["M4",
  FontFamily->"Courier"],
 ":"
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "3", ",", "0"}], "}"}], ",", 
   RowBox[{"g", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD"}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"g", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`g[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilong", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilong[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric \"\>", "\[InvisibleSpace]", 
   RowBox[{"Tetrag", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tetrametric ", 
   $CellContext`Tetrag[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tetrametric \"\>", "\[InvisibleSpace]", 
   RowBox[{"Tetrag\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tetrametric ", 
   $CellContext`Tetrag\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "Weyl tensor ", 
   $CellContext`WeylCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 4],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 4],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 4],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detg", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detg[], ". ", "Determinant."],
  Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
 "We define a spin structure comprising the complex vector bundle ",
 StyleBox["Spin", "Input",
  FontWeight->"Plain"],
 " whose abstract indices are ",
 StyleBox["{A, B, C, D, F, H, L, P, Q}", "Input",
  FontWeight->"Plain"],
 StyleBox[".", "Input"],
 " We need to supply as arguments the name of the metric tensor of the \
background Lorentzian manifold",
 StyleBox[" ", "Input"],
 StyleBox["g[-a,-b]", "Input",
  FontWeight->"Plain"],
 ", the name of the antisymmetric spin metric",
 StyleBox[" ", "Input"],
 StyleBox["\[Epsilon][-A,-B]", "Input",
  FontWeight->"Plain"],
 StyleBox[",", "Input"],
 " the name of the soldering form ",
 StyleBox["\[Sigma][a,-A,-A\[Dagger]]", "Input",
  FontWeight->"Plain"],
 " and the name of the spin covariant derivative ",
 StyleBox["CDe[-a].", "Input",
  FontWeight->"Plain"],
 "The option ",
 StyleBox["SpinorPrefix", "Input",
  FontWeight->"Plain"],
 StyleBox[" ", "Input"],
 StyleBox["is used to introduce a tag which will be  prepended automatically \
to  any spinor related to a spacetime tensor and t", "Text",
  FontWeight->"Plain"],
 "he option ",
 StyleBox["SpinorMark", "Input",
  FontWeight->"Plain"],
 StyleBox[" introduces a formatting of the tag ", "Text",
  FontWeight->"Plain"],
 StyleBox["SP", "Input",
  FontWeight->"Plain"],
 StyleBox[" different to the default one. The long output of ", "Text",
  FontWeight->"Plain"],
 StyleBox["DefSpinStructure", "Input",
  FontWeight->"Plain"],
 StyleBox[" can be suppressed  by adding the option ", "Text",
  FontWeight->"Plain"],
 StyleBox["Info->False", "Text",
  FontFamily->"Courier",
  FontWeight->"Plain"],
 StyleBox[" which deactivates the information of the definition commands .", 
  "Text",
  FontWeight->"Plain"]
}], "Text",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefSpinStructure", "[", 
  RowBox[{"g", ",", "Spin", ",", 
   RowBox[{"{", 
    RowBox[{
    "A", ",", "B", ",", "C", ",", "D", ",", "F", ",", "H", ",", "L", ",", "P",
      ",", "Q"}], "}"}], ",", "\[Epsilon]", ",", "\[Sigma]", ",", "CDe", ",", 
   
   RowBox[{"SpinorPrefix", "->", "SP"}], ",", 
   RowBox[{"SpinorMark", "->", "\"\<S\>\""}], ",", 
   RowBox[{"Info", "->", "False"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "Spin", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`Spin, ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"ValidateSymbol", "::", "capital"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"System name \[NoBreak]\\!\\(\\\"C\\\"\\)\[NoBreak] is \
overloaded as an abstract index.\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"ValidateSymbol", "::", "capital"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"System name \[NoBreak]\\!\\(\\\"D\\\"\\)\[NoBreak] is \
overloaded as an abstract index.\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle \"\>", "\[InvisibleSpace]", 
   "Spin\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Assuming fixed anti-isomorphism between Spin and \
Spin\[Dagger]\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "conjugated vbundle ", $CellContext`Spin\[Dagger], ". ", 
   "Assuming fixed anti-isomorphism between Spin and Spin\[Dagger]"],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Sigma]", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Sigma][$CellContext`a, -C, -$CellContext`D\[Dagger]], ". ", 
   ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Sigma]\[Dagger]", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Sigma]\[Dagger][$CellContext`a, -$CellContext`C\[Dagger], -
    D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Epsilon]", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Epsilon][-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Epsilon]\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Epsilon]\[Dagger][-$CellContext`A\[Dagger], -$CellContext`B\
\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Sigma\[Sigma]", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Sigma\[Sigma][$CellContext`a, $CellContext`b, \
-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Sigma\[Sigma]\[Dagger]", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Sigma\[Sigma]\[Dagger][$CellContext`a, $CellContext`b, \
-$CellContext`A\[Dagger], -$CellContext`B\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]g", "[", 
    RowBox[{"A", ",", "C\[Dagger]", ",", "B", ",", "D\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]g[$CellContext`A, $CellContext`C\[Dagger], \
$CellContext`B, $CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]g\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", "C", ",", "B\[Dagger]", ",", "D"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]g\[Dagger][$CellContext`A\[Dagger], 
    C, $CellContext`B\[Dagger], D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]epsilong", "[", 
    RowBox[{
    "A", ",", "A\[Dagger]", ",", "B", ",", "B\[Dagger]", ",", "C", ",", 
     "C\[Dagger]", ",", "D", ",", "D\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]epsilong[$CellContext`A, $CellContext`A\
\[Dagger], $CellContext`B, $CellContext`B\[Dagger], 
    C, $CellContext`C\[Dagger], D, $CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]epsilong\[Dagger]", "[", 
    RowBox[{
    "A\[Dagger]", ",", "A", ",", "B\[Dagger]", ",", "B", ",", "C\[Dagger]", 
     ",", "C", ",", "D\[Dagger]", ",", "D"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]epsilong\[Dagger][$CellContext`A\[Dagger], \
$CellContext`A, $CellContext`B\[Dagger], $CellContext`B, $CellContext`C\
\[Dagger], C, $CellContext`D\[Dagger], D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CDe", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CDe[-$CellContext`a], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDCDe", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing tensor ", 
   $CellContext`ChristoffelCDCDe[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelCDe", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelCDe[$CellContext`A, -$CellContext`b, -C], ". ", 
   ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelCDe\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelCDe\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"FRiemannCDe", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}], ",", "D"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "FRiemann tensor ", 
   $CellContext`FRiemannCDe[-$CellContext`a, -$CellContext`b, -C, D], ". ", 
   "Antisymmetric only in the first pair."],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"FRiemannCDe\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", "D\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "FRiemann tensor ", 
   $CellContext`FRiemannCDe\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`C\[Dagger], $CellContext`D\[Dagger]], ". ", 
   "Antisymmetric only in the first pair."],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]FRiemannCDe", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]FRiemannCDe[-$CellContext`A, -$CellContext`A\
\[Dagger], -$CellContext`B, -$CellContext`B\[Dagger], -C, -D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]FRiemannCDe\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]FRiemannCDe\[Dagger][-$CellContext`A\[Dagger]\
, -$CellContext`A, -$CellContext`B\[Dagger], -$CellContext`B, -$CellContext`C\
\[Dagger], -$CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChiCDe", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChiCDe[-$CellContext`A, -$CellContext`B, -C, -D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChiCDe\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChiCDe\[Dagger][-$CellContext`A\[Dagger], -$CellContext`B\
\[Dagger], -$CellContext`C\[Dagger], -$CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe[-$CellContext`A, -$CellContext`B, -$CellContext`C\
\[Dagger], -$CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe\[Dagger][-$CellContext`A\[Dagger], -$CellContext`B\
\[Dagger], -C, -D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"LambdaCDe", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`LambdaCDe[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe[-$CellContext`A, -$CellContext`B, -C, -D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe\[Dagger][-$CellContext`A\[Dagger], -$CellContext`B\
\[Dagger], -$CellContext`C\[Dagger], -$CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]RiemannCD[-$CellContext`A, -$CellContext`A\
\[Dagger], -$CellContext`B, -$CellContext`B\[Dagger], -
    C, -$CellContext`C\[Dagger], -D, -$CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]RiemannCD\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D\[Dagger]"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]RiemannCD\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`A, -$CellContext`B\[Dagger], -$CellContext`B, -$CellContext`C\
\[Dagger], -C, -$CellContext`D\[Dagger], -D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]RicciCD[-$CellContext`A, -$CellContext`A\
\[Dagger], -$CellContext`B, -$CellContext`B\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]RicciCD\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]RicciCD\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`A, -$CellContext`B\[Dagger], -$CellContext`B], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]WeylCD[-$CellContext`A, -$CellContext`A\
\[Dagger], -$CellContext`B, -$CellContext`B\[Dagger], -
    C, -$CellContext`C\[Dagger], -D, -$CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]WeylCD\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D\[Dagger]"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]WeylCD\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`A, -$CellContext`B\[Dagger], -$CellContext`B, -$CellContext`C\
\[Dagger], -C, -$CellContext`D\[Dagger], -D], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]TFRicciCD[-$CellContext`A, -$CellContext`A\
\[Dagger], -$CellContext`B, -$CellContext`B\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"SP\[UnderBracket]TFRicciCD\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`SP\[UnderBracket]TFRicciCD\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`A, -$CellContext`B\[Dagger], -$CellContext`B], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
Pretty printing options for the spinor indices and the spin metric.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"PrintAs", "[", "A\[Dagger]", "]"}], "^=", "\"\<A'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "B\[Dagger]", "]"}], "^=", "\"\<B'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "C\[Dagger]", "]"}], "^=", "\"\<C'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "D\[Dagger]", "]"}], "^=", "\"\<D'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "F\[Dagger]", "]"}], "^=", "\"\<F'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "H\[Dagger]", "]"}], "^=", "\"\<H'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "L\[Dagger]", "]"}], "^=", "\"\<L'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "M\[Dagger]", "]"}], "^=", "\"\<M'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "P\[Dagger]", "]"}], "^=", "\"\<P'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "Q\[Dagger]", "]"}], "^=", "\"\<Q'\>\""}], ",", 
    RowBox[{
     RowBox[{"PrintAs", "[", "\[Epsilon]\[Dagger]", "]"}], "^=", 
     "\"\<\!\(\*OverscriptBox[\(\[Epsilon]\), \(__\)]\)\>\""}]}], "}"}], 
  ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["2. The Newman-Penrose formalism", "Subsection"],

Cell[TextData[{
 "The Newman-Penrose formalism can be regarded as an application of the \
spinor calculus. In this section we show how all the equations of this \
formalism can be derived sytematically and efficiently using ",
 StyleBox["Spinors`", "Output"],
 "  in conjuntion with the package ",
 StyleBox["xCoba`", "Output"],
 ". Full explanations about the usage of ",
 StyleBox["xCoba`", "Output"],
 "  can be found in the package documentation. "
}], "Text",
 TextJustification->1.,
 FontWeight->"Plain"],

Cell[TextData[{
 "We load ",
 StyleBox["xCoba`", "Output"],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "xAct`xCoba`"}]], "Input"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xCoba`  version \"\>", 
   "\[InvisibleSpace]", "\<\"0.8.1\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2013", ",", "7", ",", "1"}], "}"}]}],
  SequenceForm["Package xAct`xCoba`  version ", "0.8.1", ", ", {2013, 7, 1}],
  Editable->False]], "Print"],

Cell[BoxData["\<\"CopyRight (C) 2005-2013, David Yllanes and Jose M. \
Martin-Garcia, under the General Public License.\"\>"], "Print"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print"]
}, Open  ]],

Cell[TextData[{
 "We turn off the verbose messages issued by ",
 StyleBox["ComponentValue", "Output"],
 StyleBox[".", "Input"]
}], "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CVVerbose", "=", "False"}], ";"}]], "Input"],

Cell[TextData[{
 "In the following subsections we introduce the spin dyad and its associated \
null tetrad. We will use  ",
 StyleBox["xCoba`",
  FontFamily->"Adobe Courier"],
 " to introduce two frames (basis), called ",
 StyleBox["Dyad",
  FontFamily->"Adobe Courier"],
 " and ",
 StyleBox["NP ",
  FontFamily->"Adobe Courier"],
 "and we shall create a parallel set of vectors for the elements in each \
basis. See the corresponding subsections for full details."
}], "Text",
 TextJustification->1.],

Cell[CellGroupData[{

Cell["Spin dyad", "Subsubsection"],

Cell["First of all, we define the elements of a spin dyad.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"DefSpinor", "[", 
     RowBox[{"#", ",", "M4"}], "]"}], "&"}], "/@", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"o", "[", "A", "]"}], ",", 
     RowBox[{"\[Iota]", "[", "A", "]"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "o\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(o\), \(__\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Iota]\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Iota]\), \(__\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"o", "[", "A", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`o[$CellContext`A], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"o\[Dagger]", "[", "A\[Dagger]", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`o\[Dagger][$CellContext`A\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Iota]", "[", "A", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Iota][$CellContext`A], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Iota]\[Dagger]", "[", "A\[Dagger]", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["These spinors belong to a spin basis which we define next.", \
"MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"Dyad", ",", "Spin", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1"}], "}"}], ",", 
   RowBox[{"BasisColor", "\[Rule]", 
    RowBox[{"RGBColor", "[", 
     RowBox[{"1", ",", "0", ",", "0"}], "]"}]}], ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "Dyad", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`Dyad, 
   ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDDyad", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDDyad[-$CellContext`a], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDDyad", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDDyad[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDDyad", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDDyad[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDDyad", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDDyad[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDDyad", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDDyad[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDDyad", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelPDDyad[$CellContext`A, -$CellContext`b, -C], ". ",
    ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelPDDyad\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelPDDyad\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing FRiemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"FRiemannPDDyad", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}], ",", "D"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing FRiemann tensor ", 
   $CellContext`FRiemannPDDyad[-$CellContext`a, -$CellContext`b, -C, D], ". ",
    ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing FRiemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"FRiemannPDDyad\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", "D\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing FRiemann tensor ", 
   $CellContext`FRiemannPDDyad\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`C\[Dagger], $CellContext`D\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", 
   "Dyad\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", 
   "basis ", $CellContext`Dyad\[Dagger], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpDyad", "[", 
    RowBox[{"A", ",", "B"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpDyad[$CellContext`A, $CellContext`B], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpDyad\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", "B\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpDyad\[Dagger][$CellContext`A\[Dagger], $CellContext`B\
\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownDyad", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownDyad[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownDyad\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownDyad\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`B\[Dagger]], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
Component indices in the conjugated basis will be also primed :\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Dyad\[Dagger]", "/:", 
   RowBox[{"CIndexForm", "[", 
    RowBox[{"i_Integer", ",", "Dyad\[Dagger]"}], "]"}], ":=", 
   RowBox[{"StringJoin", "[", 
    RowBox[{
     RowBox[{"ToString", "[", "i", "]"}], ",", "\"\<'\>\""}], "]"}]}], 
  ";"}]], "Input"],

Cell["\<\
By definition, these are the equations relating both sets of objects:\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"eqs", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"o", "[", "A", "]"}], "\[Equal]", 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", 
         RowBox[{"-", "Dyad"}]}], "}"}], ",", "A"}], "]"}]}], ",", 
    RowBox[{
     RowBox[{"\[Iota]", "[", "A", "]"}], "\[Equal]", 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "Dyad"}]}], "}"}], ",", "A"}], "]"}]}]}], 
   "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"o", 
         StyleBox[GridBox[{
            {"A"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o[$CellContext`A],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "A"},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{0, -$CellContext`Dyad}, $CellContext`A],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Iota]", 
         StyleBox[GridBox[{
            {"A"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota][$CellContext`A],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "A"},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`Dyad}, $CellContext`A],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "From them, we read and store the components of the spinors ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["o", " ", "A"],
    AutoSpacing->False],
   $CellContext`o[$CellContext`A],
   Editable->False]], "Output",
  CellChangeTimes->{3.4238257047278767`*^9, 3.423825778247744*^9, 
   3.423825861535986*^9}],
 ", ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Iota]", " ", "A"],
    AutoSpacing->False],
   $CellContext`\[Iota][$CellContext`A],
   Editable->False]], "Output",
  CellChangeTimes->{3.4238257047278767`*^9, 3.423825778247744*^9, 
   3.423825861535986*^9}],
 " ."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "eqs", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Det\[Epsilon]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Det\[Epsilon][], ". ", "Determinant."],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Det\[Epsilon]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Det\[Epsilon]Dyad[], ". ", "Determinant."],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"o", 
         StyleBox[GridBox[{
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      0, -$CellContext`Dyad}, {$CellContext`A, $CellContext`Dyad}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Iota]", 
         StyleBox[GridBox[{
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      1, -$CellContext`Dyad}, {$CellContext`A, $CellContext`Dyad}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We now do the same for the conjugated equations:", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], ",", 
   RowBox[{"Dagger", "[", "eqs", "]"}], ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Det\[Epsilon]\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Det\[Epsilon]\[Dagger][], ". ", "Determinant."],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Det\[Epsilon]\[Dagger]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Det\[Epsilon]\[Dagger]Dyad\[Dagger][], ". ", "Determinant."],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["o", "__"], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      0, -$CellContext`Dyad\[Dagger]}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Iota]", "__"], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      1, -$CellContext`Dyad\[Dagger]}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "@", 
  RowBox[{"Join", "[", 
   RowBox[{"%%", ",", "%"}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{0, $CellContext`Dyad}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{1, $CellContext`Dyad}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{0, $CellContext`Dyad}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{1, $CellContext`Dyad}],
       Editable->False], "\[Equal]", "1"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{0, $CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{1, $CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{0, $CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{1, $CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Equal", "->", "ComponentValue"}]}], ";"}]], "Input"],

Cell[TextData[{
 "Since ",
 Cell[BoxData[
  RowBox[{"{", 
   StyleBox[
    RowBox[{
     SubsuperscriptBox["o", " ", "A"], ","}],
    AutoSpacing->False]}]], "Output",
  CellChangeTimes->{{3.423826323021446*^9, 3.423826346868227*^9}, 
    3.423826487957116*^9, {3.423826557584194*^9, 3.42382657235732*^9}, 
    3.423826805717166*^9, 3.4238273982616997`*^9, 3.423903800260459*^9, 
    3.4239043784188433`*^9, 3.4239046714212503`*^9, 3.42392536234076*^9, 
    3.42397833572017*^9, 3.4239934551341877`*^9, 3.4240051077502537`*^9, 
    3.4240611902849293`*^9, 3.424061340660375*^9, 3.4240708360967007`*^9, 
    3.4240760648380957`*^9, 3.424076474325555*^9, 3.424415057111891*^9, 
    3.424430450536281*^9, 3.4244468338964863`*^9, 3.42444740893721*^9, 
    3.4245794840971746`*^9, 3.424595355535811*^9, 3.424596536942671*^9, 
    3.424599108414694*^9, 3.424668078687495*^9, 3.424776230261053*^9, 
    3.424780974004918*^9, 3.424791882204584*^9, 3.424963735863695*^9, 
    3.425017466626889*^9, 3.425381932441518*^9, 3.425460126841139*^9, 
    3.425460884149899*^9, 3.4254685798737793`*^9, 3.4254737073383636`*^9, 
    3.425496991712772*^9, 3.425550852667362*^9, 3.4255793867147217`*^9, 
    3.4256217205863113`*^9, 3.4257077110767193`*^9, 3.42579024440254*^9, 
    3.4258789930781193`*^9, 3.426274866355981*^9, 3.426796529786223*^9, 
    3.426796830331471*^9, 3.426797048510157*^9, 3.426798569075074*^9, 
    3.426798700887356*^9, 3.426798824952031*^9, 3.4267988794887877`*^9, 
    3.4267989522929487`*^9, 3.426799131075737*^9, 3.426800596828487*^9, 
    3.427087625511497*^9, 3.4270954501415253`*^9, 3.427174249661171*^9}],
 Cell[BoxData[
  RowBox[{
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Iota]", " ", "A"],
     AutoSpacing->False],
    $CellContext`\[Iota][$CellContext`A],
    Editable->False], "}"}]], "Output",
  CellChangeTimes->{{3.423826323021446*^9, 3.423826346868227*^9}, 
    3.423826487957116*^9, {3.423826557584194*^9, 3.42382657235732*^9}, 
    3.423826805717166*^9, 3.4238273982616997`*^9, 3.423903800260459*^9, 
    3.4239043784188433`*^9, 3.4239046714212503`*^9, 3.42392536234076*^9, 
    3.42397833572017*^9, 3.4239934551341877`*^9, 3.4240051077502537`*^9, 
    3.4240611902849293`*^9, 3.424061340660375*^9, 3.4240708360967007`*^9, 
    3.4240760648380957`*^9, 3.424076474325555*^9, 3.424415057111891*^9, 
    3.424430450536281*^9, 3.4244468338964863`*^9, 3.42444740893721*^9, 
    3.4245794840971746`*^9, 3.424595355535811*^9, 3.424596536942671*^9, 
    3.424599108414694*^9, 3.424668078687495*^9, 3.424776230261053*^9, 
    3.424780974004918*^9, 3.424791882204584*^9, 3.424963735863695*^9, 
    3.425017466626889*^9, 3.425381932441518*^9, 3.425460126841139*^9, 
    3.425460884149899*^9, 3.4254685798737793`*^9, 3.4254737073383636`*^9, 
    3.425496991712772*^9, 3.425550852667362*^9, 3.4255793867147217`*^9, 
    3.4256217205863113`*^9, 3.4257077110767193`*^9, 3.42579024440254*^9, 
    3.4258789930781193`*^9, 3.426274866355981*^9, 3.426796529786223*^9, 
    3.426796830331471*^9, 3.426797048510157*^9, 3.426798569075074*^9, 
    3.426798700887356*^9, 3.426798824952031*^9, 3.4267988794887877`*^9, 
    3.4267989522929487`*^9, 3.426799131075737*^9, 3.426800596828487*^9, 
    3.427087625511497*^9, 3.4270954501415253`*^9, 3.427174249661171*^9}],
 " is a spin dyad, the spin metric can be written in the form"
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"eqs", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"\[Epsilon]", "[", 
     RowBox[{"A", ",", "B"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"o", "[", "A", "]"}], " ", 
      RowBox[{"\[Iota]", "[", "B", "]"}]}], "-", 
     RowBox[{
      RowBox[{"o", "[", "B", "]"}], " ", 
      RowBox[{"\[Iota]", "[", "A", "]"}]}]}]}], "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Epsilon]", 
        StyleBox[GridBox[{
           {"A", "B"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Epsilon][$CellContext`A, $CellContext`B],
    Editable->False], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"B"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`B],
       Editable->False]}], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"B"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`B],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
The components of the spin metric and its complex conjugate can now be \
computed and stored.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "eqs", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Epsilon]", 
        StyleBox[GridBox[{
           {
            StyleBox["A",
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox["B",
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Epsilon][{$CellContext`A, $CellContext`Dyad}, \
{$CellContext`B, $CellContext`Dyad}],
    Editable->False], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False]}], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["B",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], ",", 
   RowBox[{"Dagger", "[", "eqs", "]"}], ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        OverscriptBox["\[Epsilon]", "__"], 
        StyleBox[GridBox[{
           {
            StyleBox[
             RowBox[{"A", "'"}],
             FontColor->RGBColor[1., 0., 0.]], 
            StyleBox[
             RowBox[{"B", "'"}],
             FontColor->RGBColor[1., 0., 0.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Epsilon]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
    Editable->False], "\[Equal]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False]}], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"B", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "@", 
  RowBox[{"ComponentArray", "@", 
   RowBox[{"Join", "[", 
    RowBox[{"%%", ",", "%"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Epsilon]", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon][{0, $CellContext`Dyad}, {
          0, $CellContext`Dyad}],
         Editable->False], "\[Equal]", "0"}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Epsilon]", 
             StyleBox[GridBox[{
                {
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon][{0, $CellContext`Dyad}, {
          1, $CellContext`Dyad}],
         Editable->False], "\[Equal]", "1"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Epsilon]", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["0",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon][{1, $CellContext`Dyad}, {
          0, $CellContext`Dyad}],
         Editable->False], "\[Equal]", 
        RowBox[{"-", "1"}]}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Epsilon]", 
             StyleBox[GridBox[{
                {
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox["1",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon][{1, $CellContext`Dyad}, {
          1, $CellContext`Dyad}],
         Editable->False], "\[Equal]", "0"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {
             OverscriptBox["\[Epsilon]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"0", "'"}],
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox[
                  RowBox[{"0", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]\[Dagger][{0, $CellContext`Dyad\[Dagger]}, {
          0, $CellContext`Dyad\[Dagger]}],
         Editable->False], "\[Equal]", "0"}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {
             OverscriptBox["\[Epsilon]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"0", "'"}],
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox[
                  RowBox[{"1", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]\[Dagger][{0, $CellContext`Dyad\[Dagger]}, {
          1, $CellContext`Dyad\[Dagger]}],
         Editable->False], "\[Equal]", "1"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {
             OverscriptBox["\[Epsilon]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"1", "'"}],
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox[
                  RowBox[{"0", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]\[Dagger][{1, $CellContext`Dyad\[Dagger]}, {
          0, $CellContext`Dyad\[Dagger]}],
         Editable->False], "\[Equal]", 
        RowBox[{"-", "1"}]}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {
             OverscriptBox["\[Epsilon]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"1", "'"}],
                  FontColor->RGBColor[1., 0., 0.]], 
                 StyleBox[
                  RowBox[{"1", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]\[Dagger][{1, $CellContext`Dyad\[Dagger]}, {
          1, $CellContext`Dyad\[Dagger]}],
         Editable->False], "\[Equal]", "0"}]}], "}"}]}], "}"}]}], 
  "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Equal", "->", "ComponentValue"}]}], ";"}]], "Input"],

Cell[TextData[{
 "We deduce from here the components of ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Epsilon]", "AB", "  "],
    AutoSpacing->False],
   $CellContext`\[Epsilon][-$CellContext`A, -$CellContext`B],
   Editable->False]], "Output",
  CellChangeTimes->{3.423827499169783*^9}],
 " by means of the relation ",
 Cell[BoxData[
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubsuperscriptBox["\[Epsilon]", "AB", "  "],
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`A, -$CellContext`B],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      SubsuperscriptBox["\[Epsilon]", "  ", "CB"],
      AutoSpacing->False],
     $CellContext`\[Epsilon][C, $CellContext`B],
     Editable->False]}], "=", 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Delta]", 
      RowBox[{"A", " "}], 
      RowBox[{" ", "C"}]],
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`A, C],
    Editable->False]}]], "Output",
  GeneratedCell->False,
  CellAutoOverwrite->False,
  CellChangeTimes->{3.423827609425281*^9}],
 ".  To that end we need to find the inverse of the transposed of the matrix \
representation of the components of ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Epsilon]", "  ", "AB"],
    AutoSpacing->False],
   $CellContext`\[Epsilon][$CellContext`A, $CellContext`B],
   Editable->False]], "Output",
  CellChangeTimes->{3.427175671011052*^9}],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"EP", "=", 
  RowBox[{"ToValues", "@", 
   RowBox[{"TableOfComponents", "[", 
    RowBox[{
     RowBox[{"\[Epsilon]", "[", 
      RowBox[{"A", ",", "B"}], "]"}], ",", "Dyad"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Inverse", "@", 
  RowBox[{"Transpose", "@", "EP"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Storage of the components of  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Epsilon]", "  ", "AB"],
    AutoSpacing->False],
   $CellContext`\[Epsilon][$CellContext`A, $CellContext`B],
   Editable->False]], "Output",
  CellChangeTimes->{3.427175671011052*^9}],
 " and its complex conjugate."
}], "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"MetricInBasis", "[", 
   RowBox[{"\[Epsilon]", ",", 
    RowBox[{"-", "Dyad"}], ",", "%"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"MetricInBasis", "[", 
   RowBox[{"\[Epsilon]\[Dagger]", ",", 
    RowBox[{"-", "Dyad\[Dagger]"}], ",", "%%"}], "]"}], ";"}]}], "Input"],

Cell[TextData[{
 "We now compute and store the components of ",
 Cell[BoxData[
  StyleBox[
   SubsuperscriptBox["o", "A", " "],
   AutoSpacing->False]], "Output",
  CellChangeTimes->{
   3.423826079206477*^9, 3.4238261160657682`*^9, {3.423826160205621*^9, 
    3.423826186371304*^9}, 3.423827098750963*^9, 3.423827411291833*^9}],
 ", ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Iota]", "A", " "],
    AutoSpacing->False],
   $CellContext`\[Iota][-$CellContext`A],
   Editable->False]], "Output",
  CellChangeTimes->{
   3.423826079206477*^9, 3.4238261160657682`*^9, {3.423826160205621*^9, 
    3.423826186371304*^9}, 3.423827098750963*^9, 3.423827411291833*^9}],
 " and their complex conjugate."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"eqs", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"o", "[", 
      RowBox[{"-", "A"}], "]"}], "\[Equal]", 
     RowBox[{
      RowBox[{"\[Epsilon]", "[", 
       RowBox[{
        RowBox[{"-", "B"}], ",", 
        RowBox[{"-", "A"}]}], "]"}], " ", 
      RowBox[{"o", "[", "B", "]"}]}]}], ",", 
    RowBox[{
     RowBox[{"\[Iota]", "[", 
      RowBox[{"-", "A"}], "]"}], "\[Equal]", 
     RowBox[{
      RowBox[{"\[Epsilon]", "[", 
       RowBox[{
        RowBox[{"-", "B"}], ",", 
        RowBox[{"-", "A"}]}], "]"}], " ", 
      RowBox[{"\[Iota]", "[", "B", "]"}]}]}]}], "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"o", 
         StyleBox[GridBox[{
            {" "},
            {"A"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o[-$CellContext`A],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"B"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`B],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Epsilon]", 
          StyleBox[GridBox[{
             {" ", " "},
             {"B", "A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon][-$CellContext`B, -$CellContext`A],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Iota]", 
         StyleBox[GridBox[{
            {" "},
            {"A"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota][-$CellContext`A],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Epsilon]", 
          StyleBox[GridBox[{
             {" ", " "},
             {"B", "A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon][-$CellContext`B, -$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"B"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`B],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "eqs", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"o", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["B",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Epsilon]", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["B",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon][{-$CellContext`B, -$CellContext`Dyad}, \
{-$CellContext`A, -$CellContext`Dyad}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Iota]", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Epsilon]", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["B",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon][{-$CellContext`B, -$CellContext`Dyad}, \
{-$CellContext`A, -$CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["B",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], ",", 
   RowBox[{"Dagger", "[", "eqs", "]"}], ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["o", "__"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o\[Dagger][{-$CellContext`A\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"B", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Epsilon]", "__"], 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox[
               RowBox[{"B", "'"}],
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]\[Dagger][{-$CellContext`B\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Iota]", "__"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota]\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Epsilon]", "__"], 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox[
               RowBox[{"B", "'"}],
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]\[Dagger][{-$CellContext`B\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"B", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", 
   RowBox[{"Join", "[", 
    RowBox[{"%%", ",", "%"}], "]"}], ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"o", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{0, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Epsilon]", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon][{
        0, -$CellContext`Dyad}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{1, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Epsilon]", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon][{
        1, -$CellContext`Dyad}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Iota]", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Epsilon]", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon][{
        0, -$CellContext`Dyad}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{0, $CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Epsilon]", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon][{
        1, -$CellContext`Dyad}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{1, $CellContext`Dyad}],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["o", "__"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`o\[Dagger][{-$CellContext`A\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{0, $CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Epsilon]", "__"], 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]\[Dagger][{
        0, -$CellContext`Dyad\[Dagger]}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{1, $CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Epsilon]", "__"], 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]\[Dagger][{
        1, -$CellContext`Dyad\[Dagger]}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Iota]", "__"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Iota]\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Epsilon]", "__"], 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]\[Dagger][{
        0, -$CellContext`Dyad\[Dagger]}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{0, $CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Epsilon]", "__"], 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]\[Dagger][{
        1, -$CellContext`Dyad\[Dagger]}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{1, $CellContext`Dyad\[Dagger]}],
       Editable->False]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{0, -$CellContext`Dyad}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{1, -$CellContext`Dyad}],
       Editable->False], "\[Equal]", "1"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{0, -$CellContext`Dyad}],
       Editable->False], "\[Equal]", 
      RowBox[{"-", "1"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{1, -$CellContext`Dyad}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", 
      RowBox[{"-", "1"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Equal", "->", "ComponentValue"}]}], ";"}]], "Input"],

Cell[TextData[{
 "We change the ",
 StyleBox["Rule", "Output"],
 " assignment to the ",
 StyleBox["Set", "Output"],
 " assignment in each ",
 StyleBox["FoldedRule", "Output"],
 " storing the components computed so far"
}], "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleToSet", "/@", 
   RowBox[{"{", 
    RowBox[{
    "\[Epsilon]", ",", "\[Epsilon]\[Dagger]", ",", "o", ",", "o\[Dagger]", 
     ",", "\[Iota]", ",", "\[Iota]\[Dagger]"}], "}"}]}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Null tetrad", "Subsubsection"],

Cell["\<\
We define now four vectors which will represent the null tetrad basis \
elements.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"lNP", "[", "a", "]"}], ",", "M4", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<l\>\""}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"lNP", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`lNP[$CellContext`a], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"nNP", "[", "a", "]"}], ",", "M4", ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<n\>\""}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"nNP", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`nNP[$CellContext`a], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"mNP", "[", "a", "]"}], ",", "M4", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}], ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<m\>\""}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"mNP", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`mNP[$CellContext`a], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"mNP\[Dagger]", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`mNP\[Dagger][$CellContext`a], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "mNP\[Dagger]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(m\), \(_\)]\)\>\""}], ";"}]], "Input"],

Cell["\<\
These vectors belong to a basis (null  tetrad) which we define next.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefBasis", "[", 
  RowBox[{"NP", ",", "TangentM4", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"BasisColor", "->", 
    RowBox[{"RGBColor", "[", 
     RowBox[{"0", ",", "0", ",", "1"}], "]"}]}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefBasis", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"basis \"\>", "\[InvisibleSpace]", "NP", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xCoba`DefBasis, ": Defining ", "basis ", $CellContext`NP, ". ", 
   ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parallel derivative \"\>", "\[InvisibleSpace]", 
   RowBox[{"PDNP", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "parallel derivative ", 
   $CellContext`PDNP[-$CellContext`a], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionPDNP", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionPDNP[$CellContext`a, -$CellContext`b, -$CellContext`c],
    ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelPDNP", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelPDNP[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannPDNP", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannPDNP[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, $CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciPDNP", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciPDNP[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric +1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaUpNP", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric +1 density ", 
   $CellContext`etaUpNP[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric -1 density \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"etaDownNP", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric -1 density ", 
   $CellContext`etaDownNP[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
 "We specify the behaviour of the elements of this basis under complex \
conjugation. This is achieved with the command ",
 StyleBox["SetDaggerMatrix", "Input"],
 "."
}], "MathCaption"],

Cell[BoxData[
 RowBox[{"SetDaggerMatrix", "[", 
  RowBox[{"NP", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}]}], "}"}]}], 
  "]"}]], "Input"],

Cell[TextData[{
 "We format a number of objects according to standard conventions. Due to the \
new way of sorting definitions in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "  6.0, we need to stop the formatting engine and start it again after \
defining these new rules."
}], "Text",
 FontWeight->"Plain"],

Cell["Stop formatting engine.", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"xTensorFormStop", "[", "]"}], ";"}]], "Input"],

Cell["\<\
Formatting of the directional derivatives with the standard NP symbols.\
\>", "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<D\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[CapitalDelta]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[Delta]\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", 
    "\"\<\!\(\*OverscriptBox[\(\[Delta]\), \(_\)]\)\>\""}], "]"}], 
  ";"}]}], "Input"],

Cell[TextData[{
 "The Newman-Penrose directional derivatives can also act on spinors. For \
that case we need to use the covariant derivative ",
 StyleBox["CDe", "Output"],
 " rather than ",
 StyleBox["PDNP", "Output"],
 ".  Therefore we need to set a formatting for the directional derivatives \
taken with respect to ",
 StyleBox["CDe", "Output"],
 " (note that when ",
 StyleBox["CDe", "Output"],
 " and ",
 StyleBox["PDNP", "Output"],
 " act on scalars or tensors they coincide)."
}], "MathCaption",
 TextJustification->1.],

Cell[BoxData[{
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"CDe", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<D\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"CDe", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[CapitalDelta]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"CDe", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[Delta]\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"CDe", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", 
    "\"\<\!\(\*OverscriptBox[\(\[Delta]\), \(_\)]\)\>\""}], "]"}], 
  ";"}]}], "Input"],

Cell["Restart formatting  engine:", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"xTensorFormStart", "[", "]"}], ";"}]], "Input"],

Cell[TextData[{
 "These are the relations between the elements of the null tetrad  and the \
elements of the basis ",
 StyleBox["NP", "Output"],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"lNP", "[", "a", "]"}], "\[Equal]", 
    RowBox[{"Basis", "[", 
     RowBox[{"a", ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "NP"}]}], "}"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"nNP", "[", "a", "]"}], "\[Equal]", 
    RowBox[{"Basis", "[", 
     RowBox[{"a", ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "NP"}]}], "}"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"mNP", "[", "a", "]"}], "\[Equal]", 
    RowBox[{"Basis", "[", 
     RowBox[{"a", ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", 
        RowBox[{"-", "NP"}]}], "}"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"mNP\[Dagger]", "[", "a", "]"}], "\[Equal]", 
    RowBox[{"Basis", "[", 
     RowBox[{"a", ",", 
      RowBox[{"{", 
       RowBox[{"4", ",", 
        RowBox[{"-", "NP"}]}], "}"}]}], "]"}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`NP}, $CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`NP}, $CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`NP}, $CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{4, -$CellContext`NP}, $CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "From this relation we compute the components of ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["l", " ", "a"],
    AutoSpacing->False],
   $CellContext`lNP[$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 ",  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["n", " ", "a"],
    AutoSpacing->False],
   $CellContext`nNP[$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 ",  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["m", " ", "a"],
    AutoSpacing->False],
   $CellContext`mNP[$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 ",  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox[
     OverscriptBox["m", "_"], " ", "a"],
    AutoSpacing->False],
   $CellContext`mNP\[Dagger][$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 " in the basis ",
 StyleBox["NP", "Output"],
 " ."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"DetgNP", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`DetgNP[], ". ", "Determinant."],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{$CellContext`a, $CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      1, -$CellContext`NP}, {$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{$CellContext`a, $CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      2, -$CellContext`NP}, {$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{$CellContext`a, $CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      3, -$CellContext`NP}, {$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{
      4, -$CellContext`NP}, {$CellContext`a, $CellContext`NP}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{1, $CellContext`NP}],
     Editable->False], "\[Equal]", "1"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{2, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{3, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{4, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{1, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{2, $CellContext`NP}],
     Editable->False], "\[Equal]", "1"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{3, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{4, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{1, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{2, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{3, $CellContext`NP}],
     Editable->False], "\[Equal]", "1"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{4, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{1, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{2, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{3, $CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{4, $CellContext`NP}],
     Editable->False], "\[Equal]", "1"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Equal", "\[Rule]", "ComponentValue"}]}], ";"}]], "Input"],

Cell["\<\
As is well known, in the null tetrad the metric takes the form\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"g", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"lNP", "[", "a", "]"}], 
    RowBox[{"nNP", "[", "b", "]"}]}], "+", 
   RowBox[{
    RowBox[{"lNP", "[", "b", "]"}], 
    RowBox[{"nNP", "[", "a", "]"}]}], "-", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"mNP", "[", "a", "]"}], 
      RowBox[{"mNP\[Dagger]", "[", "b", "]"}]}], "+", 
     RowBox[{
      RowBox[{"mNP", "[", "b", "]"}], 
      RowBox[{"mNP\[Dagger]", "[", "a", "]"}]}]}], ")"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`g[$CellContext`a, $CellContext`b],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[$CellContext`b],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`b],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`b],
     Editable->False]}]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "We compute and store the components of ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["g", "  ", "ab"],
    AutoSpacing->False],
   $CellContext`g[$CellContext`a, $CellContext`b],
   Editable->False]], "Output",
  CellChangeTimes->{{3.457247975119607*^9, 3.457247985098529*^9}, {
    3.4572480842554502`*^9, 3.457248099409466*^9}, 3.457248317855564*^9}],
 " in the basis ",
 StyleBox["NP", "Output"],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`g[{$CellContext`a, $CellContext`NP}, {$CellContext`b, \
$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{$CellContext`a, $CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{$CellContext`b, $CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{$CellContext`b, $CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{$CellContext`a, $CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{$CellContext`b, $CellContext`NP}],
     Editable->False]}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Equal]", 
      RowBox[{"-", "1"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Equal]", 
      RowBox[{"-", "1"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Equal]", "0"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We store the components of the metric tensor", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Equal", "->", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {1, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {2, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {3, $CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, $CellContext`NP}, {4, $CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "We compute the components of ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["g", "ab", "  "],
    AutoSpacing->False],
   $CellContext`g[-$CellContext`a, -$CellContext`b],
   Editable->False]], "Output",
  CellChangeTimes->{3.4572484089309483`*^9}],
 "  and store their values."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MatrixForm", "@", 
  RowBox[{"ToValues", "@", 
   RowBox[{"TableOfComponents", "[", 
    RowBox[{
     RowBox[{"g", "[", 
      RowBox[{"a", ",", "b"}], "]"}], ",", "NP"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "1", "0", "0"},
     {"1", "0", "0", "0"},
     {"0", "0", "0", 
      RowBox[{"-", "1"}]},
     {"0", "0", 
      RowBox[{"-", "1"}], "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MatrixForm", "@", 
  RowBox[{"Inverse", "@", "%"}]}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"0", "1", "0", "0"},
     {"1", "0", "0", "0"},
     {"0", "0", "0", 
      RowBox[{"-", "1"}]},
     {"0", "0", 
      RowBox[{"-", "1"}], "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricInBasis", "[", 
  RowBox[{"g", ",", 
   RowBox[{"-", "NP"}], ",", "%"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", "1"}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{4, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Finally we change the ",
 StyleBox["Rule", "Output"],
 " assignment to the ",
 StyleBox["Set", "Output"],
 " assignment in the ",
 StyleBox["FoldedRule", "Output"],
 " storing all the components of the metric tensor."
}], "MathCaption",
 TextJustification->1.],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "g"}]], "Input"],

Cell["\<\
We also need the components for the elements of the null tetrad with the \
indices downstairs. They are computed from those with the indices upstairs. \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"lNP", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"g", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"lNP", "[", "b", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"nNP", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"g", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"nNP", "[", "b", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"mNP", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"g", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"mNP", "[", "b", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"mNP\[Dagger]", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"g", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"mNP\[Dagger]", "[", "b", "]"}]}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[-$CellContext`a, -$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[$CellContext`b],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[-$CellContext`a, -$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[$CellContext`b],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[-$CellContext`a, -$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[$CellContext`b],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[-$CellContext`a, -$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][$CellContext`b],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, \
-$CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, \
-$CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, \
-$CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, \
-$CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{1, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{2, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{3, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{4, $CellContext`NP}],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{1, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{2, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{3, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{4, $CellContext`NP}],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{1, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{2, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{3, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{4, $CellContext`NP}],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{-$CellContext`a, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{1, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{2, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{3, $CellContext`NP}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`g[{-$CellContext`a, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{4, $CellContext`NP}],
       Editable->False]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "@", 
  RowBox[{"Flatten", "@", 
   RowBox[{"ComponentArray", "@", "%"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{1, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{2, -$CellContext`NP}],
     Editable->False], "\[Equal]", "1"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{3, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{4, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{1, -$CellContext`NP}],
     Editable->False], "\[Equal]", "1"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{2, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{3, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{4, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{1, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{2, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{3, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", "1"}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{1, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{2, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", "1"}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{4, -$CellContext`NP}],
     Editable->False], "\[Equal]", "0"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Equal", "\[Rule]", "ComponentValue"}]}], ";"}]], "Input"],

Cell["\<\
Automatic replacement of the values of the components of the elements of the \
null tetrad.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleToSet", "/@", 
   RowBox[{"{", 
    RowBox[{"lNP", ",", "nNP", ",", "mNP", ",", "mNP\[Dagger]"}], "}"}]}], 
  ";"}]], "Input"],

Cell[TextData[{
 "It is convenient to have a rule which relates  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["l", " ", "a"],
    AutoSpacing->False],
   $CellContext`lNP[$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 ",  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["n", " ", "a"],
    AutoSpacing->False],
   $CellContext`nNP[$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 ",  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["m", " ", "a"],
    AutoSpacing->False],
   $CellContext`mNP[$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 ",  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox[
     OverscriptBox["m", "_"], " ", "a"],
    AutoSpacing->False],
   $CellContext`mNP\[Dagger][$CellContext`a],
   Editable->False]], "Output",
  CellChangeTimes->{3.443600000373357*^9, 3.443600614034006*^9, 
   3.44360110601153*^9, 3.4444578983998013`*^9, 3.444971906279531*^9, 
   3.4450800572925777`*^9, 3.445238951063208*^9, 3.44523920656565*^9, 
   3.447147750178196*^9, 3.447148302618504*^9, 3.4471593828182697`*^9, 
   3.447998369063972*^9, 3.448171329884437*^9, 3.448271854453495*^9, 
   3.448341708076147*^9, 3.449288679662628*^9, 3.449816979737742*^9, 
   3.4504101371642933`*^9, 3.450417087980618*^9, 3.4518175014621572`*^9, 
   3.452242362453743*^9, 3.4524082882882566`*^9, 3.45724482695157*^9}],
 " and the elements of the basis ",
 StyleBox["NP", "Output"],
 ". We compute such a rule next."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"lNP", "[", "a", "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"lNP", "[", "a", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"nNP", "[", "a", "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"nNP", "[", "a", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"mNP", "[", "a", "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"mNP", "[", "a", "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"mNP\[Dagger]", "[", "a", "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"mNP\[Dagger]", "[", "a", "]"}]}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`NP}, $CellContext`a],
      
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`NP}, $CellContext`a],
      
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`NP}, $CellContext`a],
      
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "a"},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`NP}, $CellContext`a],
      
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][{$CellContext`b, $CellContext`NP}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`NP}, $CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`NP}, $CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`NP}, $CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{4, -$CellContext`NP}, $CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Cases", "[", 
     RowBox[{"%", ",", 
      RowBox[{"Basis", "[", 
       RowBox[{
        RowBox[{"{", "__", "}"}], ",", "a_"}], "]"}], ",", "Infinity"}], 
     "]"}]}], "]"}], "//", "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{1, -$CellContext`NP}, $CellContext`a],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{2, -$CellContext`NP}, $CellContext`a],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{3, -$CellContext`NP}, $CellContext`a],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "a"},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{4, -$CellContext`NP}, $CellContext`a],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{
     RowBox[{"HoldPattern", "@", 
      RowBox[{"Rule", "[", 
       RowBox[{"X_", ",", "Y_"}], "]"}]}], "\[RuleDelayed]", 
     RowBox[{"MakeRule", "[", 
      RowBox[{"{", 
       RowBox[{"X", ",", "Y"}], "}"}], "]"}]}]}], "//", "Flatten"}], "//", 
  "Union"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{1, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{2, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{3, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{4, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][$CellContext`a],
       Editable->False]}], "]"}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"BasisToVectors", "=", "%"}], ";"}]], "Input"],

Cell["We need the same rule with the indices downstairs.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"lNP", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"lNP", "[", 
      RowBox[{"-", "a"}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"nNP", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"nNP", "[", 
      RowBox[{"-", "a"}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"mNP", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"mNP", "[", 
      RowBox[{"-", "a"}], "]"}]}]}], ",", 
   RowBox[{
    RowBox[{"mNP\[Dagger]", "[", 
     RowBox[{"-", "a"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{"SeparateBasis", "[", "NP", "]"}], "@", 
     RowBox[{"mNP\[Dagger]", "[", 
      RowBox[{"-", "a"}], "]"}]}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {" "},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[{-$CellContext`b, -$CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[{-$CellContext`b, -$CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {" "},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[{-$CellContext`b, -$CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {$CellContext`b, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {" "},
             {
              StyleBox["b",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][{-$CellContext`b, -$CellContext`NP}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {4, $CellContext`NP}],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][-$CellContext`a],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`NP}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Cases", "[", 
     RowBox[{"%", ",", 
      RowBox[{"Basis", "[", "__", "]"}], ",", "Infinity"}], "]"}]}], "]"}], "//",
   "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {2, $CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[-$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {1, $CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[-$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {4, $CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[-$CellContext`a],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[-$CellContext`a, {3, $CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][-$CellContext`a],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{
     RowBox[{"HoldPattern", "@", 
      RowBox[{"Rule", "[", 
       RowBox[{"X_", ",", "Y_"}], "]"}]}], "\[RuleDelayed]", 
     RowBox[{"MakeRule", "[", 
      RowBox[{"{", 
       RowBox[{"X", ",", "Y"}], "}"}], "]"}]}]}], "//", "Flatten"}], "//", 
  "Union"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ], {1, $CellContext`NP}],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ], {2, $CellContext`NP}],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ], {3, $CellContext`NP}],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            OverscriptBox["m", "_"], 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`mNP\[Dagger][$CellContext`a],
        Editable->False]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ], {4, $CellContext`NP}],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"m", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`mNP[$CellContext`a],
        Editable->False]}]}], "]"}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We combine these relations with the relations found above for the indices \
upstairs.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AppendTo", "[", 
  RowBox[{"BasisToVectors", ",", "%"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{1, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{2, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{3, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              UnderscriptBox[
               UnderscriptBox["a", "_"], "_"]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{4, -$CellContext`NP}, 
       PatternTest[
        Pattern[$CellContext`a, 
         Blank[]], TangentM4`pmQ]],
      Editable->False], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][$CellContext`a],
       Editable->False]}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"HoldPattern", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {
                UnderscriptBox[
                 UnderscriptBox["a", "_"], "_"], 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[
         PatternTest[
          Pattern[$CellContext`a, 
           Blank[]], TangentM4`pmQ], {1, $CellContext`NP}],
        Editable->False], "]"}], "\[RuleDelayed]", 
      RowBox[{"Module", "[", 
       RowBox[{
        RowBox[{"{", "}"}], ",", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"n", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`nNP[$CellContext`a],
         Editable->False]}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"HoldPattern", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {
                UnderscriptBox[
                 UnderscriptBox["a", "_"], "_"], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[
         PatternTest[
          Pattern[$CellContext`a, 
           Blank[]], TangentM4`pmQ], {2, $CellContext`NP}],
        Editable->False], "]"}], "\[RuleDelayed]", 
      RowBox[{"Module", "[", 
       RowBox[{
        RowBox[{"{", "}"}], ",", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"l", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`lNP[$CellContext`a],
         Editable->False]}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"HoldPattern", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {
                UnderscriptBox[
                 UnderscriptBox["a", "_"], "_"], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[
         PatternTest[
          Pattern[$CellContext`a, 
           Blank[]], TangentM4`pmQ], {3, $CellContext`NP}],
        Editable->False], "]"}], "\[RuleDelayed]", 
      RowBox[{"Module", "[", 
       RowBox[{
        RowBox[{"{", "}"}], ",", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {
              OverscriptBox["m", "_"], 
              StyleBox[GridBox[{
                 {"a"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`mNP\[Dagger][$CellContext`a],
          Editable->False]}]}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"HoldPattern", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"e", 
            StyleBox[GridBox[{
               {
                UnderscriptBox[
                 UnderscriptBox["a", "_"], "_"], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]},
               {" ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`Basis[
         PatternTest[
          Pattern[$CellContext`a, 
           Blank[]], TangentM4`pmQ], {4, $CellContext`NP}],
        Editable->False], "]"}], "\[RuleDelayed]", 
      RowBox[{"Module", "[", 
       RowBox[{
        RowBox[{"{", "}"}], ",", 
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"m", 
              StyleBox[GridBox[{
                 {"a"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`mNP[$CellContext`a],
          Editable->False]}]}], "]"}]}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Relation between the spin dyad and the null tetrad", "Subsubsection"],

Cell["\<\
The relation between the elements of the spinor dyad and the elements of the \
null tetrad is as follows\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SpinDyadToNullTetrad", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"\[Sigma]", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "A"}], ",", 
        RowBox[{"-", "A\[Dagger]"}]}], "]"}], 
      RowBox[{"o", "[", "A", "]"}], 
      RowBox[{"Dagger", "@", 
       RowBox[{"o", "[", "A", "]"}]}]}], "\[Equal]", 
     RowBox[{"lNP", "[", "a", "]"}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"\[Sigma]", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "A"}], ",", 
        RowBox[{"-", "A\[Dagger]"}]}], "]"}], 
      RowBox[{"\[Iota]", "[", "A", "]"}], 
      RowBox[{"Dagger", "@", 
       RowBox[{"\[Iota]", "[", "A", "]"}]}]}], "\[Equal]", 
     RowBox[{"nNP", "[", "a", "]"}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"\[Sigma]", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "A"}], ",", 
        RowBox[{"-", "A\[Dagger]"}]}], "]"}], 
      RowBox[{"o", "[", "A", "]"}], 
      RowBox[{"Dagger", "@", 
       RowBox[{"\[Iota]", "[", "A", "]"}]}]}], "\[Equal]", 
     RowBox[{"mNP", "[", "a", "]"}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"\[Sigma]", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "A"}], ",", 
        RowBox[{"-", "A\[Dagger]"}]}], "]"}], 
      RowBox[{"Dagger", "[", 
       RowBox[{
        RowBox[{"o", "[", "A", "]"}], 
        RowBox[{"Dagger", "@", 
         RowBox[{"\[Iota]", "[", "A", "]"}]}]}], "]"}]}], "\[Equal]", 
     RowBox[{"mNP\[Dagger]", "[", "a", "]"}]}]}], "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Using this relation we find and store the components of the soldering form ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Sigma]", 
     RowBox[{" ", 
      RowBox[{"AA", "'"}]}], 
     RowBox[{"a", "   "}]],
    AutoSpacing->False],
   $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
   Editable->False]], "Output",
  CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQBuL7Enb/kha/dGRhcdFJAdKPpmr76S556Vj9wioeRCeb
z54SAaTD/u3wbgHSXYJue3YC6Se2Dv9BtLCrw6UDQHrSRZnLMcteOj7QKttf
AaRDX/2+CKKPvRZ/UQOkaw5tKaoD0ttiol41A+k2KdHWC0B6gg0DX9nyl44H
plYdaADSTh/2Mi4B0ixrWuwYV7x09HI9c1AVSJvFxN8C0eFv0k1PrHzpeEor
xeDwqpeOAVLO/78DaXmBP1NcV7905P/7eguI9vvwcs7LdS8dN/G9ewiiHZ4b
er8H0to5Nl86Nrx0fJ+cknEXSPf5P6/l2/jSsSd2xUwDIP0lRvfvv00vHb8/
X/GJYctLx9/rbCcrbX3peC1I6ocqkJ5Rc/zwxe0vHRluZpgv3PHS0e2Rms0n
IP1Suuei7iGgvsir2duBNADncb9o
   "]],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, -$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, -$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, -$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, -$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\[Dagger]}],
      
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\[Dagger]}],
      
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\[Dagger]}],
      
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, {-$CellContext`A, \
-$CellContext`Dyad}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\[Dagger]}],
      
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, {
      0, -$CellContext`Dyad}, {0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, {
      0, -$CellContext`Dyad}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[{$CellContext`a, $CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, {
      1, -$CellContext`Dyad}, {0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ComponentArray", "@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> 
           {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Equal]", "1"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Equal", "\[Rule]", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "1"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{1, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{2, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{3, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{4, $CellContext`NP}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", "1"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We also need the components of the soldering form with the spacetime index \
downstairs and the spin indices upstairs.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NullTetradToSpinDyad", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"Map", "[", 
     RowBox[{"PutSolderingForm", ",", "SpinDyadToNullTetrad", ",", 
      RowBox[{"{", "2", "}"}]}], "]"}], "/.", 
    RowBox[{"B", "->", "A"}]}], "/.", 
   RowBox[{"B\[Dagger]", "->", "A\[Dagger]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", "A", 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", "A", 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", "A", 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", "A", 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, $CellContext`A\[Dagger]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, $CellContext`A\[Dagger]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, $CellContext`A\[Dagger]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              RowBox[{"A", "'"}]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, $CellContext`A\[Dagger]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, {$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, {$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, {$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][-$CellContext`a, {$CellContext`A, \
$CellContext`Dyad}, {$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`lNP[{$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`nNP[{$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP[{$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{
      1, -$CellContext`NP}, {$CellContext`A, $CellContext`Dyad}, \
{$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{
      2, -$CellContext`NP}, {$CellContext`A, $CellContext`Dyad}, \
{$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{
      3, -$CellContext`NP}, {$CellContext`A, $CellContext`Dyad}, \
{$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{
      4, -$CellContext`NP}, {$CellContext`A, $CellContext`Dyad}, \
{$CellContext`A\[Dagger], $CellContext`Dyad\[Dagger]}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"1", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{1, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{1, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{1, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{1, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{2, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{2, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{2, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"1", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{2, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{3, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"1", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{3, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{3, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{3, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{4, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{4, -$CellContext`NP}, {0, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"1", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{4, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      0, $CellContext`Dyad\[Dagger]}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{4, -$CellContext`NP}, {1, $CellContext`Dyad}, {
      1, $CellContext`Dyad\[Dagger]}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Solve", "[", 
  RowBox[{"%", ",", 
   RowBox[{"Cases", "[", 
    RowBox[{"%", ",", "_\[Sigma]", ",", "Infinity"}], "]"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}]}], "}"}], "}"}]], "Output"]
}, Open  ]],

Cell["We store these values.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{1, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{2, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{3, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {0, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"0", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       0, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "1"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"1", "'"}],
               FontColor->RGBColor[1., 0., 0.]]},
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][{4, -$CellContext`NP}, {1, $CellContext`Dyad}, {
       1, $CellContext`Dyad\[Dagger]}],
      Editable->False], "\[Rule]", "0"}]}], "}"}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Change of the ",
 StyleBox["Rule", "Output"],
 " assignment to the ",
 StyleBox["Set", "Output"],
 " assignment in the folded rule."
}], "MathCaption"],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "\[Sigma]"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Newman-Penrose spin coefficients", "Subsubsection"],

Cell["\<\
 We define the 12 Newmann-Penrose spin coefficients, and their complex \
conjugates:\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Kappa]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Kappa]Dyad", "]"}], "^=", "\"\<\[Kappa]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Kappa]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Kappa]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Kappa]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Kappa]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Kappa]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Kappa]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Epsilon]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Epsilon]Dyad", "]"}], "^=", 
   "\"\<\[Epsilon]\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Epsilon]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Epsilon]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Epsilon]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Epsilon]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Epsilon]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Epsilon]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Pi]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Pi]Dyad", "]"}], "^=", "\"\<\[Pi]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Pi]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Pi]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Pi]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Pi]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Pi]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Pi]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Tau]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Tau]Dyad", "]"}], "^=", "\"\<\[Tau]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Tau]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Tau]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Tau]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Tau]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Tau]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Tau]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Gamma]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Gamma]Dyad", "]"}], "^=", "\"\<\[Gamma]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Gamma]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Gamma]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Gamma]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Gamma]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Gamma]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Gamma]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Nu]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Nu]Dyad", "]"}], "^=", "\"\<\[Nu]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Nu]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Nu]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Nu]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Nu]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Nu]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Nu]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Sigma]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Sigma]Dyad", "]"}], "^=", "\"\<\[Sigma]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Sigma]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Sigma]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Sigma]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Sigma]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Sigma]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Sigma]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Beta]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Beta]Dyad", "]"}], "^=", "\"\<\[Beta]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Beta]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Beta]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Beta]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Beta]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Beta]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Beta]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Mu]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Mu]Dyad", "]"}], "^=", "\"\<\[Mu]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Mu]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Mu]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Mu]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Mu]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Mu]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Mu]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Rho]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Rho]Dyad", "]"}], "^=", "\"\<\[Rho]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Rho]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Rho]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Rho]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Rho]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Rho]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Rho]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Alpha]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Alpha]Dyad", "]"}], "^=", "\"\<\[Alpha]\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Alpha]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Alpha]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Alpha]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Alpha]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Alpha]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Alpha]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"DefSpinor", "[", 
   RowBox[{
    RowBox[{"\[Lambda]Dyad", "[", "]"}], ",", "M4"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Lambda]Dyad", "]"}], "^=", 
   "\"\<\[Lambda]\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "\[Lambda]Dyad\[Dagger]", "]"}], "^=", 
   "\"\<\!\(\*OverscriptBox[\(\[Lambda]\), \(_\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Lambda]Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Lambda]Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Lambda]Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Lambda]Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["This is full the list of Newman-Penrose spin coefficients.", \
"MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPSpinCoefficients", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"\[Alpha]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Beta]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Gamma]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Epsilon]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Kappa]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Lambda]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Mu]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Nu]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Pi]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Rho]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Sigma]Dyad", "[", "]"}], ",", 
    RowBox[{"\[Tau]Dyad", "[", "]"}]}], "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Alpha]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Beta]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Beta]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Gamma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Gamma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Epsilon]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Epsilon]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Kappa]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Kappa]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Lambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Lambda]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Mu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Mu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Nu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Nu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Pi]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Pi]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Rho]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Rho]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Sigma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Sigma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Tau]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Tau]Dyad[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Length"}]], "Input"],

Cell[BoxData["12"], "Output"]
}, Open  ]],

Cell["\<\
The spin coefficients are defined through the derivations of the elements of \
the spin tetrad by means of the following relations.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DyadDerivatives", "=", 
  RowBox[{"{", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"o", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", "  ", 
     RowBox[{
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Epsilon]Dyad", "[", "]"}]}], "-", 
      RowBox[{
       RowBox[{"\[Iota]", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Kappa]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Iota]", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        RowBox[{"\[Iota]", "[", 
         RowBox[{"-", "A"}], "]"}]}], 
       RowBox[{"\[Epsilon]Dyad", "[", "]"}]}], "+", 
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Pi]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"2", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"o", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", "  ", 
     RowBox[{
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Gamma]Dyad", "[", "]"}]}], "-", 
      RowBox[{
       RowBox[{"\[Iota]", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Tau]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"2", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Iota]", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        RowBox[{"\[Iota]", "[", 
         RowBox[{"-", "A"}], "]"}]}], 
       RowBox[{"\[Gamma]Dyad", "[", "]"}]}], "+", 
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Nu]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"3", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"o", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", "  ", 
     RowBox[{
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Beta]Dyad", "[", "]"}]}], "-", 
      RowBox[{
       RowBox[{"\[Iota]", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Sigma]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"3", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Iota]", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        RowBox[{"\[Iota]", "[", 
         RowBox[{"-", "A"}], "]"}]}], 
       RowBox[{"\[Beta]Dyad", "[", "]"}]}], "+", 
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Mu]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"4", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"o", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", "  ", 
     RowBox[{
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Alpha]Dyad", "[", "]"}]}], "-", 
      RowBox[{
       RowBox[{"\[Iota]", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Rho]Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"CDe", "[", 
       RowBox[{"{", 
        RowBox[{"4", ",", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Iota]", "[", 
       RowBox[{"-", "A"}], "]"}], "]"}], "\[Equal]", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        RowBox[{"\[Iota]", "[", 
         RowBox[{"-", "A"}], "]"}]}], 
       RowBox[{"\[Alpha]Dyad", "[", "]"}]}], "+", 
      RowBox[{
       RowBox[{"o", "[", 
        RowBox[{"-", "A"}], "]"}], 
       RowBox[{"\[Lambda]Dyad", "[", "]"}]}]}]}]}], "\[IndentingNewLine]", 
   "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{1, -$CellContext`NP}][
      $CellContext`o[-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{1, -$CellContext`NP}][
      $CellContext`\[Iota][-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{2, -$CellContext`NP}][
      $CellContext`o[-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{2, -$CellContext`NP}][
      $CellContext`\[Iota][-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{3, -$CellContext`NP}][
      $CellContext`o[-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{3, -$CellContext`NP}][
      $CellContext`\[Iota][-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{4, -$CellContext`NP}][
      $CellContext`o[-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {" "},
             {"A"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[{4, -$CellContext`NP}][
      $CellContext`\[Iota][-$CellContext`A]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][-$CellContext`A],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {"A"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[-$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "From this expression we compute and store the components of the spin \
covariant derivative \"intrinsic\" connection (these are the components of \
the \"inner\" Christoffel symbols of the covariant derivative ",
 StyleBox["CDe", "Output"],
 "). "
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"D", GridBox[{
           {"o", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{1, -$CellContext`NP}][
        $CellContext`o[-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"D", GridBox[{
           {"\[Iota]", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{1, -$CellContext`NP}][
        $CellContext`\[Iota][-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[CapitalDelta]", GridBox[{
           {"o", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{2, -$CellContext`NP}][
        $CellContext`o[-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[CapitalDelta]", GridBox[{
           {"\[Iota]", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{2, -$CellContext`NP}][
        $CellContext`\[Iota][-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Delta]", GridBox[{
           {"o", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{3, -$CellContext`NP}][
        $CellContext`o[-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Delta]", GridBox[{
           {"\[Iota]", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{3, -$CellContext`NP}][
        $CellContext`\[Iota][-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         OverscriptBox["\[Delta]", "_"], GridBox[{
           {"o", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{4, -$CellContext`NP}][
        $CellContext`o[-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", "B"},
             {
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]], " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[{-$CellContext`A, -$CellContext`Dyad}, \
$CellContext`B],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         OverscriptBox["\[Delta]", "_"], GridBox[{
           {"\[Iota]", 
            StyleBox[GridBox[{
               {" "},
               {"B"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{4, -$CellContext`NP}][
        $CellContext`\[Iota][-$CellContext`B]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"AChristoffelCDePDDyad", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`AChristoffelCDePDDyad[$CellContext`A, -$CellContext`b, -C], 
   ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"AChristoffelCDePDDyad\[Dagger]", "[", 
    RowBox[{"A\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "C\[Dagger]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`AChristoffelCDePDDyad\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         1, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{1, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         1, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{1, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         2, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{2, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         2, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{2, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         3, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{3, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         3, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{3, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         4, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{4, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
            StyleBox[GridBox[{
               {
                StyleBox["B",
                 FontColor->RGBColor[1., 0., 0.]], " ", " "},
               {" ", 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["A",
                 FontColor->RGBColor[1., 0., 0.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`AChristoffelCDePDDyad[{$CellContext`B, \
$CellContext`Dyad}, {
         4, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`B, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{4, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"TraceBasisDummy", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
        1, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{1, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
       1, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{1, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
        2, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{2, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
       2, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{2, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
        3, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{3, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
       3, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{3, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
        4, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{4, -$CellContext`NP}][
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["A",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
       4, -$CellContext`NP}, {-$CellContext`A, -$CellContext`Dyad}],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{4, -$CellContext`NP}][
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {" "},
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{-$CellContext`A, -$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       1, -$CellContext`NP}, {0, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Kappa]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       1, -$CellContext`NP}, {1, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Epsilon]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Epsilon]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       2, -$CellContext`NP}, {0, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Tau]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Tau]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       2, -$CellContext`NP}, {1, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Gamma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Gamma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       3, -$CellContext`NP}, {0, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Sigma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       3, -$CellContext`NP}, {1, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Beta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Beta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["0",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       4, -$CellContext`NP}, {0, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Rho]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Rho]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
       4, -$CellContext`NP}, {1, -$CellContext`Dyad}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Cases", "[", 
     RowBox[{"%", ",", "_AChristoffelCDePDDyad", ",", "Infinity"}], "]"}]}], 
   "]"}], "//", "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Epsilon]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Gamma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Beta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "->", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Epsilon]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      1, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Gamma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      2, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Beta]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      3, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{1, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Alpha]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"A", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad[{0, $CellContext`Dyad}, {
      4, -$CellContext`NP}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We need the complex conjugate of all these relations.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"Dagger", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Epsilon]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Gamma]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Beta]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Mu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Alpha]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "->", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Epsilon]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {1, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Gamma]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Gamma]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {2, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Beta]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Beta]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {4, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Mu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      1, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Alpha]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Alpha]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"A", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
         StyleBox[GridBox[{
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`AChristoffelCDePDDyad\[Dagger][{
      0, $CellContext`Dyad\[Dagger]}, {3, -$CellContext`NP}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Automatic replacement of the components of  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox[
     RowBox[{"A", "[", 
      RowBox[{"\[EmptyDownTriangle]", ",", 
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[1, 0, 0]]}], "]"}], 
     RowBox[{" ", "bC"}], 
     RowBox[{"A", "  "}]],
    AutoSpacing->False],
   $CellContext`AChristoffelCDePDDyad[$CellContext`A, -$CellContext`b, -C],
   Editable->False]], "Output",
  CellChangeTimes->{{3.457678748806048*^9, 3.457678766406042*^9}}],
 " and ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox[
     RowBox[{
      RowBox[{"A", "[", 
       RowBox[{"\[EmptyDownTriangle]", ",", 
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[1, 0, 0]]}], "]"}], "\[Dagger]"}], 
     RowBox[{"  ", 
      RowBox[{"bC", "'"}]}], 
     RowBox[{
      RowBox[{"A", "'"}], "   "}]],
    AutoSpacing->False],
   $CellContext`AChristoffelCDePDDyad\[Dagger][$CellContext`A\[Dagger], \
-$CellContext`b, -$CellContext`C\[Dagger]],
   Editable->False]], "Output",
  CellChangeTimes->{{3.457678748806048*^9, 3.457678778908732*^9}}],
 "."
}], "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleToSet", "@", "AChristoffelCDePDDyad"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleToSet", "@", "AChristoffelCDePDDyad\[Dagger]"}], 
  ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Components of the Levi-Civita connection in the null tetrad", \
"Subsubsection"],

Cell["\<\
We need to obtain the components of the Levi-Civita connection (Ricci \
rotation coefficients) in the null tetrad. This is done by taking the \
covariant derivative of the set of equations relating the null tetrad \
elements and the spin dyad elements and using the expressions for the \
derivatives of the latter in terms of the Newman-Penrose spin coefficients.\
\>", "Text",
 TextJustification->1.],

Cell["\<\
As we know the relation between the spinor dyad and the null tetrad is \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData["SpinDyadToNullTetrad"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"l", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`nNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"o", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o[$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["\[Iota]", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"m", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP[$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          OverscriptBox["o", "__"], 
          StyleBox[GridBox[{
             {
              RowBox[{"A", "'"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Iota]", 
          StyleBox[GridBox[{
             {"A"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Iota][$CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Sigma]", 
          StyleBox[GridBox[{
             {"a", " ", " "},
             {" ", "A", 
              RowBox[{"A", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["m", "_"], 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`mNP\[Dagger][$CellContext`a],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We take now the derivative of each of these equations and expand them in \
components.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"CDe", "[", 
    RowBox[{"-", "b"}], "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`o[$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {"l", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[-$CellContext`b][
      $CellContext`lNP[$CellContext`a]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`\[Iota][$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[-$CellContext`b][
      $CellContext`nNP[$CellContext`a]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`o[$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {"m", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[-$CellContext`b][
      $CellContext`mNP[$CellContext`a]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][$CellContext`a, -$CellContext`A, -$CellContext`A\
\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[-$CellContext`b][
         $CellContext`\[Iota][$CellContext`A]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {
          OverscriptBox["m", "_"], 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CDe[-$CellContext`b][
      $CellContext`mNP\[Dagger][$CellContext`a]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]], GridBox[{
           {"l", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
        $CellContext`lNP[$CellContext`c]],
       Editable->False], ")"}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]], GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
        $CellContext`nNP[$CellContext`c]],
       Editable->False], ")"}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]], GridBox[{
           {"m", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
        $CellContext`mNP[$CellContext`c]],
       Editable->False], ")"}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][$CellContext`A]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"e", 
          StyleBox[GridBox[{
             {" ", 
              StyleBox["a",
               FontColor->RGBColor[0., 0., 1.]]},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Basis[-$CellContext`c, {$CellContext`a, $CellContext`NP}],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", 
          StyleBox["b",
           FontColor->RGBColor[0., 0., 1.]]], GridBox[{
           {
            OverscriptBox["m", "_"], 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
        $CellContext`mNP\[Dagger][$CellContext`c]],
       Editable->False], ")"}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDePDNP", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCDePDNP[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`lNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`nNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[$CellContext`A]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {"A"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][$CellContext`A],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "A", 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
-$CellContext`A, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {"A"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][$CellContext`A]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`lNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`nNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 RowBox[{"A", "'"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][$CellContext`A\[Dagger]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               RowBox[{"A", "'"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][$CellContext`A\[Dagger]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               RowBox[{"A", "'"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, -$CellContext`A\[Dagger]],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`lNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> 
           {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`nNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CDe[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "All the ocurrences of the spin covariant derivative are now acting on \
scalar functions. Therefore we may replace them by any other covariant \
derivative, in particular ",
 StyleBox["PDNP", "Output"],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"CDe", "\[Rule]", "PDNP"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`lNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"l", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`lNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`nNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`nNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["\[Iota]", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"o", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o[{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"o", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o[{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["\[Iota]", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota]\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP[{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {"m", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP[{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"A", "[", 
             RowBox[{"\[EmptyDownTriangle]", ",", 
              StyleBox["\[ScriptCapitalD]",
               FontColor->RGBColor[1., 0., 0.]]}], "]"}], "\[Dagger]"}], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad\[Dagger][{$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"B", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`B\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"A", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[1., 0., 0.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`AChristoffelCDePDDyad[{$CellContext`A, \
$CellContext`Dyad}, {-$CellContext`b, -$CellContext`NP}, {-$CellContext`B, \
-$CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["B",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`B, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Iota]", 
           StyleBox[GridBox[{
              {
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {
             OverscriptBox["o", "__"], 
             StyleBox[GridBox[{
                {
                 StyleBox[
                  RowBox[{"A", "'"}],
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["o", "__"], 
           StyleBox[GridBox[{
              {
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`o\[Dagger][{$CellContext`A\[Dagger], $CellContext`Dyad\
\[Dagger]}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Sigma]", 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["A",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"A", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma][{$CellContext`a, $CellContext`NP}, \
{-$CellContext`A, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox[
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]], GridBox[{
            {"\[Iota]", 
             StyleBox[GridBox[{
                {
                 StyleBox["A",
                  FontColor->RGBColor[1., 0., 0.]]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
         $CellContext`\[Iota][{$CellContext`A, $CellContext`Dyad}]],
        Editable->False], ")"}]}]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{$CellContext`a, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`mNP\[Dagger][{$CellContext`c, $CellContext`NP}],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           OverscriptBox["m", "_"], 
           StyleBox[GridBox[{
              {
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`mNP\[Dagger][{$CellContext`a, $CellContext`NP}]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Map", "[", 
   RowBox[{"TraceBasisDummy", ",", "%", ",", 
    RowBox[{"{", "2", "}"}]}], "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Mu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Mu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad\[Dagger][],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{"0", "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{"%", ",", 
     RowBox[{"Union", "@", 
      RowBox[{"Cases", "[", 
       RowBox[{"%", ",", "_ChristoffelCDePDNP", ",", "Infinity"}], "]"}]}]}], 
    "]"}], "//", "Flatten"}], "//", "Simplify"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Mu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Mu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", "0"}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
      4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
These are the components of the Levi-Civita connection in the null tetrad \
written in terms of the spin coefficients. We store them.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}], ";"}]], "Input"],

Cell["We show the list of Ricci rotation coefficients.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TensorValues", "@", "ChristoffelCDePDNP"}]], "Input"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        1, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False], "+", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Epsilon]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        1, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        1, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        1, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        2, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False], "+", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Gamma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        2, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        2, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        2, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        3, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False], "+", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        3, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        3, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        3, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        4, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False], "+", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Beta]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        4, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        4, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{1, $CellContext`NP}, {
        4, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        1, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        1, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Epsilon]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        1, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Kappa]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Kappa]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        1, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Kappa]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Kappa]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        2, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        2, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Gamma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        2, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Tau]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Tau]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        2, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Tau]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Tau]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        3, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        3, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        3, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Sigma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Sigma]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        3, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Rho]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        4, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        4, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Beta]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        4, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{2, $CellContext`NP}, {
        4, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Sigma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Sigma]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        1, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Kappa]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Kappa]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        1, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        1, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Epsilon]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        1, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        2, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Tau]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Tau]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        2, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        2, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Gamma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        2, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        3, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Rho]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        3, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        3, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        3, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        4, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Sigma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Sigma]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        4, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        4, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Beta]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{3, $CellContext`NP}, {
        4, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        1, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Kappa]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Kappa]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        1, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        1, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        1, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Epsilon]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        2, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Tau]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Tau]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        2, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        2, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        2, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Gamma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad\[Dagger][],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        3, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Sigma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Sigma]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        3, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        3, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        3, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False], "-", 
       InterpretationBox[
        StyleBox["\[Beta]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        4, -$CellContext`NP}, {1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho]Dyad[],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        4, -$CellContext`NP}, {2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        4, -$CellContext`NP}, {3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[{4, $CellContext`NP}, {
        4, -$CellContext`NP}, {4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False]}], "+", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Beta]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad\[Dagger][],
        Editable->False]}]}]}], "}"}]}], "]"}]], "Output"]
}, Open  ]],

Cell["\<\
From now on, each Ricci rotation coefficient will be automatically replaced \
by its value in terms of the Newman-Penrose spin coefficients.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleToSet", "@", "ChristoffelCDePDNP"}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutation relations", "Subsubsection"],

Cell[TextData[{
 "In this subsection we calculate the commutation coefficients of the \
elements of the null tetrad in terms of the spin coefficients. This is \
achieved by calculating the brackets of the elements of the null tetrad which \
in turn requires the commutation coefficients of the null tetrad elements. \
These coeffients are the components of the tensor",
 StyleBox[" TorsionPDNP", "Output"],
 StyleBox[".", "Input"]
}], "Text",
 TextJustification->1.,
 FontWeight->"Plain"],

Cell[TextData[{
 "The first step is the creation of a folded rule containing the independent \
components of ",
 StyleBox["TorsionPDNP", "Output"],
 " in the null tetrad."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "@", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], "@", 
   RowBox[{"TorsionPDNP", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{1, $CellContext`NP}, {
         1, -$CellContext`NP}, {2, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{1, $CellContext`NP}, {
         1, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{1, $CellContext`NP}, {
         2, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{1, $CellContext`NP}, {
         1, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{1, $CellContext`NP}, {
         2, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{1, $CellContext`NP}, {
         3, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{2, $CellContext`NP}, {
         1, -$CellContext`NP}, {2, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{2, $CellContext`NP}, {
         1, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{2, $CellContext`NP}, {
         2, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{2, $CellContext`NP}, {
         1, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{2, $CellContext`NP}, {
         2, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{2, $CellContext`NP}, {
         3, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{3, $CellContext`NP}, {
         1, -$CellContext`NP}, {2, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{3, $CellContext`NP}, {
         1, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{3, $CellContext`NP}, {
         2, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{3, $CellContext`NP}, {
         1, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{3, $CellContext`NP}, {
         2, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{3, $CellContext`NP}, {
         3, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{4, $CellContext`NP}, {
         1, -$CellContext`NP}, {2, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{4, $CellContext`NP}, {
         1, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{4, $CellContext`NP}, {
         2, -$CellContext`NP}, {3, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
        1, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["1",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{4, $CellContext`NP}, {
         1, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["2",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{4, $CellContext`NP}, {
         2, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            StyleBox["T",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox[GridBox[{
               {
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]], " ", " "},
               {" ", 
                StyleBox["3",
                 FontColor->RGBColor[0., 0., 1.]], 
                StyleBox["4",
                 FontColor->RGBColor[0., 0., 1.]]}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TorsionPDNP[{4, $CellContext`NP}, {
         3, -$CellContext`NP}, {4, -$CellContext`NP}],
        Editable->False]}]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", "0"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        2, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["1",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
        3, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["2",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["3",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["4",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
        4, -$CellContext`NP}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output"]
}, Open  ]],

Cell["\<\
Next we compute the explicit expression for the bracket of each pair of \
elements of the null tetrad.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Bracket", "[", 
    RowBox[{
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "c"}], ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", "a"}], "]"}], ",", 
     RowBox[{"Basis", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "d"}], ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", "a"}], "]"}]}], "]"}], "[", "b", 
   "]"}], "==", 
  RowBox[{"ChangeTorsion", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Bracket", "[", 
      RowBox[{
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "NP"}]}], "}"}], ",", "a"}], "]"}], ",", 
       RowBox[{"Basis", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", "d"}], ",", 
           RowBox[{"-", "NP"}]}], "}"}], ",", "a"}], "]"}]}], "]"}], "[", "b",
      "]"}], ",", "PDNP", ",", "CDe"}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox["T",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionPDNP[$CellContext`b, {-$CellContext`c, \
-$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDePDNP[$CellContext`b, {-$CellContext`c, \
-$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDePDNP[$CellContext`b, {-$CellContext`d, \
-$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox["T",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox[GridBox[{
           {
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionPDNP[{$CellContext`b, $CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDePDNP[{$CellContext`b, $CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", 
          StyleBox["\[ScriptCapitalD]",
           FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
        StyleBox[GridBox[{
           {
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDePDNP[{$CellContext`b, $CellContext`NP}, \
{-$CellContext`d, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Finally we compute and store the components of  ",
 StyleBox["TorsionPDNP", "Input",
  FontWeight->"Plain"],
 StyleBox[" ", "Input"],
 "in the null tetrad."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Nu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Nu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Nu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Nu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {4, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Kappa]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Kappa]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {4, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Sigma]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Lambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Lambda]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {4, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Sigma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Lambda]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Lambda]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
       1, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {4, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", "0"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "@", 
  RowBox[{"ToValues", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Nu]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Nu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Nu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Nu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Nu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Kappa]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Kappa]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Kappa]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Sigma]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Lambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Lambda]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox["\[Sigma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Lambda]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Lambda]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          StyleBox["T",
           FontColor->RGBColor[0., 0., 1.]], 
          StyleBox[GridBox[{
             {
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
       4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Union", "@", 
     RowBox[{"Cases", "[", 
      RowBox[{"%", ",", "_TorsionPDNP", ",", "Infinity"}], "]"}]}]}], "]"}], "//",
   "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Pi]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Pi]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Pi]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Nu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Nu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Nu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Nu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{1, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Kappa]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Kappa]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{2, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Tau]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Rho]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Sigma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox["\[Mu]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["\[Lambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{3, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
      2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox["\[Tau]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Tau]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox["\[Sigma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Sigma]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {1, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox["\[Rho]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Rho]Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {2, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Mu]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Mu]Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{4, $CellContext`NP}, {3, -$CellContext`NP}, {
      4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We store these values.", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleToSet", "@", "TorsionPDNP"}], ";"}]], "Input"],

Cell["\<\
We can now write down the equations arising from the commutation relations. \
First we define a scalar function upon which the differential operators will \
act.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"\[Psi]", "[", "]"}], ",", "M4"}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Psi]", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Psi][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
 "By default, ",
 StyleBox["xTensor`", "Output"],
 " ",
 StyleBox[" ",
  FontSlant->"Italic"],
 "commutes automatically pairs of covariant derivatives when they act on \
scalars. "
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PDNP", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"PDNP", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{"\[Psi]", "[", "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[0., 0., 1.]], "b"], 
     SubscriptBox[
      StyleBox["\[ScriptCapitalD]",
       FontColor->RGBColor[0., 0., 1.]], "a"], "\[Psi]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`PDNP[-$CellContext`b][
    $CellContext`PDNP[-$CellContext`a][
     $CellContext`\[Psi][]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox["T",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionPDNP[$CellContext`c, -$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], "c"], "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[-$CellContext`c][
      $CellContext`\[Psi][]],
     Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "We need to remove this automatic commutation in our calculations.  This is \
done by setting the global variable ",
 StyleBox["$CommuteCovDsOnScalars", "Input"],
 " to ",
 StyleBox["False", "Input"],
 "."
}], "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CommuteCovDsOnScalars", "=", "False"}], ";"}]], "Input"],

Cell[TextData[{
 "Commutation of the operator ",
 StyleBox["PDNP", "Output"],
 " on the scalar ",
 StyleBox["\[Psi]", "Output"],
 "."
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"-", "a"}], "]"}], "@", 
    RowBox[{
     RowBox[{"PDNP", "[", 
      RowBox[{"-", "b"}], "]"}], "@", 
     RowBox[{"\[Psi]", "[", "]"}]}]}], "-", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"-", "b"}], "]"}], "@", 
    RowBox[{
     RowBox[{"PDNP", "[", 
      RowBox[{"-", "a"}], "]"}], "@", 
     RowBox[{"\[Psi]", "[", "]"}]}]}]}], "==", 
  RowBox[{"SortCovDs", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"PDNP", "[", 
      RowBox[{"-", "a"}], "]"}], "@", 
     RowBox[{
      RowBox[{"PDNP", "[", 
       RowBox[{"-", "b"}], "]"}], "@", 
      RowBox[{"\[Psi]", "[", "]"}]}]}], "-", 
    RowBox[{
     RowBox[{"PDNP", "[", 
      RowBox[{"-", "b"}], "]"}], "@", 
     RowBox[{
      RowBox[{"PDNP", "[", 
       RowBox[{"-", "a"}], "]"}], "@", 
      RowBox[{"\[Psi]", "[", "]"}]}]}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "a"], 
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "b"], "\[Psi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`a][
     $CellContext`PDNP[-$CellContext`b][
      $CellContext`\[Psi][]]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "b"], 
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "a"], "\[Psi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`b][
     $CellContext`PDNP[-$CellContext`a][
      $CellContext`\[Psi][]]],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox["T",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionPDNP[$CellContext`c, -$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], "c"], "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[-$CellContext`c][
      $CellContext`\[Psi][]],
     Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell["\<\
The components of this relation in the null tetrad yield the Newman-Penrose \
commutation relations.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`NP}, $CellContext`c],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["a",
          FontColor->RGBColor[0., 0., 1.]]], 
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], "c"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
       $CellContext`PDNP[-$CellContext`c][
        $CellContext`\[Psi][]]],
      Editable->False], ")"}]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "c"},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`NP}, $CellContext`c],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], 
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], "c"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`PDNP[-$CellContext`c][
        $CellContext`\[Psi][]]],
      Editable->False], ")"}]}]}], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox["T",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionPDNP[{$CellContext`c, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`a, -$CellContext`NP}],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox["c",
         FontColor->RGBColor[0., 0., 1.]]], "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{-$CellContext`c, -$CellContext`NP}][
      $CellContext`\[Psi][]],
     Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["a",
        FontColor->RGBColor[0., 0., 1.]]], 
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["b",
        FontColor->RGBColor[0., 0., 1.]]], "\[Psi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
     $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
      $CellContext`\[Psi][]]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["b",
        FontColor->RGBColor[0., 0., 1.]]], 
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["a",
        FontColor->RGBColor[0., 0., 1.]]], "\[Psi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
     $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
      $CellContext`\[Psi][]]],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        StyleBox["T",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], " ", " "},
           {" ", 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionPDNP[{$CellContext`c, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`a, -$CellContext`NP}],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox[
        StyleBox["\[ScriptCapitalD]",
         FontColor->RGBColor[0., 0., 1.]], 
        StyleBox["c",
         FontColor->RGBColor[0., 0., 1.]]], "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{-$CellContext`c, -$CellContext`NP}][
      $CellContext`\[Psi][]],
     Editable->False], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["a",
        FontColor->RGBColor[0., 0., 1.]]], 
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["b",
        FontColor->RGBColor[0., 0., 1.]]], "\[Psi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
     $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
      $CellContext`\[Psi][]]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["b",
        FontColor->RGBColor[0., 0., 1.]]], 
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["a",
        FontColor->RGBColor[0., 0., 1.]]], "\[Psi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
     $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
      $CellContext`\[Psi][]]],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{
      1, $CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{
      2, $CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Psi][]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{
      3, $CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Psi][]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{
      4, $CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Psi][]],
      Editable->False], ")"}]}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "@", 
  RowBox[{"Flatten", "@", 
   RowBox[{"ComponentArray", "@", "%"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`PDNP[{2, -$CellContext`NP}][
          $CellContext`\[Psi][]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`PDNP[{3, -$CellContext`NP}][
          $CellContext`\[Psi][]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`PDNP[{3, -$CellContext`NP}][
          $CellContext`\[Psi][]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Nu]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad\[Dagger][],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", 
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`PDNP[{4, -$CellContext`NP}][
          $CellContext`\[Psi][]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "D", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", 
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`PDNP[{4, -$CellContext`NP}][
          $CellContext`\[Psi][]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Nu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad[],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", 
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`PDNP[{4, -$CellContext`NP}][
          $CellContext`\[Psi][]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We extract the independent equations from this set.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Union", "@", 
     RowBox[{"Cases", "[", 
      RowBox[{"%", ",", 
       RowBox[{
        RowBox[{"PDNP", "[", "_", "]"}], "@", 
        RowBox[{
         RowBox[{"PDNP", "[", "_", "]"}], "@", 
         RowBox[{"\[Psi]", "[", "]"}]}]}], ",", "Infinity"}], "]"}]}]}], 
   "]"}], "//", "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Solve", "::", "svars"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Equations may not give solutions for all \\\"solve\\\" \
variables. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Solve/svars\\\", \
ButtonNote -> \\\"Solve::svars\\\"]\\)\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[CapitalDelta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Nu]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad\[Dagger][],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[CapitalDelta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Nu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad[],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Delta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Mu]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Mu]Dyad[],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
These equations come in complex-conjugate pairs.  Therefore we only need to \
consider a subset of them.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Expand", "//@", "%"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Union", "[", 
   RowBox[{"%", ",", 
    RowBox[{"SameTest", "\[Rule]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"Dagger", "/@", "#1"}], "===", "#2"}], "&"}], ")"}]}]}], 
   "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "//@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[CapitalDelta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Nu]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad\[Dagger][],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Delta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Therefore, the full set of independent commutation relations is\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPCommutators", "=", 
  RowBox[{"%", "/.", 
   RowBox[{"Rule", "->", "Equal"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[CapitalDelta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Nu]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad\[Dagger][],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Delta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Components of the Riemann tensor in terms of the components of the curvature \
spinors\
\>", "Subsubsection"],

Cell["\<\
We are going to write the components of the Riemann tensor in the null tetrad \
in terms of the spinor curvature components expressed in the spin dyad.  \
First we compute folded rules containing the independent components of each \
of the quantities involved. \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "@", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], "@", 
   RowBox[{"PsiCDe", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "PsiCDe"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "@", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], "@", 
   RowBox[{"PsiCDe\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "A\[Dagger]"}], ",", 
     RowBox[{"-", "B\[Dagger]"}], ",", 
     RowBox[{"-", "C\[Dagger]"}], ",", 
     RowBox[{"-", "D\[Dagger]"}]}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{
            RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
            "\[Dagger]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
        1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "PsiCDe\[Dagger]"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AllComponentValues", "@", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], "@", 
   RowBox[{
    RowBox[{"ToBasis", "[", "Dyad", "]"}], "@", 
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"-", "A"}], ",", 
      RowBox[{"-", "B"}], ",", 
      RowBox[{"-", "A\[Dagger]"}], ",", 
      RowBox[{"-", "B\[Dagger]"}]}], "]"}]}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"FoldedRule", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["0",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"0", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False], "\[Rule]", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox["1",
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]], 
               StyleBox[
                RowBox[{"1", "'"}],
                FontColor->RGBColor[1., 0., 0.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
        1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
       Editable->False]}]}], "}"}]}], "]"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "PhiCDe"}]], "Input"],

Cell["\<\
We introduce the scalars which will represent the components of the Weyl \
spinor\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"DefTensor", "[", 
     RowBox[{"#", ",", "M4", ",", 
      RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}], "&"}], "/@", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"PsiCDe0Dyad", "[", "]"}], ",", 
     RowBox[{"PsiCDe1Dyad", "[", "]"}], ",", 
     RowBox[{"PsiCDe2Dyad", "[", "]"}], ",", 
     RowBox[{"PsiCDe3Dyad", "[", "]"}], ",", 
     RowBox[{"PsiCDe4Dyad", "[", "]"}]}], "}"}]}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe0Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe0Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe0Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe0Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe1Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe1Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe1Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe1Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe2Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe2Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe2Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe2Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe3Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe3Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe3Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe3Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe4Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe4Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PsiCDe4Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PsiCDe4Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
The different components of the Weyl spinor are formatted with their \
conventional names.\
\>", "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe0Dyad"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPsi]\), \(0\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe1Dyad"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPsi]\), \(1\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe2Dyad"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPsi]\), \(2\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe3Dyad"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPsi]\), \(3\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe4Dyad"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPsi]\), \(4\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe0Dyad\[Dagger]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[OverscriptBox[\(\[CapitalPsi]\), \(_\)], \
\(0\)]\)\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe1Dyad\[Dagger]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[OverscriptBox[\(\[CapitalPsi]\), \(_\)], \
\(1\)]\)\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe2Dyad\[Dagger]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[OverscriptBox[\(\[CapitalPsi]\), \(_\)], \
\(2\)]\)\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe3Dyad\[Dagger]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[OverscriptBox[\(\[CapitalPsi]\), \(_\)], \
\(3\)]\)\>\""}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "@", "PsiCDe4Dyad\[Dagger]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[OverscriptBox[\(\[CapitalPsi]\), \(_\)], \
\(4\)]\)\>\""}], ";"}]}], "Input"],

Cell[BoxData[
 RowBox[{"SetToRule", "@", "PsiCDe"}]], "Input"],

Cell[BoxData[
 RowBox[{"SetToRule", "@", "PsiCDe\[Dagger]"}]], "Input"],

Cell["Explicit definition of the components of the Weyl spinor.", \
"MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"PsiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}]}], "]"}], "->", 
    RowBox[{"PsiCDe0Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PsiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}]}], "]"}], "->", 
    RowBox[{"PsiCDe1Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PsiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}]}], "]"}], "->", 
    RowBox[{"PsiCDe2Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PsiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}]}], "]"}], "->", 
    RowBox[{"PsiCDe3Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PsiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}]}], "]"}], "->", 
    RowBox[{"PsiCDe4Dyad", "[", "]"}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "1"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe1Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "2"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe2Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "3"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe3Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "1"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe1Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "2"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe2Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "3"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe3Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"Dagger", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe1Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe2Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe3Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad\[Dagger][],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe1Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe2Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{0, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe3Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{
          RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
          "\[Dagger]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe\[Dagger][{1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}, {
      1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad\[Dagger][],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "PsiCDe"}]], "Input"],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "PsiCDe\[Dagger]"}]], "Input"],

Cell["\<\
We introduce the scalars which will represent the components of the Ricci \
spinor.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"DefTensor", "[", 
     RowBox[{"#", ",", "M4", ",", 
      RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}], "&"}], "/@", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"PhiCDe00Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe01Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe02Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe10Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe11Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe12Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe20Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe21Dyad", "[", "]"}], ",", 
     RowBox[{"PhiCDe22Dyad", "[", "]"}]}], "}"}]}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe00Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe00Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe00Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe00Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe01Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe01Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe01Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe01Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe02Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe02Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe02Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe02Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe10Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe10Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe10Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe10Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe11Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe11Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe11Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe11Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe12Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe12Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe12Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe12Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe20Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe20Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe20Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe20Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe21Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe21Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe21Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe21Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe22Dyad", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe22Dyad[], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"PhiCDe22Dyad\[Dagger]", "[", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`PhiCDe22Dyad\[Dagger][], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
The components of the Ricci spinor are formatted with their conventional \
names.\
\>", "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe00Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(00\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe01Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(01\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe02Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(02\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe10Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(10\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe11Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(11\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe12Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(12\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe20Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(20\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe21Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(21\)]\)\>\""}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "PhiCDe22Dyad", "]"}], "^=", 
   "\"\<\!\(\*SubscriptBox[\(\[CapitalPhi]\), \(22\)]\)\>\""}], 
  ";"}]}], "Input"],

Cell[BoxData[
 RowBox[{"SetToRule", "@", "PhiCDe"}]], "Input"],

Cell["Explicit definition of the components of the Ricci spinor.", \
"MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe00Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe01Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe02Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe10Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe20Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe11Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe21Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"0", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe12Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PhiCDe", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "Dyad\[Dagger]"}]}], "}"}]}], "]"}], "\[Rule]", 
    RowBox[{"PhiCDe22Dyad", "[", "]"}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "00"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "01"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {0, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "02"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "10"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad\[Dagger]}, {0, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "20"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "11"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"0", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      0, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "21"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["0",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{0, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "12"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["1",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"1", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{1, -$CellContext`Dyad}, {1, -$CellContext`Dyad}, {
      1, -$CellContext`Dyad\[Dagger]}, {1, -$CellContext`Dyad\[Dagger]}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Rule", "\[Rule]", "ComponentValue"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "@", 
  RowBox[{"Map", "[", 
   RowBox[{"Dagger", ",", "%", ",", 
    RowBox[{"{", "2", "}"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "00"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe00Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "10"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe01Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "20"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe02Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "01"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe10Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "02"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe20Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "11"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe11Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "12"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe21Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "21"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe12Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox["PhiCDe22Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad\[Dagger][],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "\[Rule]", "Equal"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "00"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe00Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "10"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe01Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "20"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe02Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "01"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe10Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "02"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe20Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "11"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe11Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "12"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe21Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "21"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe12Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox["PhiCDe22Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad\[Dagger][],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"PhiCDe00Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe01Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe02Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe10Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe11Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe12Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe20Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe21Dyad\[Dagger]", "[", "]"}], ",", 
      RowBox[{"PhiCDe22Dyad\[Dagger]", "[", "]"}]}], "}"}]}], "]"}], "//", 
  "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe00Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "00"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe01Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "10"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe02Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "20"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe10Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "01"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe01Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe11Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "11"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe12Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "21"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe20Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "02"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe21Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe21Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "12"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox["PhiCDe22Dyad\[Dagger]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad\[Dagger][],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Rule", "\[Rule]", "Set"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "PhiCDe"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", "LambdaCDe", "]"}], "^=", 
   "\"\<\[CapitalLambda]\>\""}], ";"}]], "Input"],

Cell["Computation of the folded rule for the Riemann tensor.", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"AllComponentValues", "@", 
   RowBox[{
    RowBox[{"ToBasis", "[", "NP", "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "RiemannCD"}]], "Input"],

Cell["\<\
Next we show how to write Riemann tensor components in terms of the \
components of curvature spinors\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " ", " ", " "},
         {"a", "b", "c", "d"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->Rational[39, 4]]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
  Editable->False]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "==", 
  RowBox[{"Expand", "@", 
   RowBox[{"Decomposition", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"SeparateSolderingForm", "[", "]"}], "[", "%", "]"}], ",", 
     "Riemann"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           "\[Dagger]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              RowBox[{"A", "'"}], 
              RowBox[{"B", "'"}], 
              RowBox[{"D", "'"}], 
              RowBox[{"C", "'"}]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe\[Dagger][-$CellContext`A\[Dagger], -$CellContext`B\
\[Dagger], -$CellContext`D\[Dagger], -$CellContext`C\[Dagger]],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"A", "B"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`A, -$CellContext`B],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"D", "C"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-D, -C],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "A", 
             RowBox[{"A", "'"}]},
            {"a", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "B", 
             RowBox[{"B", "'"}]},
            {"b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`b, $CellContext`B, $CellContext`B\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "C", 
             RowBox[{"C", "'"}]},
            {"c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`c, C, $CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "D", 
             RowBox[{"D", "'"}]},
            {"d", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`d, D, $CellContext`D\[Dagger]],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"A", "B", 
             RowBox[{"D", "'"}], 
             RowBox[{"C", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[-$CellContext`A, -$CellContext`B, -$CellContext`D\
\[Dagger], -$CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"D", "C"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-D, -C],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"A", "'"}], 
             RowBox[{"B", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`B\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "A", 
             RowBox[{"A", "'"}]},
            {"a", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "B", 
             RowBox[{"B", "'"}]},
            {"b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`b, $CellContext`B, $CellContext`B\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "C", 
             RowBox[{"C", "'"}]},
            {"c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`c, C, $CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "D", 
             RowBox[{"D", "'"}]},
            {"d", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`d, D, $CellContext`D\[Dagger]],
     Editable->False]}], "-", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["\[CapitalLambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`LambdaCDe[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"A", "D"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`A, -D],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"B", "C"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`B, -C],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"A", "'"}], 
             RowBox[{"D", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`D\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"B", "'"}], 
             RowBox[{"C", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`B\[Dagger], \
-$CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "A", 
             RowBox[{"A", "'"}]},
            {"a", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "B", 
             RowBox[{"B", "'"}]},
            {"b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`b, $CellContext`B, $CellContext`B\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "C", 
             RowBox[{"C", "'"}]},
            {"c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`c, C, $CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "D", 
             RowBox[{"D", "'"}]},
            {"d", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`d, D, $CellContext`D\[Dagger]],
     Editable->False]}], "+", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["\[CapitalLambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`LambdaCDe[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"A", "C"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`A, -C],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"B", "D"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`B, -D],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"A", "'"}], 
             RowBox[{"C", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"B", "'"}], 
             RowBox[{"D", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`B\[Dagger], \
-$CellContext`D\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "A", 
             RowBox[{"A", "'"}]},
            {"a", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "B", 
             RowBox[{"B", "'"}]},
            {"b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`b, $CellContext`B, $CellContext`B\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "C", 
             RowBox[{"C", "'"}]},
            {"c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`c, C, $CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "D", 
             RowBox[{"D", "'"}]},
            {"d", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`d, D, $CellContext`D\[Dagger]],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"D", "C", 
             RowBox[{"A", "'"}], 
             RowBox[{"B", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[-D, -
      C, -$CellContext`A\[Dagger], -$CellContext`B\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"A", "B"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][-$CellContext`A, -$CellContext`B],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"D", "'"}], 
             RowBox[{"C", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`D\[Dagger], \
-$CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "A", 
             RowBox[{"A", "'"}]},
            {"a", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "B", 
             RowBox[{"B", "'"}]},
            {"b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`b, $CellContext`B, $CellContext`B\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "C", 
             RowBox[{"C", "'"}]},
            {"c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`c, C, $CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "D", 
             RowBox[{"D", "'"}]},
            {"d", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`d, D, $CellContext`D\[Dagger]],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"A", "B", "D", "C"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[-$CellContext`A, -$CellContext`B, -D, -C],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"A", "'"}], 
             RowBox[{"B", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`A\[Dagger], \
-$CellContext`B\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             RowBox[{"D", "'"}], 
             RowBox[{"C", "'"}]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][-$CellContext`D\[Dagger], \
-$CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "A", 
             RowBox[{"A", "'"}]},
            {"a", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`a, $CellContext`A, $CellContext`A\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "B", 
             RowBox[{"B", "'"}]},
            {"b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`b, $CellContext`B, $CellContext`B\
\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "C", 
             RowBox[{"C", "'"}]},
            {"c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`c, C, $CellContext`C\[Dagger]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", "D", 
             RowBox[{"D", "'"}]},
            {"d", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][-$CellContext`d, D, $CellContext`D\[Dagger]],
     Editable->False]}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "Dyad\[Dagger]", "]"}], "/@", 
  RowBox[{
   RowBox[{"ToBasis", "[", "Dyad", "]"}], "/@", 
   RowBox[{
    RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["c",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["d",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}, \
{-$CellContext`d, -$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{
           RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
           "\[Dagger]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox[
               RowBox[{"A", "'"}],
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"B", "'"}],
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"D", "'"}],
               FontColor->RGBColor[1., 0., 0.]], 
              StyleBox[
               RowBox[{"C", "'"}],
               FontColor->RGBColor[1., 0., 0.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`B\[Dagger], -$CellContext`Dyad\
\[Dagger]}, {-$CellContext`D\[Dagger], -$CellContext`Dyad\[Dagger]}, \
{-$CellContext`C\[Dagger], -$CellContext`Dyad\[Dagger]}],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-$CellContext`A, -$CellContext`Dyad}, \
{-$CellContext`B, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-D, -$CellContext`Dyad}, {-
       C, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`b, -$CellContext`NP}, \
{$CellContext`B, $CellContext`Dyad}, {$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`c, -$CellContext`NP}, {
      C, $CellContext`Dyad}, {$CellContext`C\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`d, -$CellContext`NP}, {
      D, $CellContext`Dyad}, {$CellContext`D\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{-$CellContext`A, -$CellContext`Dyad}, \
{-$CellContext`B, -$CellContext`Dyad}, {-$CellContext`D\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`C\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-D, -$CellContext`Dyad}, {-
       C, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`B\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`b, -$CellContext`NP}, \
{$CellContext`B, $CellContext`Dyad}, {$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`c, -$CellContext`NP}, {
      C, $CellContext`Dyad}, {$CellContext`C\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`d, -$CellContext`NP}, {
      D, $CellContext`Dyad}, {$CellContext`D\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False]}], "-", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["\[CapitalLambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`LambdaCDe[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-$CellContext`A, -$CellContext`Dyad}, {-
       D, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-$CellContext`B, -$CellContext`Dyad}, {-
       C, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`D\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`B\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`C\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`b, -$CellContext`NP}, \
{$CellContext`B, $CellContext`Dyad}, {$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`c, -$CellContext`NP}, {
      C, $CellContext`Dyad}, {$CellContext`C\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`d, -$CellContext`NP}, {
      D, $CellContext`Dyad}, {$CellContext`D\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False]}], "+", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox["\[CapitalLambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`LambdaCDe[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-$CellContext`A, -$CellContext`Dyad}, {-
       C, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-$CellContext`B, -$CellContext`Dyad}, {-
       D, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`C\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`B\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`D\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`b, -$CellContext`NP}, \
{$CellContext`B, $CellContext`Dyad}, {$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`c, -$CellContext`NP}, {
      C, $CellContext`Dyad}, {$CellContext`C\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`d, -$CellContext`NP}, {
      D, $CellContext`Dyad}, {$CellContext`D\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPhi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe[{-D, -$CellContext`Dyad}, {-
       C, -$CellContext`Dyad}, {-$CellContext`A\[Dagger], -$CellContext`Dyad\
\[Dagger]}, {-$CellContext`B\[Dagger], -$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Epsilon]", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon][{-$CellContext`A, -$CellContext`Dyad}, \
{-$CellContext`B, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`D\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`C\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`b, -$CellContext`NP}, \
{$CellContext`B, $CellContext`Dyad}, {$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`c, -$CellContext`NP}, {
      C, $CellContext`Dyad}, {$CellContext`C\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`d, -$CellContext`NP}, {
      D, $CellContext`Dyad}, {$CellContext`D\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalPsi]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe[{-$CellContext`A, -$CellContext`Dyad}, \
{-$CellContext`B, -$CellContext`Dyad}, {-D, -$CellContext`Dyad}, {-
       C, -$CellContext`Dyad}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`A\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`B\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         OverscriptBox["\[Epsilon]", "__"], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Epsilon]\[Dagger][{-$CellContext`D\[Dagger], \
-$CellContext`Dyad\[Dagger]}, {-$CellContext`C\[Dagger], -$CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["A",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"A", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`a, -$CellContext`NP}, \
{$CellContext`A, $CellContext`Dyad}, {$CellContext`A\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["B",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"B", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`b, -$CellContext`NP}, \
{$CellContext`B, $CellContext`Dyad}, {$CellContext`B\[Dagger], \
$CellContext`Dyad\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["C",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"C", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`c, -$CellContext`NP}, {
      C, $CellContext`Dyad}, {$CellContext`C\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Sigma]", 
         StyleBox[GridBox[{
            {" ", 
             StyleBox["D",
              FontColor->RGBColor[1., 0., 0.]], 
             StyleBox[
              RowBox[{"D", "'"}],
              FontColor->RGBColor[1., 0., 0.]]},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma][{-$CellContext`d, -$CellContext`NP}, {
      D, $CellContext`Dyad}, {$CellContext`D\[Dagger], $CellContext`Dyad\
\[Dagger]}],
     Editable->False]}]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "/@", "%"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Flatten", "@", 
   RowBox[{"ComponentArray", "@", "%"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}, {2, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}, {3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
       1, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
       2, -$CellContext`NP}, {3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
       2, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
       3, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
       1, -$CellContext`NP}, {3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
       1, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "00"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}, {3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "02"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
       3, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
       1, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}, {3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "20"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe20Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["1",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}, {3, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
       2, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
       3, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}, {
       2, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["2",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["3",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["4",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}, {
       3, -$CellContext`NP}, {4, -$CellContext`NP}],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Union", "@", 
     RowBox[{"Cases", "[", 
      RowBox[{"%", ",", "_RiemannCD", ",", "Infinity"}], "]"}]}]}], "]"}], "//",
   "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "00"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe00Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad\[Dagger][],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}, {
      2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad[],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}, {
      3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We add the values just found to the folded rule containg the components of \
the Riemann tensor. \
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{"SetToRule", "@", "RiemannCD"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%%", "/.", 
   RowBox[{"Rule", "->", "ComponentValue"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "RiemannCD"}]], "Input"],

Cell["\<\
From previous computations we can obtain and store the values for the \
components of the Ricci tensor.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"AllComponentValues", "@", 
   RowBox[{
    RowBox[{"ToBasis", "[", "NP", "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{"g", "[", 
    RowBox[{"p", ",", "q"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "p"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "q"}]}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"p", "q"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[$CellContext`p, $CellContext`q],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "p", "b", "q"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`p, -$CellContext`b, \
-$CellContext`q],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[{$CellContext`c, $CellContext`NP}, {$CellContext`d, \
$CellContext`NP}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["c",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["d",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`d, -$CellContext`NP}],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TraceBasisDummy", "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, {
     1, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, {
     2, -$CellContext`NP}],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["2",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["1",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, {
     2, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, {
     1, -$CellContext`NP}],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["4",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, {
     3, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, {
     4, -$CellContext`NP}],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["4",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["3",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, {
     4, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}, {
     3, -$CellContext`NP}],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"ComponentArray", "@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "6"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "6"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "6"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "6"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[{4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Equal", "\[Rule]", "ComponentValue"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "RicciCD"}]], "Input"],

Cell["The scalar curvature is worked out in a similar fashion.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciScalarCD", "[", "]"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"ToBasis", "[", "NP", "]"}], "@", 
    RowBox[{"g", "[", 
     RowBox[{"a", ",", "b"}], "]"}]}], 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarCD[],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[{$CellContext`a, $CellContext`NP}, {$CellContext`b, \
$CellContext`NP}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarCD[],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[{$CellContext`a, $CellContext`NP}, {$CellContext`b, \
$CellContext`NP}],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "/@", "%"}], "//", "Expand"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarCD[],
   Editable->False], "\[Equal]", 
  RowBox[{"24", " ", 
   InterpretationBox[
    StyleBox["\[CapitalLambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`LambdaCDe[],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Equal", "\[Rule]", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarCD[],
   Editable->False], "\[Rule]", 
  RowBox[{"24", " ", 
   InterpretationBox[
    StyleBox["\[CapitalLambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`LambdaCDe[],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "RicciScalarCD"}]], "Input"],

Cell["\<\
We may also need the components of the traceless part of the Ricci tensor.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"AllComponentValues", "@", 
   RowBox[{
    RowBox[{"ToBasis", "[", "NP", "]"}], "@", 
    RowBox[{"TFRicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TFRicciCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "==", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"TFRicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "//", "TFRicciToRicci"}], ")"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "6"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[-$CellContext`b, -$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[CapitalLambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`LambdaCDe[],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TFRicciCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "6"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`b, -$CellContext`NP}, {-$CellContext`a, \
-$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[CapitalLambda]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`LambdaCDe[],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {
            StyleBox["a",
             FontColor->RGBColor[0., 0., 1.]], 
            StyleBox["b",
             FontColor->RGBColor[0., 0., 1.]]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToValues", "@", 
  RowBox[{"Flatten", "@", 
   RowBox[{"ComponentArray", "@", "%"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`LambdaCDe[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False]}], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`LambdaCDe[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False]}], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[CapitalLambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`LambdaCDe[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False]}], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Equal", "->", "ComponentValue"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {1, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["1",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{1, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {2, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["2",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{2, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {3, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["3",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{3, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["4",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TFRicciCD[{4, -$CellContext`NP}, {4, -$CellContext`NP}],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", "2"}], " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"RuleToSet", "@", "TFRicciCD"}]], "Input"],

Cell["\<\
This is the complete list of the components of the Riemann tensor. \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPCVComponents", "=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"PsiCDe0Dyad", "[", "]"}], ",", 
    RowBox[{"PsiCDe1Dyad", "[", "]"}], ",", 
    RowBox[{"PsiCDe2Dyad", "[", "]"}], ",", 
    RowBox[{"PsiCDe3Dyad", "[", "]"}], ",", 
    RowBox[{"PsiCDe4Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe00Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe01Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe02Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe10Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe11Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe12Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe20Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe21Dyad", "[", "]"}], ",", 
    RowBox[{"PhiCDe22Dyad", "[", "]"}], ",", 
    RowBox[{"LambdaCDe", "[", "]"}]}], "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "0"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe0Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "1"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe1Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "2"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe2Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "3"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe3Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "4"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe4Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "00"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe00Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "01"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe01Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "02"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe02Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "10"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe10Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "11"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe11Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "12"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe12Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "20"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe20Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "21"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe21Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "22"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe22Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[CapitalLambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`LambdaCDe[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox[" The Newman-Penrose equations", \
"SubsubsectionIcon"]], "Subsubsection"],

Cell["\<\
We now calculate the Riemann tensor with respect to the null tetrad \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}]}], "]"}], "\[Equal]", 
  RowBox[{"ChangeCurvature", "[", 
   RowBox[{
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], ",", "CD", ",", "PDNP"}], "]"}]}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDPDNP", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCDPDNP[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[-$CellContext`d, -$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDNP[$CellContext`f, -$CellContext`b, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"h", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDNP[$CellContext`f, -$CellContext`a, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"h", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "h", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDPDNP[$CellContext`f, -$CellContext`h, \
-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {"h", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[-$CellContext`a][
       $CellContext`ChristoffelCDPDNP[$CellContext`f, -$CellContext`b, \
-$CellContext`c]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[-$CellContext`b][
       $CellContext`ChristoffelCDPDNP[$CellContext`f, -$CellContext`a, \
-$CellContext`c]],
      Editable->False]}], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"ChristoffelCDPDNP", "->", "ChristoffelCDePDNP"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[-$CellContext`d, -$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[$CellContext`f, -$CellContext`b, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"h", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[$CellContext`f, -$CellContext`a, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"h", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "h", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCDePDNP[$CellContext`f, -$CellContext`h, \
-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           StyleBox["T",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox[GridBox[{
              {"h", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TorsionPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[-$CellContext`a][
       $CellContext`ChristoffelCDePDNP[$CellContext`f, -$CellContext`b, \
-$CellContext`c]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[-$CellContext`b][
       $CellContext`ChristoffelCDePDNP[$CellContext`f, -$CellContext`a, \
-$CellContext`c]],
      Editable->False]}], ")"}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["c",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["d",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}, \
{-$CellContext`d, -$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`h, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`h, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "h"},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`b, -$CellContext`NP}, $CellContext`h],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "j"},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`NP}, $CellContext`j],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["a",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["f",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "h", "j"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
       $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
-$CellContext`h, -$CellContext`j]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "h"},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`a, -$CellContext`NP}, $CellContext`h],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"e", 
         StyleBox[GridBox[{
            {" ", "j"},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Basis[{-$CellContext`c, -$CellContext`NP}, $CellContext`j],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["f",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", "h", "j"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
-$CellContext`h, -$CellContext`j]],
      Editable->False], ")"}]}]}]}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {
           StyleBox["a",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["b",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["c",
            FontColor->RGBColor[0., 0., 1.]], 
           StyleBox["d",
            FontColor->RGBColor[0., 0., 1.]]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->Rational[39, 4]]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}, \
{-$CellContext`d, -$CellContext`NP}],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`h, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`h, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         StyleBox["T",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionPDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["a",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["f",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["b",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
       $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`g[{-$CellContext`d, -$CellContext`NP}, {-$CellContext`f, \
-$CellContext`NP}],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox[
         StyleBox["\[ScriptCapitalD]",
          FontColor->RGBColor[0., 0., 1.]], 
         StyleBox["b",
          FontColor->RGBColor[0., 0., 1.]]], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", 
            RowBox[{"\[EmptyDownTriangle]", ",", 
             StyleBox["\[ScriptCapitalD]",
              FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
           StyleBox[GridBox[{
              {
               StyleBox["f",
                FontColor->RGBColor[0., 0., 1.]], " ", " "},
              {" ", 
               StyleBox["a",
                FontColor->RGBColor[0., 0., 1.]], 
               StyleBox["c",
                FontColor->RGBColor[0., 0., 1.]]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
       $CellContext`ChristoffelCDePDNP[{$CellContext`f, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}]],
      Editable->False], ")"}]}]}]}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "/@", "%"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Flatten", "@", 
   RowBox[{"ComponentArray", "@", "%"}]}], ";"}]], "Input"],

Cell["From here we obtain directly the Newman-Penrose equations", \
"MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Union", "[", 
   RowBox[{"Simplify", "//@", 
    RowBox[{"Union", "[", "%", "]"}]}], "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPRicci", "=", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Lambda]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Pi]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Mu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Pi]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Nu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Pi]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Rho]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Kappa]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Sigma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Kappa]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Tau]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Kappa]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Pi]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Nu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Nu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Lambda]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Tau]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Rho]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Epsilon]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Kappa]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Sigma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Lambda]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Mu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Nu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Mu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Rho]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Sigma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Tau]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad\[Dagger][]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Sigma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad\[Dagger][]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Not all the equations in this bunch are independent. To extract the \
independent set of equations we just eliminate a certain set of variables. \
The variables which we are going to eliminate are \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "/@", 
  RowBox[{"Flatten", "@", 
   RowBox[{"Outer", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"PDNP", "[", "#1", "]"}], "@", "#2"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3", ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4", ",", 
         RowBox[{"-", "NP"}]}], "}"}]}], "}"}], ",", "NPSpinCoefficients", 
     ",", "1"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Alpha]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Alpha]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Beta]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Beta]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Gamma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Epsilon]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Epsilon]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Kappa]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Lambda]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Mu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Nu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Pi]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Rho]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Sigma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      OverscriptBox["\[Tau]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Alpha]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Alpha]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Beta]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Beta]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Gamma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Epsilon]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Epsilon]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Kappa]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Lambda]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Mu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Nu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Pi]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Rho]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Sigma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      OverscriptBox["\[Tau]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Alpha]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Alpha]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Beta]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Beta]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Gamma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Epsilon]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Epsilon]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Kappa]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Lambda]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Mu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Nu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Pi]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Rho]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Sigma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      OverscriptBox["\[Tau]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Alpha]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Alpha]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Beta]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Beta]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Gamma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Epsilon]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Epsilon]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Kappa]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Lambda]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Mu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Nu]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Pi]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Rho]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Sigma]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      OverscriptBox["\[Tau]", "_"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad\[Dagger][]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Eliminate", "[", 
    RowBox[{"NPRicci", ",", "%"}], "]"}], "//", "Simplify"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"List", "@@", "%"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Solve", "[", 
    RowBox[{"%", ",", 
     RowBox[{"Union", "@", 
      RowBox[{"Cases", "[", 
       RowBox[{"%", ",", 
        RowBox[{
         RowBox[{"PDNP", "[", "_", "]"}], "@", "_"}], ",", "Infinity"}], 
       "]"}]}]}], "]"}], "//", "Flatten"}], "//", "Simplify"}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Solve", "::", "svars"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Equations may not give solutions for all \\\"solve\\\" \
variables. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Solve/svars\\\", \
ButtonNote -> \\\"Solve::svars\\\"]\\)\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Beta]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"NPRicci", "=", 
   RowBox[{"(", 
    RowBox[{"%", "/.", 
     RowBox[{"Rule", "->", "Equal"}]}], ")"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"FullSimplify", ",", "NPRicci", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "3"}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Beta]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We may need the complex conjugate set in future computations.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPRicci\[Dagger]", "=", 
  RowBox[{"Map", "[", 
   RowBox[{"Dagger", ",", "NPRicci", ",", 
    RowBox[{"{", "2", "}"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       OverscriptBox["\[Epsilon]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Gamma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       OverscriptBox["\[Kappa]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Tau]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       OverscriptBox["\[Pi]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Nu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       OverscriptBox["\[Gamma]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       OverscriptBox["\[Epsilon]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Beta]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       OverscriptBox["\[Kappa]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Sigma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       OverscriptBox["\[Nu]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Mu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       OverscriptBox["\[Pi]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Mu]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       OverscriptBox["\[Tau]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Sigma]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Beta]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Beta]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Gamma]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox[
         OverscriptBox["\[CapitalPsi]", "_"], "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Epsilon]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Alpha]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Kappa]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Rho]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Mu]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad\[Dagger][],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Lambda]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Nu]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox[
         OverscriptBox["\[CapitalPsi]", "_"], "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Lambda]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Pi]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Lambda]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Sigma]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad\[Dagger][],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        OverscriptBox["\[Rho]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad\[Dagger][]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       OverscriptBox["\[Tau]", "_"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad\[Dagger][]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox[
        OverscriptBox["\[CapitalPsi]", "_"], "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad\[Dagger][],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Rho]", "_"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad\[Dagger][]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Bianchi identities", "Subsubsection"],

Cell["\<\
We take next the components of the second Bianchi identity in the null tetrad.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"3", "*", 
   RowBox[{"Antisymmetrize", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "a"}], "]"}], "@", 
      RowBox[{"RiemannCD", "[", 
       RowBox[{
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "c"}], ",", 
        RowBox[{"-", "d"}], ",", 
        RowBox[{"-", "f"}]}], "]"}]}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "}"}]}], "]"}]}], "\[Equal]", "0"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"b", "c", "d", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`a][
       $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"c", "b", "d", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`a][
       $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, \
-$CellContext`d, -$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"a", "c", "d", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`b][
       $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"c", "a", "d", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`b][
       $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, \
-$CellContext`d, -$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"a", "b", "d", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`c][
       $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"b", "a", "d", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`c][
       $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, \
-$CellContext`d, -$CellContext`f]],
      Editable->False]}], ")"}]}], "\[Equal]", "0"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`a][
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`b][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`c][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`f]],
    Editable->False]}], "\[Equal]", "0"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ChangeCovD", "[", 
    RowBox[{"#", ",", "CD", ",", "PDNP"}], "]"}], "&"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
          StyleBox[GridBox[{
             {"h", " ", " "},
             {" ", "c", "f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`f],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "h", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`h, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "h", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`h, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "h", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`h, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"h", "b", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`h, -$CellContext`b, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"h", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`h, -$CellContext`c, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"h", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`h, -$CellContext`c, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`a][
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "b"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`b][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "c"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`c][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`f]],
    Editable->False]}], "\[Equal]", "0"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
          StyleBox[GridBox[{
             {"h", " ", " "},
             {" ", "c", "f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`f],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "f", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`f, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "f", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`f, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "f", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`f, -$CellContext`h],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"c", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDPDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"c", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`a][
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "b"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`b][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "c"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`c][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`f]],
    Editable->False]}], "\[Equal]", "0"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"ChristoffelCDPDNP", "->", "ChristoffelCDePDNP"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
          StyleBox[GridBox[{
             {"h", " ", " "},
             {" ", "c", "f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`f],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "f", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`f, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "f", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`f, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`h],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "f", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`f, -$CellContext`h],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`c, \
-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`a, \
-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"c", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {"h", " ", " "},
            {" ", "b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[$CellContext`h, -$CellContext`b, \
-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"c", "h", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`h, \
-$CellContext`d, -$CellContext`f],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "a"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"b", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`a][
     $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "b"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "c", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`b][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, \
-$CellContext`d, -$CellContext`f]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], "c"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[-$CellContext`c][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`f]],
    Editable->False]}], "\[Equal]", "0"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToBasis", "[", "NP", "]"}], "/@", 
  RowBox[{
   RowBox[{"ToBasis", "[", "NP", "]"}], "/@", "%"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", 
           RowBox[{"\[EmptyDownTriangle]", ",", 
            StyleBox["\[ScriptCapitalD]",
             FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
          StyleBox[GridBox[{
             {
              StyleBox["h",
               FontColor->RGBColor[0., 0., 1.]], " ", " "},
             {" ", 
              StyleBox["c",
               FontColor->RGBColor[0., 0., 1.]], 
              StyleBox["f",
               FontColor->RGBColor[0., 0., 1.]]}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`f, -$CellContext`NP}],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`f, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`f, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`f, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`f, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`f, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`c, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`a, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`a, -$CellContext`NP}, {-$CellContext`b, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`c, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", 
          RowBox[{"\[EmptyDownTriangle]", ",", 
           StyleBox["\[ScriptCapitalD]",
            FontColor->RGBColor[0., 0., 1.]]}], "]"}], 
         StyleBox[GridBox[{
            {
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], " ", " "},
            {" ", 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCDePDNP[{$CellContext`h, $CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`a, -$CellContext`NP}],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["h",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[{-$CellContext`c, -$CellContext`NP}, \
{-$CellContext`h, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["a",
        FontColor->RGBColor[0., 0., 1.]]], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`a, -$CellContext`NP}][
     $CellContext`RiemannCD[{-$CellContext`b, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["b",
        FontColor->RGBColor[0., 0., 1.]]], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["c",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`b, -$CellContext`NP}][
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`c, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox[
       StyleBox["\[ScriptCapitalD]",
        FontColor->RGBColor[0., 0., 1.]], 
       StyleBox["c",
        FontColor->RGBColor[0., 0., 1.]]], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {
             StyleBox["a",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["b",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["d",
              FontColor->RGBColor[0., 0., 1.]], 
             StyleBox["f",
              FontColor->RGBColor[0., 0., 1.]]}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{-$CellContext`c, -$CellContext`NP}][
     $CellContext`RiemannCD[{-$CellContext`a, -$CellContext`NP}, \
{-$CellContext`b, -$CellContext`NP}, {-$CellContext`d, -$CellContext`NP}, \
{-$CellContext`f, -$CellContext`NP}]],
    Editable->False]}], "\[Equal]", "0"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"TraceBasisDummy", "/@", "%"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Flatten", "@", 
   RowBox[{"ComponentArray", "@", "%"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"NPBianchi", "=", 
   RowBox[{"Union", "[", 
    RowBox[{"Simplify", "//@", "%"}], "]"}]}], ";"}]], "Input"],

Cell["Not all the equations here are independent.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "/@", 
  RowBox[{"Flatten", "@", 
   RowBox[{"Outer", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"PDNP", "[", "#1", "]"}], "@", "#2"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3", ",", 
         RowBox[{"-", "NP"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4", ",", 
         RowBox[{"-", "NP"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"NPCVComponents", "[", 
      RowBox[{"[", 
       RowBox[{"Range", "[", 
        RowBox[{"1", ",", "5"}], "]"}], "]"}], "]"}], ",", "1"}], 
    "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Join", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"PDNP", "[", 
      RowBox[{"{", 
       RowBox[{"4", ",", 
        RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
     RowBox[{"LambdaCDe", "[", "]"}]}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "0"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe0Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "2"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe2Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "3"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe3Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", 
      SubscriptBox[
       OverscriptBox["\[CapitalPsi]", "_"], "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad\[Dagger][]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[CapitalLambda]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`LambdaCDe[]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"List", "@@", 
   RowBox[{"Eliminate", "[", 
    RowBox[{"NPBianchi", ",", "%"}], "]"}]}], ";"}]], "Input"],

Cell["The final result for the Bianchi identities is", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPBianchi", "=", 
  RowBox[{"Map", "[", 
   RowBox[{"FullSimplify", ",", "%", ",", 
    RowBox[{"{", "2", "}"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "01"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe01Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             OverscriptBox["\[Alpha]", "_"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha]Dyad\[Dagger][],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox["\[Beta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Beta]Dyad[],
            Editable->False]}], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "22"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe22Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "20"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe20Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Epsilon]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Epsilon]Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             OverscriptBox["\[Epsilon]", "_"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Epsilon]Dyad\[Dagger][],
            Editable->False]}], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPsi]", "4"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPhi]", "00"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PhiCDe00Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPhi]", "21"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PhiCDe21Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "4"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         OverscriptBox["\[Delta]", "_"], 
         SubscriptBox["\[CapitalPhi]", "11"]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`PhiCDe11Dyad[]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"9", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", 
          SubscriptBox["\[CapitalPsi]", "3"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`PsiCDe3Dyad[]],
        Editable->False], ")"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", 
          SubscriptBox["\[CapitalPhi]", "10"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`PhiCDe10Dyad[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], 
          SubscriptBox["\[CapitalPsi]", "2"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`PsiCDe2Dyad[]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["3. The GHP formalism", "Subsection"],

Cell[TextData[{
 StyleBox["The Geroch-Held-Penrose formalism (known as GHP formalism) is a \
particularisation of the Newman-Penrose formalism in which the spin \
coefficients are separated according to their ",
  FontWeight->"Plain"],
 StyleBox["boost ",
  FontWeight->"Plain",
  FontSlant->"Italic"],
 StyleBox["and",
  FontWeight->"Plain"],
 StyleBox[" spin weight. ",
  FontWeight->"Plain",
  FontSlant->"Italic"],
 StyleBox["Also the direcctional derivatives associated to the null tetrad \
are replaced by differential operators which behave well under boost and spin \
transformations and therefore we can speak of the boost and spin weight of \
these operators. This enables us to re-write the basic equations used in the \
Newman-Penrose formalism in terms of boost-weighted quantities only. In this \
section we show how this is accomplished in the framework of ",
  FontWeight->"Plain"],
 StyleBox["Spinors`", "Output",
  FontWeight->"Plain"],
 StyleBox[" and ",
  FontWeight->"Plain"],
 StyleBox["xCoba`", "Output",
  FontWeight->"Plain"],
 StyleBox[" . ",
  FontWeight->"Plain"]
}], "Text",
 TextJustification->1.],

Cell[CellGroupData[{

Cell["Boost and spin weights", "Subsubsection"],

Cell[TextData[{
 "The GHP formalism only admits ",
 StyleBox["weighted quantities",
  FontSlant->"Italic"],
 ". In this subsection we introduce a command, called ",
 StyleBox["GHPWeightOf", "Input"],
 ", which computes the boost and spin weight of any quantity. These are \
integers which are kept in a list.  The command ",
 StyleBox["GHPWeightOf", "Input"],
 " also checks that the GHP weight of the terms of any sum all agree, \
throwing an error otherwise. "
}], "Text",
 TextJustification->1.],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "::", "usage"}], "=", 
   "\"\<GHPWeightOf[expr] returns the GHP boost and spin weight of any \
expression. If the expression is not weighted, then an error message is \
returned\>\""}], ";"}]], "Input"],

Cell["Weights of integers, rationals and complex numbers.", "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "[", 
    RowBox[{"num_", "?", "IntegerQ"}], " ", "]"}], ":=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "[", "_Rational", "]"}], ":=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "[", "_Complex", "]"}], ":=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}]}], ";"}]}], "Input"],

Cell["Weights of products.", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "[", "expr_Times", "]"}], ":=", 
   RowBox[{"Apply", "[", 
    RowBox[{"Plus", ",", 
     RowBox[{"GHPWeightOf", "/@", 
      RowBox[{"Level", "[", 
       RowBox[{"expr", ",", 
        RowBox[{"{", "1", "}"}]}], "]"}]}]}], "]"}]}], ";"}]], "Input"],

Cell["Weights of sums", "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", "expr_Plus"}], ":=", 
   RowBox[{"If", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Length", "@", 
       RowBox[{"Union", "[", 
        RowBox[{"GHPWeightOf", "/@", 
         RowBox[{"Level", "[", 
          RowBox[{"expr", ",", "1"}], "]"}]}], "]"}]}], ">", "1"}], ",", 
     RowBox[{"Throw", "@", 
      RowBox[{"Message", "[", 
       RowBox[{"GHPWeightOf", "::", "Error"}], "]"}]}], ",", 
     RowBox[{"First", "@", 
      RowBox[{"Union", "[", 
       RowBox[{"GHPWeightOf", "/@", 
        RowBox[{"Level", "[", 
         RowBox[{"expr", ",", "1"}], "]"}]}], "]"}]}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "::", "Error"}], "=", 
   "\"\<Inconsistent weights on Plus expression\>\""}], ";"}]}], "Input"],

Cell["Weights of powers", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "[", 
    RowBox[{"expr_", "^", 
     RowBox[{"(", 
      RowBox[{"n_", "?", "IntegerQ"}], ")"}]}], "]"}], ":=", 
   RowBox[{"n", " ", 
    RowBox[{"GHPWeightOf", "@", "expr"}]}]}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Weights of the spin coefficients", "Subsubsection"],

Cell["\<\
Some of the spin coefficients are weighted quantities. These are in the \
following list.\
\>", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GHPSpinCoefficients", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"NPSpinCoefficients", "[", 
     RowBox[{"[", "#", "]"}], "]"}], "&"}], "/@", 
   RowBox[{"Range", "[", 
    RowBox[{"5", ",", "12"}], "]"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[Kappa]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Kappa]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Lambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Lambda]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Mu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Mu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Nu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Nu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Pi]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Pi]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Rho]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Rho]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Sigma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Sigma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Tau]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Tau]Dyad[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
 These are the  spin and boost weights of these quantities : \
\>", "MathCaption",
 TextJustification->1.],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Kappa]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"3", ",", "1"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Lambda]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "3"}], ",", "1"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Mu]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", 
       RowBox[{"-", "1"}]}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Nu]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "3"}], ",", 
       RowBox[{"-", "1"}]}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Pi]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1"}], ",", "1"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Rho]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"1", ",", "1"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Sigma]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "1"}]}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"\[Tau]Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "1"}]}], "}"}]}]}], "}"}], ";"}]], "Input"],

Cell["\<\
From here we compute the weights of the complex conjugate of these quantities\
\
\>", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "/@", "GHPSpinCoefficients"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Kappa]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Kappa]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Lambda]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Lambda]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Mu]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Mu]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Nu]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Nu]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Pi]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Pi]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Rho]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Rho]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Sigma]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Sigma]Dyad\[Dagger][],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     OverscriptBox["\[Tau]", "_"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Tau]Dyad\[Dagger][],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"UpSet", "[", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", "#"}], ",", 
     RowBox[{"Reverse", "@", 
      RowBox[{"GHPWeightOf", "@", 
       RowBox[{"Dagger", "@", "#"}]}]}]}], "]"}], "&"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "3"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "3"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", "3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", "1"}], "}"}]}], "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Weights of the components of the curvature spinors", "Subsubsection"],

Cell["\<\
It is easy to deduce the weights of the components of the curvature spinors. \
These are their values.\
\>", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe0Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"4", ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe1Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"2", ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe2Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe3Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "2"}], ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe4Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "4"}], ",", "0"}], "}"}]}]}], "}"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe0Dyad\[Dagger]", "[", "]"}]}], "^=", 
     RowBox[{"Reverse", "@", 
      RowBox[{"{", 
       RowBox[{"4", ",", "0"}], "}"}]}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe1Dyad\[Dagger]", "[", "]"}]}], "^=", 
     RowBox[{"Reverse", "@", 
      RowBox[{"{", 
       RowBox[{"2", ",", "0"}], "}"}]}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe2Dyad\[Dagger]", "[", "]"}]}], "^=", 
     RowBox[{"Reverse", "@", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0"}], "}"}]}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe3Dyad\[Dagger]", "[", "]"}]}], "^=", 
     RowBox[{"Reverse", "@", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "2"}], ",", "0"}], "}"}]}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PsiCDe4Dyad\[Dagger]", "[", "]"}]}], "^=", 
     RowBox[{"Reverse", "@", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "4"}], ",", "0"}], "}"}]}]}]}], "}"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe00Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"2", ",", "2"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe10Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"0", ",", "2"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe01Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"2", ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe11Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe20Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "2"}], ",", "2"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe21Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "2"}], ",", "0"}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe02Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "2"}]}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe12Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "2"}]}], "}"}]}], ",", 
    RowBox[{
     RowBox[{"GHPWeightOf", "@", 
      RowBox[{"PhiCDe22Dyad", "[", "]"}]}], "^=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "2"}], ",", 
       RowBox[{"-", "2"}]}], "}"}]}]}], "}"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", 
    RowBox[{"LambdaCDe", "[", "]"}]}], "^=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}]}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["GHP differential operators", "Subsubsection"],

Cell[TextData[{
 "The GHP differential operators are constructed by adding some extra terms \
to the four ",
 StyleBox["NP", "Output"],
 " direccional derivatives. The result is a set of four differential \
operators having good transformation properties under spin and boost. "
}], "Text",
 TextJustification->1.,
 FontWeight->"Plain"],

Cell[TextData[{
 "The GHP differential operators can be regarded as the \"components\" of a \
certain differential operator which we define next.  This  operator is not a \
true covariant derivative but just a differential operator acting on scalars, \
 so we need to use some primitive ",
 StyleBox["xAct`", "Output"],
 " instructions in order to create it. We use the notation ",
 Cell[BoxData[
  SubscriptBox["\[CapitalTheta]", "a"]], "Output",
  CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAmIBr5bssvqXjuvqQupB9HqjOe9B9DfW+h8gmkNb/R+I
vp87dWUFkP7ld+1EFZD+sKJl3hwgzbTH7xiI9rrDoz65+aVj++9AGxDN1Rt9
FkQrXd8wYy6QduqQmrwVSO/7eTpgG5Deu82sfTuQvnFf/PshIP1rycwnz4H0
pv8FTpwtLx19X8zlsQPSy6wuxjoAabPbmze6AmmhZ56zZwLpeX5x1zcD6Tnf
RJzSWl867vCfWN8HpJ3OeclMANJCBmZzJwFpmT9qf+YCaYmVth47gHT1vsdm
x4C0x5Stj24BaZVfNfdZ2l46+phr6ugCaSuRB40RQJorio1HuuOl44ngn+Hr
gPQ3Pn2x30A691hJ5x8gLfB600m2TqC7om0S2YF0sbdKLogGAK3csIk=
   "]],
 " as a representation for it."
}], "MathCaption",
 TextJustification->1.],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"CovDQ", "[", "\[CapitalTheta]", "]"}], "^=", "True"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"xAct`xTensor`Private`MakeLinearDerivative", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"\[CapitalTheta]", "[", "A_", "]"}], ",", 
      RowBox[{"\[CapitalTheta]", "[", "A", "]"}]}], "}"}], ",", "True"}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SymmetryGroupOfCovD", "[", "\[CapitalTheta]", "]"}], "^=", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ManifoldOfCovD", "[", "\[CapitalTheta]", "]"}], "^=", 
   RowBox[{"ManifoldOfCovD", "[", "CDe", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"VBundlesOfCovD", "[", "\[CapitalTheta]", "]"}], "^=", 
   RowBox[{"{", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Dagger", "[", 
    RowBox[{
     RowBox[{"\[CapitalTheta]", "[", "A_", "]"}], "[", "expr_", "]"}], "]"}], 
   "^:=", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"DaggerIndex", "[", "A", "]"}], "]"}], "[", 
    RowBox[{"Dagger", "[", "expr", "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SymbolOfCovD", "[", "\[CapitalTheta]", "]"}], "^=", 
   RowBox[{"{", "\"\<\[CapitalTheta]\>\"", "}"}]}], ";"}]}], "Input"],

Cell["Formatting of the GHP operators with their standard symbols.", \
"MathCaption"],

Cell[BoxData[{
 RowBox[{"xTensorFormStop", "[", "CovD", "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[Thorn]\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[Thorn]'\>\""}], "]"}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[Eth]\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"FormatBasis", "[", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], ",", "\"\<\[Eth]'\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"xTensorFormStart", "[", "CovD", "]"}]}], "Input"],

Cell["\<\
The following constants represent respectively generic boost and spin weights.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "w", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol \"\>", "\[InvisibleSpace]", "w", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefConstantSymbol, ": Defining ", 
   "constant symbol ", $CellContext`w, ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "z", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol \"\>", "\[InvisibleSpace]", "z", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefConstantSymbol, ": Defining ", 
   "constant symbol ", $CellContext`z, ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell["We introduce a complex scalar of weight (p,q).", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefSpinor", "[", 
  RowBox[{
   RowBox[{"\[Eta]", "[", 
    RowBox[{"LI", "@", 
     RowBox[{"{", 
      RowBox[{"p", ",", "q"}], "}"}]}], "]"}], ",", "M4"}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Eta]", "[", 
    RowBox[{"LI", "[", 
     RowBox[{"{", 
      RowBox[{"p", ",", "q"}], "}"}], "]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Eta][
    xAct`xTensor`LI[{$CellContext`p, $CellContext`q}]], ". ", ""],
  Editable->False]], "Print"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Eta]\[Dagger]", "[", 
    RowBox[{"LI", "[", 
     RowBox[{"{", 
      RowBox[{"p", ",", "q"}], "}"}], "]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Eta]\[Dagger][
    xAct`xTensor`LI[{$CellContext`p, $CellContext`q}]], ". ", ""],
  Editable->False]], "Print"]
}, Open  ]],

Cell[TextData[{
 "We specify the weight of the scalar quantity ",
 Cell[BoxData[
  StyleBox[
   SubsuperscriptBox["\[Eta]", " ", 
    RowBox[{"{", 
     RowBox[{"w", ",", " ", "z"}], "}"}]],
   AutoSpacing->False]], "Output",
  CellChangeTimes->{{3.424602713885338*^9, 3.42460276070536*^9}, 
    3.424602810071467*^9, {3.424602864476454*^9, 3.424602872054472*^9}, {
    3.424602923793524*^9, 3.424602926937171*^9}, 3.424603733735276*^9, 
    3.424604177361084*^9, 3.4246042277494717`*^9, 3.4246053245022573`*^9, 
    3.424605562886141*^9, {3.4246057526364*^9, 3.4246057598558702`*^9}, 
    3.424605899660397*^9, {3.424605975024331*^9, 3.424605986433363*^9}, 
    3.424668181220231*^9, {3.424668375406006*^9, 3.424668381044478*^9}, 
    3.424674281118836*^9, 3.424674362073228*^9, {3.424674528370613*^9, 
    3.424674538654908*^9}, {3.4246746135989122`*^9, 3.424674640331546*^9}, 
    3.424675045908537*^9, 3.424776294307295*^9, 3.424781053667054*^9, 
    3.4247919520979223`*^9, 3.4249638047210903`*^9, 3.425017532462819*^9, 
    3.4253829432418528`*^9, 3.4254659003126793`*^9, 3.425468644706599*^9, 
    3.4254737710847073`*^9, 3.4254970585513163`*^9, 3.425550915128848*^9, 
    3.4255794484667873`*^9, 3.425621782855205*^9, 3.42570777312621*^9, 
    3.425790306536113*^9, 3.4258790562580643`*^9, 3.426801864417285*^9, 
    3.427103550912269*^9, 3.427260594807111*^9, 3.4272635648685093`*^9}],
 " (note that the integers w and z are zero weighted). "
}], "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", 
    RowBox[{"\[Eta]", "[", 
     RowBox[{"LI", "@", 
      RowBox[{"{", 
       RowBox[{"w", ",", "z"}], "}"}]}], "]"}]}], "^=", 
   RowBox[{"{", 
    RowBox[{"w", ",", "z"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", "w"}], "^=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}]}], ";", 
  RowBox[{
   RowBox[{"GHPWeightOf", "@", "z"}], "^=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}]}], ";"}]}], "Input"],

Cell[TextData[{
 "The action of the differential operator ",
 Cell[BoxData[
  SubscriptBox["\[CapitalTheta]", "a"]], "Output",
  CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAmIBr5bssvqXjuvqQupB9HqjOe9B9DfW+h8gmkNb/R+I
vp87dWUFkP7ld+1EFZD+sKJl3hwgzbTH7xiI9rrDoz65+aVj++9AGxDN1Rt9
FkQrXd8wYy6QduqQmrwVSO/7eTpgG5Deu82sfTuQvnFf/PshIP1rycwnz4H0
pv8FTpwtLx19X8zlsQPSy6wuxjoAabPbmze6AmmhZ56zZwLpeX5x1zcD6Tnf
RJzSWl867vCfWN8HpJ3OeclMANJCBmZzJwFpmT9qf+YCaYmVth47gHT1vsdm
x4C0x5Stj24BaZVfNfdZ2l46+phr6ugCaSuRB40RQJorio1HuuOl44ngn+Hr
gPQ3Pn2x30A691hJ5x8gLfB600m2TqC7om0S2YF0sbdKLogGAK3csIk=
   "]],
 " on a weighted scalar ",
 Cell[BoxData[
  RowBox[{" ", 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Eta]", " ", 
      RowBox[{"{", 
       RowBox[{"w", ",", " ", "z"}], "}"}]],
     AutoSpacing->False],
    $CellContext`\[Eta][
     xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
    Editable->False]}]], "Output",
  CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB2IBr5bssvqXjuvqQupB9HqjOe9B9DfW+h8gmkNb/R+I
vp87dWUFkP7ld+1EFZD+sKJl3hwgzbTH7xiI9rrDoz65+aVj++9AGxDN1Rt9
FkQrXd8wYy6QduqQmrwVSO/7eTpgG5Deu82sfTuQvnFf/PshIP1rycwnz4H0
pv8FTpwtLx19X8zlsQPSy6wuxjoAabPbmze6AmmhZ56zZwLpeX5x1zcD6Tnf
RJzSWl867vCfWN8HpJ3OeclMANJCBmZzJwFpmT9qf+YCaYmVth47gHT1vsdm
x4C0x5Stj24BaZVfNfdZ2l46+phr6ugCaSuRB40RQJorio1HuuOl44ngn+Hr
gPQ3Pn2x30A691hJ5x8gDQCKuKZm
   "]],
 " is, by definition "
}], "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"\[CapitalTheta]", "[", 
      RowBox[{"{", 
       RowBox[{"1", ",", " ", 
        RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
     RowBox[{"\[Eta]", "[", 
      RowBox[{"LI", "[", 
       RowBox[{"{", 
        RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}], " ", "->", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{"w", "*", 
           RowBox[{"\[Epsilon]Dyad", "[", "]"}]}], ")"}]}], " ", "-", " ", 
        RowBox[{"z", "*", 
         RowBox[{"\[Epsilon]Dyad\[Dagger]", "[", "]"}]}]}], ")"}], "*", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}]}], "+", " ", 
     RowBox[{
      RowBox[{"PDNP", "[", 
       RowBox[{"{", 
        RowBox[{"1", ",", " ", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}]}]}], ",", 
   " ", "\n", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"\[CapitalTheta]", "[", 
      RowBox[{"{", 
       RowBox[{"2", ",", " ", 
        RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
     RowBox[{"\[Eta]", "[", 
      RowBox[{"LI", "[", 
       RowBox[{"{", 
        RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}], " ", "->", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{"w", "*", 
           RowBox[{"\[Gamma]Dyad", "[", "]"}]}], ")"}]}], "-", " ", 
        RowBox[{"z", "*", 
         RowBox[{"\[Gamma]Dyad\[Dagger]", "[", "]"}]}]}], ")"}], "*", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}]}], "+", 
     RowBox[{
      RowBox[{"PDNP", "[", 
       RowBox[{"{", 
        RowBox[{"2", ",", " ", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}]}]}], ",", 
   " ", "\n", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"\[CapitalTheta]", "[", 
      RowBox[{"{", 
       RowBox[{"3", ",", " ", 
        RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
     RowBox[{"\[Eta]", "[", 
      RowBox[{"LI", "[", 
       RowBox[{"{", 
        RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}], " ", "->", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{"z", "*", 
           RowBox[{"\[Alpha]Dyad\[Dagger]", "[", "]"}]}], ")"}]}], "-", 
        RowBox[{"w", "*", 
         RowBox[{"\[Beta]Dyad", "[", "]"}]}]}], ")"}], "*", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}]}], "+", 
     RowBox[{
      RowBox[{"PDNP", "[", 
       RowBox[{"{", 
        RowBox[{"3", ",", " ", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}]}]}], ",", 
   " ", "\n", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"\[CapitalTheta]", "[", 
      RowBox[{"{", 
       RowBox[{"4", ",", " ", 
        RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
     RowBox[{"\[Eta]", "[", 
      RowBox[{"LI", "[", 
       RowBox[{"{", 
        RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}], " ", "->", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          RowBox[{"w", "*", 
           RowBox[{"\[Alpha]Dyad", "[", "]"}]}], ")"}]}], "-", " ", 
        RowBox[{"z", "*", 
         RowBox[{"\[Beta]Dyad\[Dagger]", "[", "]"}]}]}], ")"}], "*", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}]}], "+", 
     RowBox[{
      RowBox[{"PDNP", "[", 
       RowBox[{"{", 
        RowBox[{"4", ",", " ", 
         RowBox[{"-", "NP"}]}], "}"}], "]"}], "[", 
      RowBox[{"\[Eta]", "[", 
       RowBox[{"LI", "[", 
        RowBox[{"{", 
         RowBox[{"w", ",", " ", "z"}], "}"}], "]"}], "]"}], "]"}]}]}]}], 
  "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"GHPDerivativesToFrame", "=", "%"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"Rule", "\[Rule]", "Equal"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Cases", "[", 
     RowBox[{"%", ",", 
      RowBox[{
       RowBox[{"PDNP", "[", "_", "]"}], "@", "_"}], ",", "Infinity"}], 
     "]"}]}], "]"}], "//", "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"FrameToGHPDerivatives", "=", "%"}], ";"}]], "Input"],

Cell["\<\
We transform previous rules in delayed rules with patterns which can be used \
for any weighted quantity\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"GHPDerivativesToFrame", "/.", 
   RowBox[{"Rule", "->", "RuleDelayed"}]}], "/.", 
  RowBox[{
   RowBox[{"\[Eta]", "[", 
    RowBox[{"LI", "@", 
     RowBox[{"{", 
      RowBox[{"w", ",", "z"}], "}"}]}], "]"}], "->", "\[Eta]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][$CellContext`\[Eta]],
     
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][$CellContext`\[Eta]],
     
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][$CellContext`\[Eta]],
     
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][$CellContext`\[Eta]],
     
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", "comp__", "]"}], "@", "\[Eta]"}], ":>", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", "comp", "]"}], "@", 
    "\[Eta]_"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GHPToNP", "=", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"%", "/.", 
     RowBox[{"w", ":>", 
      RowBox[{"First", "@", 
       RowBox[{"GHPWeightOf", "@", "\[Eta]"}]}]}]}], "/.", 
    RowBox[{"z", ":>", 
     RowBox[{"Last", "@", 
      RowBox[{"GHPWeightOf", "@", "\[Eta]"}]}]}]}], ")"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"First", "[", 
           RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"Last", "[", 
          RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"First", "[", 
           RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"Last", "[", 
          RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"Last", "[", 
           RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"First", "[", 
          RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"First", "[", 
           RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"Last", "[", 
          RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"FrameToGHPDerivatives", "/.", 
   RowBox[{"Rule", "->", "RuleDelayed"}]}], "/.", 
  RowBox[{
   RowBox[{"\[Eta]", "[", 
    RowBox[{"LI", "@", 
     RowBox[{"{", 
      RowBox[{"w", ",", "z"}], "}"}]}], "]"}], "->", "\[Eta]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][$CellContext`\[Eta]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][$CellContext`\[Eta]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][$CellContext`\[Eta]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Eta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][$CellContext`\[Eta]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{
   RowBox[{
    RowBox[{"PDNP", "[", "comp__", "]"}], "@", "\[Eta]"}], ":>", 
   RowBox[{
    RowBox[{"PDNP", "[", "comp", "]"}], "@", "\[Eta]_"}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["z",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`z,
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPToGHP", "=", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"%", "/.", 
     RowBox[{"w", ":>", 
      RowBox[{"First", "@", 
       RowBox[{"GHPWeightOf", "@", "\[Eta]"}]}]}]}], "/.", 
    RowBox[{"z", ":>", 
     RowBox[{"Last", "@", 
      RowBox[{"GHPWeightOf", "@", "\[Eta]"}]}]}]}], ")"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"First", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      RowBox[{"Last", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"First", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      RowBox[{"Last", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"Last", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      RowBox[{"First", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Eta]_"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      Pattern[$CellContext`\[Eta], 
       Blank[]]],
     Editable->False], "\[RuleDelayed]", 
    RowBox[{
     RowBox[{
      RowBox[{"First", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", "\[Eta]"}], "+", 
     RowBox[{
      RowBox[{"Last", "[", 
       RowBox[{"GHPWeightOf", "[", "\[Eta]", "]"}], "]"}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", "\[Eta]"}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][$CellContext`\[Eta]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Each of the quantities ",
 Cell[BoxData[
  RowBox[{"\[Thorn]", 
   SubsuperscriptBox["\[Eta]", " ", 
    RowBox[{"{", 
     RowBox[{"w", ",", " ", "z"}], "}"}]]}]], "Output",
  CellChangeTimes->{3.42460184609977*^9, 3.4246681812094793`*^9, 
   3.424776294298366*^9, 3.4247810402212877`*^9, 3.424791952089621*^9, 
   3.424963804641203*^9, 3.425017532422454*^9, 3.425382931026432*^9, 
   3.42546544701396*^9, 3.425465878441057*^9, 3.425468644694871*^9, 
   3.425473771073071*^9, 3.4254970585399942`*^9, 3.425550915120521*^9, 
   3.425579448430015*^9, 3.42562178284708*^9, 3.425707773118067*^9, 
   3.425790306522882*^9, 3.425879056249755*^9, 3.4268018543998213`*^9, 
   3.427103522745451*^9, 3.427218423536128*^9, 3.427231157128652*^9, 
   3.427231824135481*^9}],
 ", ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"\[Thorn]", "'"}], 
   SubsuperscriptBox["\[Eta]", " ", 
    RowBox[{"{", 
     RowBox[{"w", ",", " ", "z"}], "}"}]]}]], "Output",
  CellChangeTimes->{3.42460184609977*^9, 3.4246681812094793`*^9, 
   3.424776294298366*^9, 3.4247810402212877`*^9, 3.424791952089621*^9, 
   3.424963804641203*^9, 3.425017532422454*^9, 3.425382931026432*^9, 
   3.42546544701396*^9, 3.425465878441057*^9, 3.425468644694871*^9, 
   3.425473771073071*^9, 3.4254970585399942`*^9, 3.425550915120521*^9, 
   3.425579448430015*^9, 3.42562178284708*^9, 3.425707773118067*^9, 
   3.425790306522882*^9, 3.425879056249755*^9, 3.4268018543998213`*^9, 
   3.427103522745451*^9, 3.427218423536128*^9, 3.427231157128652*^9, 
   3.427231824135481*^9}],
 ", ",
 Cell[BoxData[
  RowBox[{"\[Eth]", 
   SubsuperscriptBox["\[Eta]", " ", 
    RowBox[{"{", 
     RowBox[{"w", ",", " ", "z"}], "}"}]]}]], "Output",
  CellChangeTimes->{3.42460184609977*^9, 3.4246681812094793`*^9, 
   3.424776294298366*^9, 3.4247810402212877`*^9, 3.424791952089621*^9, 
   3.424963804641203*^9, 3.425017532422454*^9, 3.425382931026432*^9, 
   3.42546544701396*^9, 3.425465878441057*^9, 3.425468644694871*^9, 
   3.425473771073071*^9, 3.4254970585399942`*^9, 3.425550915120521*^9, 
   3.425579448430015*^9, 3.42562178284708*^9, 3.425707773118067*^9, 
   3.425790306522882*^9, 3.425879056249755*^9, 3.4268018543998213`*^9, 
   3.427103522745451*^9, 3.427218423536128*^9, 3.427231157128652*^9, 
   3.427231824135481*^9}],
 ", ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"\[Eth]", "'"}], 
   SubsuperscriptBox["\[Eta]", " ", 
    RowBox[{"{", 
     RowBox[{"w", ",", " ", "z"}], "}"}]]}]], "Output",
  CellChangeTimes->{3.42460184609977*^9, 3.4246681812094793`*^9, 
   3.424776294298366*^9, 3.4247810402212877`*^9, 3.424791952089621*^9, 
   3.424963804641203*^9, 3.425017532422454*^9, 3.425382931026432*^9, 
   3.42546544701396*^9, 3.425465878441057*^9, 3.425468644694871*^9, 
   3.425473771073071*^9, 3.4254970585399942`*^9, 3.425550915120521*^9, 
   3.425579448430015*^9, 3.42562178284708*^9, 3.425707773118067*^9, 
   3.425790306522882*^9, 3.425879056249755*^9, 3.4268018543998213`*^9, 
   3.427103522745451*^9, 3.427218423536128*^9, 3.427231157128652*^9, 
   3.427231824135481*^9}],
 " is a spin weighted quantity. These are the specific values for the weights \
of each of them."
}], "MathCaption",
 TextJustification->1.],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", 
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}]}], "^=", 
   RowBox[{"{", 
    RowBox[{"1", ",", "1"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", 
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}]}], "^=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     RowBox[{"-", "1"}]}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", 
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}]}], "^=", 
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"-", "1"}]}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "@", 
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}]}], "^=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", "1"}], "}"}]}], ";"}]}], "Input"],

Cell["\<\
We also need to teach GHPWeightOf how to behave when acting on differential \
expressions.\
\>", "MathCaption",
 TextJustification->1.],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"GHPWeightOf", "[", 
    RowBox[{
     RowBox[{"\[CapitalTheta]", "[", 
      RowBox[{"ind_", "?", "CIndexQ"}], "]"}], "@", "sc_"}], "]"}], ":=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"Part", "[", 
       RowBox[{
        RowBox[{"GHPWeightOf", "@", 
         RowBox[{"\[CapitalTheta]", "[", "ind", "]"}]}], ",", "1"}], "]"}], 
      "+", 
      RowBox[{"Part", "[", 
       RowBox[{
        RowBox[{"GHPWeightOf", "@", "sc"}], ",", "1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"Part", "[", 
       RowBox[{
        RowBox[{"GHPWeightOf", "@", 
         RowBox[{"\[CapitalTheta]", "[", "ind", "]"}]}], ",", "2"}], "]"}], 
      "+", 
      RowBox[{"Part", "[", 
       RowBox[{
        RowBox[{"GHPWeightOf", "@", "sc"}], ",", "2"}], "]"}]}]}], "}"}]}], 
  ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Commutation of the GHP differential operators", "Subsubsection"],

Cell[TextData[{
 "In this subsection we work out the commutation relations of each of the GHP \
operators. The commutator of two GHP operators is a new differential operator \
which acts on a generic spin weighted scalar ",
 Cell[BoxData[
  StyleBox[
   SubsuperscriptBox["\[Eta]", " ", 
    RowBox[{"{", 
     RowBox[{"w", ",", " ", "z"}], "}"}]],
   AutoSpacing->False]], "Output",
  CellChangeTimes->{{3.424602713885338*^9, 3.42460276070536*^9}, 
    3.424602810071467*^9, {3.424602864476454*^9, 3.424602872054472*^9}, {
    3.424602923793524*^9, 3.424602926937171*^9}, 3.424603733735276*^9, 
    3.424604177361084*^9, 3.4246042277494717`*^9, 3.4246053245022573`*^9, 
    3.424605562886141*^9, {3.4246057526364*^9, 3.4246057598558702`*^9}, 
    3.424605899660397*^9, {3.424605975024331*^9, 3.424605986433363*^9}, 
    3.424668181220231*^9, {3.424668375406006*^9, 3.424668381044478*^9}, 
    3.424674281118836*^9, 3.424674362073228*^9, {3.424674528370613*^9, 
    3.424674538654908*^9}, {3.4246746135989122`*^9, 3.424674640331546*^9}, 
    3.424675045908537*^9, 3.424776294307295*^9, 3.424781053667054*^9, 
    3.4247919520979223`*^9, 3.4249638047210903`*^9, 3.425017532462819*^9, 
    3.4253829432418528`*^9, 3.4254659003126793`*^9, 3.425468644706599*^9, 
    3.4254737710847073`*^9, 3.4254970585513163`*^9, 3.425550915128848*^9, 
    3.4255794484667873`*^9, 3.425621782855205*^9, 3.42570777312621*^9, 
    3.425790306536113*^9, 3.4258790562580643`*^9, 3.426801864417285*^9, 
    3.427103550912269*^9, 3.427260594807111*^9, 3.4272635648685093`*^9}],
 ".   "
}], "Text",
 FontWeight->"Plain"],

Cell["This is the commutation between two generic GHP operators.", \
"MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[CapitalTheta]", "[", 
    RowBox[{"-", 
     RowBox[{"{", 
      RowBox[{"a", ",", "NP"}], "}"}]}], "]"}], "@", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"b", ",", "NP"}], "}"}]}], "]"}], "@", 
    RowBox[{"\[Eta]", "[", 
     RowBox[{"LI", "@", 
      RowBox[{"{", 
       RowBox[{"w", ",", "z"}], "}"}]}], "]"}]}]}], "-", 
  RowBox[{
   RowBox[{"\[CapitalTheta]", "[", 
    RowBox[{"-", 
     RowBox[{"{", 
      RowBox[{"b", ",", "NP"}], "}"}]}], "]"}], "@", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"-", 
      RowBox[{"{", 
       RowBox[{"a", ",", "NP"}], "}"}]}], "]"}], "@", 
    RowBox[{"\[Eta]", "[", 
     RowBox[{"LI", "@", 
      RowBox[{"{", 
       RowBox[{"w", ",", "z"}], "}"}]}], "]"}]}]}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[CapitalTheta]", 
      StyleBox["a",
       FontColor->RGBColor[0., 0., 1.]]], 
     SubscriptBox["\[CapitalTheta]", 
      StyleBox["b",
       FontColor->RGBColor[0., 0., 1.]]], GridBox[{
       {"\[Eta]", 
        StyleBox[GridBox[{
           {
            RowBox[{"{", 
             RowBox[{"w", ",", " ", "z"}], "}"}]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[CapitalTheta][{-$CellContext`a, -$CellContext`NP}][
    $CellContext`\[CapitalTheta][{-$CellContext`b, -$CellContext`NP}][
     $CellContext`\[Eta][
      xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[CapitalTheta]", 
      StyleBox["b",
       FontColor->RGBColor[0., 0., 1.]]], 
     SubscriptBox["\[CapitalTheta]", 
      StyleBox["a",
       FontColor->RGBColor[0., 0., 1.]]], GridBox[{
       {"\[Eta]", 
        StyleBox[GridBox[{
           {
            RowBox[{"{", 
             RowBox[{"w", ",", " ", "z"}], "}"}]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->Rational[39, 4]]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[CapitalTheta][{-$CellContext`b, -$CellContext`NP}][
    $CellContext`\[CapitalTheta][{-$CellContext`a, -$CellContext`NP}][
     $CellContext`\[Eta][
      xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
   Editable->False]}]], "Output"]
}, Open  ]],

Cell["\<\
We expand this into components in order to obtain the set of commutation \
relations. \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Delete", "[", 
  RowBox[{
   RowBox[{"Union", "@", 
    RowBox[{"Flatten", "@", 
     RowBox[{"ComponentArray", "@", "%"}]}]}], ",", "1"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Thorn]", 
         RowBox[{"\[Thorn]", "'"}], GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Thorn]", 
         RowBox[{"\[Eth]", "'"}], GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         RowBox[{"\[Thorn]", "'"}], 
         RowBox[{"\[Eth]", "'"}], GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Eth]", 
         RowBox[{"\[Eth]", "'"}], GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
       Editable->False], ")"}]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "[", 
  RowBox[{"%", ",", 
   RowBox[{"SameTest", "\[Rule]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "#1"}], "===", "#2"}], "&"}], ")"}]}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Union", "[", 
  RowBox[{"%", ",", 
   RowBox[{"SameTest", "\[Rule]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"Dagger", "@", "#1"}], "===", "#2"}], "&"}], ")"}]}]}], 
  "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       RowBox[{"\[Thorn]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       RowBox[{"\[Eth]", "'"}], GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
         {"\[Eta]", 
          StyleBox[GridBox[{
             {
              RowBox[{"{", 
               RowBox[{"w", ",", " ", "z"}], "}"}]},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->Rational[39, 4]]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Equal", "[", 
    RowBox[{"#", ",", 
     RowBox[{"#", "//.", "GHPToNP"}]}], "]"}], "&"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", "\[Gamma]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", 
             OverscriptBox["\[Gamma]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"D", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{1, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalDelta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", 
             OverscriptBox["\[Epsilon]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[CapitalDelta]", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{2, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "D", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", 
             OverscriptBox["\[Alpha]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", "\[Beta]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad[]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", 
           InterpretationBox[
            StyleBox["w",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`w,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"D", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{1, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Delta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", "\[Epsilon]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", 
             OverscriptBox["\[Epsilon]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[Delta]", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{3, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "D", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", 
             OverscriptBox["\[Alpha]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", "\[Beta]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad[]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["w",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`w,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[CapitalDelta]", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{2, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Delta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", "\[Gamma]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", 
             OverscriptBox["\[Gamma]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[Delta]", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{3, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[CapitalDelta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", "\[Alpha]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", 
             OverscriptBox["\[Beta]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"D", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{1, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], 
             OverscriptBox["\[Epsilon]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           OverscriptBox["\[Delta]", "_"], GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{4, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "D", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", 
             OverscriptBox["\[Beta]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[CapitalDelta]", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{2, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], 
             OverscriptBox["\[Gamma]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"-", "1"}], "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"1", "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           OverscriptBox["\[Delta]", "_"], GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{4, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[CapitalDelta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", "\[Alpha]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", 
             OverscriptBox["\[Beta]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["w",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["z",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`z,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["w",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[Delta]", GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{3, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Delta]", "_"], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], 
             OverscriptBox["\[Alpha]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad[]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox["z",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`z,
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          RowBox[{"(", 
           RowBox[{"1", "+", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False]}], ")"}]}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", "1"}], "+", 
           InterpretationBox[
            StyleBox["w",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`w,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"-", 
             InterpretationBox[
              StyleBox["w",
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`w,
              Editable->False]}], " ", 
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           OverscriptBox["\[Delta]", "_"], GridBox[{
             {"\[Eta]", 
              StyleBox[GridBox[{
                 {
                  RowBox[{"{", 
                   RowBox[{"w", ",", " ", "z"}], "}"}]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->Rational[39, 4]]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PDNP[{4, -$CellContext`NP}][
          $CellContext`\[Eta][
           xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Delta]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"NPCommutators", "/.", 
      RowBox[{
       RowBox[{"\[Psi]", "[", "]"}], "->", 
       RowBox[{"\[Eta]", "[", 
        RowBox[{"LI", "@", 
         RowBox[{"{", 
          RowBox[{"w", ",", "z"}], "}"}]}], "]"}]}]}], "/.", 
     RowBox[{"Equal", "->", "Rule"}]}], ")"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"Map", "[", 
       RowBox[{"Dagger", ",", "NPCommutators", ",", 
        RowBox[{"{", "2", "}"}]}], "]"}], "/.", 
      RowBox[{
       RowBox[{"\[Psi]", "[", "]"}], "->", 
       RowBox[{"\[Eta]", "[", 
        RowBox[{"LI", "@", 
         RowBox[{"{", 
          RowBox[{"w", ",", "z"}], "}"}]}], "]"}]}]}], "/.", 
     RowBox[{"Equal", "->", "Rule"}]}], ")"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"FullSimplify", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", "\[Gamma]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad[]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", 
             OverscriptBox["\[Gamma]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", 
             OverscriptBox["\[Epsilon]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["w",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`w,
            Editable->False], "-", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              InterpretationBox[
               StyleBox["w",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`w,
               Editable->False], "+", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z,
               Editable->False]}], ")"}], " ", 
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", 
             OverscriptBox["\[Alpha]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             RowBox[{"D", "\[Beta]"}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PDNP[{1, -$CellContext`NP}][
             $CellContext`\[Beta]Dyad[]],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             RowBox[{"\[Delta]", "\[Epsilon]"}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PDNP[{3, -$CellContext`NP}][
             $CellContext`\[Epsilon]Dyad[]],
            Editable->False]}], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", 
             OverscriptBox["\[Epsilon]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", 
             OverscriptBox["\[Alpha]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", "\[Beta]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad[]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", "\[Gamma]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad[]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[Delta]", 
             OverscriptBox["\[Gamma]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{3, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Epsilon]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", "\[Alpha]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad[]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"D", 
             OverscriptBox["\[Beta]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{1, -$CellContext`NP}][
            $CellContext`\[Beta]Dyad\[Dagger][]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad[]],
           Editable->False], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], 
             OverscriptBox["\[Epsilon]", "_"]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Epsilon]Dyad\[Dagger][]],
           Editable->False], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["w",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`w,
            Editable->False], "-", 
           InterpretationBox[
            StyleBox["z",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`z,
            Editable->False]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              InterpretationBox[
               StyleBox["w",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`w,
               Editable->False], "+", 
              InterpretationBox[
               StyleBox["z",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`z,
               Editable->False]}], ")"}], " ", 
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False]}], "+", 
           RowBox[{"2", " ", 
            InterpretationBox[
             StyleBox["z",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`z,
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False]}]}], ")"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{2, -$CellContext`NP}][
            $CellContext`\[Alpha]Dyad[]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`PDNP[{4, -$CellContext`NP}][
            $CellContext`\[Gamma]Dyad[]],
           Editable->False], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             InterpretationBox[
              StyleBox[
               RowBox[{"\[CapitalDelta]", 
                OverscriptBox["\[Beta]", "_"]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDNP[{2, -$CellContext`NP}][
               $CellContext`\[Beta]Dyad\[Dagger][]],
              Editable->False], ")"}]}], "+", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              OverscriptBox["\[Delta]", "_"], 
              OverscriptBox["\[Gamma]", "_"]}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PDNP[{4, -$CellContext`NP}][
             $CellContext`\[Gamma]Dyad\[Dagger][]],
            Editable->False]}], ")"}]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            RowBox[{"(", 
             RowBox[{
              InterpretationBox[
               StyleBox[
                OverscriptBox["\[Alpha]", "_"],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Alpha]Dyad\[Dagger][],
               Editable->False], "-", 
              RowBox[{"2", " ", 
               InterpretationBox[
                StyleBox["\[Beta]",
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Beta]Dyad[],
                Editable->False]}]}], ")"}]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               RowBox[{"-", "2"}], " ", 
               InterpretationBox[
                StyleBox[
                 OverscriptBox["\[Alpha]", "_"],
                 ShowAutoStyles->False,
                 AutoSpacing->False],
                $CellContext`\[Alpha]Dyad\[Dagger][],
                Editable->False]}], "+", 
              InterpretationBox[
               StyleBox["\[Beta]",
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`\[Beta]Dyad[],
               Editable->False]}], ")"}], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             InterpretationBox[
              StyleBox[
               RowBox[{"\[Delta]", 
                OverscriptBox["\[Beta]", "_"]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDNP[{3, -$CellContext`NP}][
               $CellContext`\[Beta]Dyad\[Dagger][]],
              Editable->False], ")"}]}], "+", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              OverscriptBox["\[Delta]", "_"], 
              OverscriptBox["\[Alpha]", "_"]}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PDNP[{4, -$CellContext`NP}][
             $CellContext`\[Alpha]Dyad\[Dagger][]],
            Editable->False]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"(", 
             InterpretationBox[
              StyleBox[
               RowBox[{"\[Delta]", "\[Alpha]"}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`PDNP[{3, -$CellContext`NP}][
               $CellContext`\[Alpha]Dyad[]],
              Editable->False], ")"}]}], "+", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PDNP[{4, -$CellContext`NP}][
             $CellContext`\[Beta]Dyad[]],
            Editable->False]}], ")"}]}]}], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We use the Newman-Penrose equations on this relation.", "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"(", 
    RowBox[{"NPRicci", "/.", 
     RowBox[{"Equal", "\[Rule]", "Rule"}]}], ")"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", 
   RowBox[{"(", 
    RowBox[{"NPRicci\[Dagger]", "/.", 
     RowBox[{"Equal", "\[Rule]", "Rule"}]}], ")"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"Simplify", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad[],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad\[Dagger][],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "01"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe01Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "1"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe1Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox[
              SubscriptBox["\[CapitalPhi]", "12"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PhiCDe12Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox[
              SubscriptBox[
               OverscriptBox["\[CapitalPsi]", "_"], "3"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PsiCDe3Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "10"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe10Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "1"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe1Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox[
              SubscriptBox["\[CapitalPsi]", "3"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PsiCDe3Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Beta]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Alpha]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox[
              SubscriptBox["\[CapitalPhi]", "21"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`PhiCDe21Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Alpha]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Alpha]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Beta]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Beta]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad[],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Epsilon]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Gamma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Epsilon]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Epsilon]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Gamma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Gamma]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Finally we write ",
 Cell[BoxData[
  RowBox[{"\[Delta]", 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Eta]", " ", 
      RowBox[{"{", 
       RowBox[{"w", ",", " ", "z"}], "}"}]],
     AutoSpacing->False],
    $CellContext`\[Eta][
     xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
    Editable->False]}]], "Output",
  CellChangeTimes->{3.424604957109913*^9, 3.42460525958708*^9, 
   3.424605350976192*^9, 3.42460603308456*^9, 3.424668382421567*^9, 
   3.42467429916698*^9, 3.424674381353918*^9, 3.4246746656331987`*^9, 
   3.424675129897169*^9, 3.42477629517249*^9, 3.42478106930527*^9, 
   3.424791953119275*^9, 3.4249638056971684`*^9, 3.4250175334461*^9, 
   3.425382944168974*^9, 3.425465914406188*^9, 3.4254686456573153`*^9, 
   3.425473771978496*^9, 3.42549705944762*^9, 3.425550916009624*^9, 
   3.425579449398937*^9, 3.4256217837118587`*^9, 3.4257077739712048`*^9, 
   3.4257903074174833`*^9, 3.425879057099744*^9, 3.4268018837933817`*^9, 
   3.427103566327231*^9, 3.427260595688903*^9, 3.427263565737082*^9}],
 ", ",
 Cell[BoxData[
  RowBox[{
   OverscriptBox["\[Delta]", "_"], 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Eta]", " ", 
      RowBox[{"{", 
       RowBox[{"w", ",", " ", "z"}], "}"}]],
     AutoSpacing->False],
    $CellContext`\[Eta][
     xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
    Editable->False]}]], "Output",
  CellChangeTimes->{3.424604957109913*^9, 3.42460525958708*^9, 
   3.424605350976192*^9, 3.42460603308456*^9, 3.424668382421567*^9, 
   3.42467429916698*^9, 3.424674381353918*^9, 3.4246746656331987`*^9, 
   3.424675129897169*^9, 3.42477629517249*^9, 3.42478106930527*^9, 
   3.424791953119275*^9, 3.4249638056971684`*^9, 3.4250175334461*^9, 
   3.425382944168974*^9, 3.425465914406188*^9, 3.4254686456573153`*^9, 
   3.425473771978496*^9, 3.42549705944762*^9, 3.425550916009624*^9, 
   3.425579449398937*^9, 3.4256217837118587`*^9, 3.4257077739712048`*^9, 
   3.4257903074174833`*^9, 3.425879057099744*^9, 3.4268018837933817`*^9, 
   3.427103566327231*^9, 3.427260595688903*^9, 3.427263565737082*^9}],
 " in terms of ",
 Cell[BoxData[
  RowBox[{"\[Eth]", 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Eta]", " ", 
      RowBox[{"{", 
       RowBox[{"w", ",", " ", "z"}], "}"}]],
     AutoSpacing->False],
    $CellContext`\[Eta][
     xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
    Editable->False]}]], "Output",
  CellChangeTimes->{3.424604460667142*^9, 3.4246046812713137`*^9, 
   3.424604970772331*^9, 3.4246052633186083`*^9, 3.424605369640905*^9, 
   3.424606040582877*^9, 3.424668382466563*^9, 3.42467430051908*^9, 
   3.424674382757556*^9, 3.424674667753281*^9, 3.42467513841271*^9, 
   3.424776295207954*^9, 3.424781070598638*^9, 3.424791953192046*^9, 
   3.424963805788774*^9, 3.425017533471662*^9, 3.425382944193553*^9, 
   3.4254659180237303`*^9, 3.4254686456977577`*^9, 3.425473772019771*^9, 
   3.425497059497922*^9, 3.425550916045393*^9, 3.425579449416657*^9, 
   3.425621783745941*^9, 3.4257077740055313`*^9, 3.425790307451231*^9, 
   3.425879057133792*^9, 3.426801884729823*^9, 3.427103567465329*^9, 
   3.427260595723021*^9, 3.427263565757406*^9}],
 ",  ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"\[Eth]", "'"}], 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[Eta]", " ", 
      RowBox[{"{", 
       RowBox[{"w", ",", " ", "z"}], "}"}]],
     AutoSpacing->False],
    $CellContext`\[Eta][
     xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
    Editable->False]}]], "Output",
  CellChangeTimes->{3.424604460667142*^9, 3.4246046812713137`*^9, 
   3.424604970772331*^9, 3.4246052633186083`*^9, 3.424605369640905*^9, 
   3.424606040582877*^9, 3.424668382466563*^9, 3.42467430051908*^9, 
   3.424674382757556*^9, 3.424674667753281*^9, 3.42467513841271*^9, 
   3.424776295207954*^9, 3.424781070598638*^9, 3.424791953192046*^9, 
   3.424963805788774*^9, 3.425017533471662*^9, 3.425382944193553*^9, 
   3.4254659180237303`*^9, 3.4254686456977577`*^9, 3.425473772019771*^9, 
   3.425497059497922*^9, 3.425550916045393*^9, 3.425579449416657*^9, 
   3.425621783745941*^9, 3.4257077740055313`*^9, 3.425790307451231*^9, 
   3.425879057133792*^9, 3.426801884729823*^9, 3.427103567465329*^9, 
   3.427260595723021*^9, 3.427263565757406*^9}],
 " and simplify the resulting expression. "
}], "MathCaption"],

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "/.", "NPToGHP"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"FullSimplify", ",", "%", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`LambdaCDe[],
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`LambdaCDe[],
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "01"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe01Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "1"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe1Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "12"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe12Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "3"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe3Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "10"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe10Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "1"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe1Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "3"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe3Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "21"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe21Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad[],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad\[Dagger][],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"GHPCommutators", "=", "%"}], ";"}]], "Input"],

Cell["We check that everything is OK", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Map", "[", 
  RowBox[{"GHPWeightOf", ",", "GHPCommutators", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "True", ",", "True", ",", "True", ",", "True", ",", "True", ",", "True"}], 
  "}"}]], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Ricci identities in the GHP formalism", "Subsubsection"],

Cell["\<\
In this subsection we use the Newman-Penrose Ricci identities (Newman-Penrose \
equations) to find a set of equations which only contain the GHP differential \
operators and spin weighted quantities. This set of equations is known as GHP \
equations or compacted spin coefficient equations.\
\>", "Text",
 TextJustification->1.,
 FontWeight->"Plain"],

Cell["These are the weighted spin coefficients.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData["GHPSpinCoefficients"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[Kappa]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Kappa]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Lambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Lambda]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Mu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Mu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Nu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Nu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Pi]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Pi]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Rho]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Rho]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Sigma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Sigma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Tau]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Tau]Dyad[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell["These are the NP differentials of these spin coefficients.", \
"MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Flatten", "@", 
  RowBox[{"Outer", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"PDNP", "[", "#1", "]"}], "[", "#2", "]"}], "&"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", 
        RowBox[{"-", "NP"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"2", ",", 
        RowBox[{"-", "NP"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", 
        RowBox[{"-", "NP"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"4", ",", 
        RowBox[{"-", "NP"}]}], "}"}]}], "}"}], ",", "GHPSpinCoefficients", 
    ",", "1"}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Lambda]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Mu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Nu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Pi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Rho]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Sigma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Tau]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Lambda]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Mu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Nu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Pi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Rho]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Sigma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Tau]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Lambda]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Mu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Nu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Pi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Rho]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Sigma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Tau]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Lambda]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Lambda]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Mu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Mu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Nu]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Nu]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Pi]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Pi]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Rho]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Rho]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Sigma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Sigma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Tau]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Tau]Dyad[]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We compute the GHP differentials of the weighted Newman-Penrose spin \
coefficients.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Rule", "[", 
    RowBox[{"#", ",", 
     RowBox[{"#", "/.", "NPToGHP"}]}], "]"}], "&"}], "/@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Lambda]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Lambda]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Epsilon]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Rho]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Rho]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Lambda]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Lambda]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Rho]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Rho]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Lambda]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Lambda]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Rho]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Rho]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Lambda]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Lambda]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Rho]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Rho]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
With this information we can render the Newman-Penrose equations in terms of \
the GHP differential operators.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPRicci", "/.", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Gamma]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Alpha]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad\[Dagger][],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Beta]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox["\[Alpha]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Alpha]Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "//@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "12"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe12Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`\[Epsilon]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
We are interested in these equations which only contain weighted spin \
coefficients and GHP operators. These are\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Part", "[", 
  RowBox[{"%", ",", 
   RowBox[{"Union", "[", 
    RowBox[{"First", "/@", 
     RowBox[{"Position", "[", 
      RowBox[{"%", ",", 
       RowBox[{
        RowBox[{"\[CapitalTheta]", "[", "_", "]"}], "@", "_"}]}], "]"}]}], 
    "]"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "22"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe22Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "02"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe02Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Kappa]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Pi]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We may write these equations in an alternative way.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Union", "@", 
     RowBox[{"Cases", "[", 
      RowBox[{"%", ",", 
       RowBox[{
        RowBox[{"\[CapitalTheta]", "[", "__", "]"}], "@", "_"}], ",", 
       "Infinity"}], "]"}]}]}], "]"}], "//", "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Solve", "::", "svars"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Equations may not give solutions for all \\\"solve\\\" \
variables. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Solve/svars\\\", \
ButtonNote -> \\\"Solve::svars\\\"]\\)\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "//@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "->", "Equal"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Previous set of equations is known as GHP equations or compacted spin \
coefficient equations.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "[", 
  RowBox[{"GHPRicci", "=", "%"}], "]"}]], "Input"],

Cell[BoxData["12"], "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Bianchi identities in the GHP formalism", "Subsubsection"],

Cell["\<\
Each of the components of the curvature spinors is a weighted scalar. Using \
this property it is possible to rewrite the Newman-Penrose form of the \
Bianchi identities in terms of spin weighted quantities. This is accomplished \
in this subsection.\
\>", "Text",
 FontWeight->"Plain"],

Cell["\<\
We write the differentials of the curvature components which appear in the NP \
Bianchi identities in terms of differentials written in ters of the GHP \
operators.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NPBianchi", "/.", "NPToGHP"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             OverscriptBox["\[Alpha]", "_"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha]Dyad\[Dagger][],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox["\[Beta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Beta]Dyad[],
            Editable->False]}], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "21"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe21Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Epsilon]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Epsilon]Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             OverscriptBox["\[Epsilon]", "_"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Epsilon]Dyad\[Dagger][],
            Editable->False]}], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "02"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe02Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "4"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         RowBox[{"\[Eth]", "'"}], 
         SubscriptBox["\[CapitalPhi]", "11"]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`PhiCDe11Dyad[]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"9", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "3"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe3Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Epsilon]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Epsilon]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{"\[Thorn]", 
           SubscriptBox["\[CapitalPsi]", "3"]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
          $CellContext`PsiCDe3Dyad[]],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "10"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe10Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           RowBox[{"\[Thorn]", "'"}], 
           SubscriptBox["\[CapitalPhi]", "10"]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
          $CellContext`PhiCDe10Dyad[]],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], 
          SubscriptBox["\[CapitalPsi]", "2"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`PsiCDe2Dyad[]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "//@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", 
          SubscriptBox["\[CapitalPsi]", "3"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`PsiCDe3Dyad[]],
        Editable->False], ")"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], 
          SubscriptBox["\[CapitalPhi]", "10"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`PhiCDe10Dyad[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"9", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], 
          SubscriptBox["\[CapitalPhi]", "11"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`PhiCDe11Dyad[]],
        Editable->False], ")"}]}], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], 
          SubscriptBox["\[CapitalPsi]", "2"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`PsiCDe2Dyad[]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["We may write these equations in an alternative way.", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "[", 
   RowBox[{"%", ",", 
    RowBox[{"Union", "@", 
     RowBox[{"Cases", "[", 
      RowBox[{"%", ",", 
       RowBox[{
        RowBox[{"\[CapitalTheta]", "[", "__", "]"}], "@", "_"}], ",", 
       "Infinity"}], "]"}]}]}], "]"}], "//", "Flatten"}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"Solve", "::", "svars"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"Equations may not give solutions for all \\\"solve\\\" \
variables. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/Solve/svars\\\", \
ButtonNote -> \\\"Solve::svars\\\"]\\)\"\>"}]], "Message", "MSG"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "01"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe01Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "12"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe12Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "22"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe22Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "0"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe0Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "21"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe21Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      FractionBox["1", "3"], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"3", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "4"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe4Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Kappa]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Kappa]Dyad[],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "22"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe22Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Kappa]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Kappa]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "01"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe01Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Lambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Lambda]Dyad[],
          Editable->False]}], "+", 
        RowBox[{"6", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "1"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe1Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Lambda]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Lambda]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "10"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe10Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "10"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe10Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Mu]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "00"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe00Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Nu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Nu]Dyad[],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "11"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe11Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"9", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "2"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe2Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "20"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe20Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "21"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe21Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"6", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "3"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe3Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "21"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe21Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Rho]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad\[Dagger][],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "12"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe12Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Sigma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Sigma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "20"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe20Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}], "+", 
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "11"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe11Dyad[],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Tau]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "-", 
     FractionBox[
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Thorn]", 
         SubscriptBox["\[CapitalPhi]", "21"]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`PhiCDe21Dyad[]],
       Editable->False], "3"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           RowBox[{"\[Thorn]", "'"}], 
           SubscriptBox["\[CapitalPhi]", "10"]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
          $CellContext`PhiCDe10Dyad[]],
         Editable->False], ")"}]}], "3"], "+", 
     FractionBox[
      InterpretationBox[
       StyleBox[
        RowBox[{"\[Eth]", 
         SubscriptBox["\[CapitalPhi]", "20"]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`PhiCDe20Dyad[]],
       Editable->False], "3"], "-", 
     FractionBox[
      RowBox[{"2", " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           RowBox[{"\[Eth]", "'"}], 
           SubscriptBox["\[CapitalPhi]", "11"]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
          $CellContext`PhiCDe11Dyad[]],
         Editable->False], ")"}]}], "3"]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FullSimplify", "//@", "%"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "01"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe01Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "1"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe1Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "3"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe3Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "01"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe01Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Tau]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Mu]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "12"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe12Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "22"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe22Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "0"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe0Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "21"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe21Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     FractionBox["1", "3"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"3", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "4"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe4Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Kappa]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Kappa]Dyad[],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "22"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe22Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Kappa]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Kappa]Dyad\[Dagger][],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "01"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe01Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Lambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Lambda]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"6", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "1"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe1Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Lambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Lambda]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "10"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe10Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "10"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe10Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "00"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe00Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Nu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Nu]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"9", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "2"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe2Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "20"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe20Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"6", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "3"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe3Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "12"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe12Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Sigma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Sigma]Dyad\[Dagger][],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "20"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe20Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", 
          SubscriptBox["\[CapitalPhi]", "21"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`PhiCDe21Dyad[]],
        Editable->False], "+", 
       RowBox[{"3", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{"\[Thorn]", 
            SubscriptBox["\[CapitalPsi]", "3"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
           $CellContext`PsiCDe3Dyad[]],
          Editable->False], ")"}]}], "+", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            RowBox[{"\[Thorn]", "'"}], 
            SubscriptBox["\[CapitalPhi]", "10"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
           $CellContext`PhiCDe10Dyad[]],
          Editable->False], ")"}]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", 
          SubscriptBox["\[CapitalPhi]", "20"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`PhiCDe20Dyad[]],
        Editable->False], "-", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            RowBox[{"\[Eth]", "'"}], 
            SubscriptBox["\[CapitalPhi]", "11"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
           $CellContext`PhiCDe11Dyad[]],
          Editable->False], ")"}]}]}], ")"}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Rule]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Rule", "->", "Equal"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "01"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe01Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "1"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe1Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "3"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe3Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "01"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe01Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Tau]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Mu]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "12"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe12Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "22"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe22Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "0"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe0Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "21"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe21Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     FractionBox["1", "3"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"3", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "4"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe4Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Kappa]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Kappa]Dyad[],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "22"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe22Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Kappa]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Kappa]Dyad\[Dagger][],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "01"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe01Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Lambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Lambda]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"6", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "1"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe1Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Lambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Lambda]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "10"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe10Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "10"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe10Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "00"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe00Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Nu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Nu]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"9", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "2"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe2Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "20"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe20Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"6", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "3"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe3Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "12"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe12Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Sigma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Sigma]Dyad\[Dagger][],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "20"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe20Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", 
          SubscriptBox["\[CapitalPhi]", "21"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`PhiCDe21Dyad[]],
        Editable->False], "+", 
       RowBox[{"3", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{"\[Thorn]", 
            SubscriptBox["\[CapitalPsi]", "3"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
           $CellContext`PsiCDe3Dyad[]],
          Editable->False], ")"}]}], "+", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            RowBox[{"\[Thorn]", "'"}], 
            SubscriptBox["\[CapitalPhi]", "10"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
           $CellContext`PhiCDe10Dyad[]],
          Editable->False], ")"}]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", 
          SubscriptBox["\[CapitalPhi]", "20"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`PhiCDe20Dyad[]],
        Editable->False], "-", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            RowBox[{"\[Eth]", "'"}], 
            SubscriptBox["\[CapitalPhi]", "11"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
           $CellContext`PhiCDe11Dyad[]],
          Editable->False], ")"}]}]}], ")"}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Previous set of equations is known as GHP Bianchi identities or compacted \
Bianchi Identities.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "[", 
  RowBox[{"GHPBianchi", "=", "%"}], "]"}]], "Input"],

Cell[BoxData["11"], "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["4. Quick start", "Subsection"],

Cell[TextData[{
 "In this section we show how to use in practical applications the NP and GHP \
equations computed in previous sections. This is intended to be a short \
tutorial which enables the user to start working as quick as possible with \
the equations of the NP and GHP formalisms, so one does not need to read the \
contents of previous sections to use it. However, to take full advantage of \
the flexibility offered by our implementation and the remaining features of \
",
 StyleBox["xAct`", "Output"],
 ",  the complete knowledge of sections 1,2 and 3 is needed "
}], "Text",
 TextJustification->1.],

Cell["\<\
The basic variables of the NP and GHP formalisms are the spin coefficients \
and the components of the curvature tensor. These are stored in lists.\
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData["NPSpinCoefficients"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Alpha]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Beta]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Beta]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Gamma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Gamma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Epsilon]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Epsilon]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Kappa]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Kappa]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Lambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Lambda]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Mu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Mu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Nu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Nu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Pi]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Pi]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Rho]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Rho]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Sigma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Sigma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Tau]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Tau]Dyad[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["NPCVComponents"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "0"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe0Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "1"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe1Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "2"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe2Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "3"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe3Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPsi]", "4"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PsiCDe4Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "00"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe00Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "01"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe01Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "02"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe02Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "10"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe10Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "11"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe11Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "12"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe12Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "20"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe20Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "21"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe21Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     SubscriptBox["\[CapitalPhi]", "22"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PhiCDe22Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[CapitalLambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`LambdaCDe[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "To type any of these quantities you need to know the imput form of them. \
This is achieved with the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " command ",
 StyleBox["InputForm", "Output"],
 ". For example, let us display the input form of the spin coefficient ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Gamma]", "", ""],
    AutoSpacing->False],
   $CellContext`\[Gamma]Dyad[],
   Editable->False]], "Output",
  CellChangeTimes->{3.458874370112237*^9}],
 ":"
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"NPSpinCoefficients", "[", 
   RowBox[{"[", "3", "]"}], "]"}], "//", "InputForm"}]], "Input"],

Cell["\[Gamma]Dyad[]", "Output"]
}, Open  ]],

Cell["\<\
Recall that Greek letters are typed by using the key \"esc\" followed by the \
Latin name of the letter followed by \"esc\". Hence \
\>", "MathCaption"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Gamma]Dyad", "[", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  StyleBox["\[Gamma]",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`\[Gamma]Dyad[],
  Editable->False]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Other essential ingredient of the NP and GHP formalisms are the ",
 StyleBox["frame differentiations.",
  FontSlant->"Italic"],
 " In the NP formalism these are just the directional derivatives associated \
to the four elements of the null tetrad. To see how they are typed, we \
consider their action on the spin coefficient  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Gamma]", "", ""],
    AutoSpacing->False],
   $CellContext`\[Gamma]Dyad[],
   Editable->False]], "Output",
  CellChangeTimes->{3.458874784747464*^9}],
 "."
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Gamma]Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Gamma]Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Gamma]Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"PDNP", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Gamma]Dyad", "[", "]"}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"D", "\[Gamma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{1, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[CapitalDelta]", "\[Gamma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{2, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Delta]", "\[Gamma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{3, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`PDNP[{4, -$CellContext`NP}][
     $CellContext`\[Gamma]Dyad[]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "The head ",
 StyleBox["PDNP", "Output"],
 " denotes a generic NP operator associated to the null tetrad, whereas the \
arguments ",
 StyleBox["{1,-NP}", "Output"],
 ", ... ",
 StyleBox["{4,-NP}", "Output"],
 " are used to refer to each of the directional derivatives of the null \
tetrad elements. In this way ",
 StyleBox["{1,-NP}", "Output"],
 " means the first element of the null tetrad, ",
 StyleBox["{2,-NP}", "Output"],
 " the second element and so on. The elements of the null tetrad are numbered \
according to the usual convention. In the case of the GHP formalism, we have \
to take into account that only weighted quantities appear in the formalism. \
All the components of the Riemann tensor are weighted, but not all the spin \
coefficients are so. The weighted spin coefficients are  stored in the \
following list."
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["GHPSpinCoefficients"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[Kappa]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Kappa]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Lambda]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Lambda]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Mu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Mu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Nu]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Nu]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Pi]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Pi]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Rho]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Rho]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Sigma]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Sigma]Dyad[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[Tau]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[Tau]Dyad[],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Let us take ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox["\[Kappa]", "", ""],
    AutoSpacing->False],
   $CellContext`\[Kappa]Dyad[],
   Editable->False]], "Output",
  CellChangeTimes->{3.458875942689981*^9}],
 ": the action of the four GHP operators on it is "
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Kappa]Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Kappa]Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Kappa]Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"\[Kappa]Dyad", "[", "]"}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Thorn]", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Eth]", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "In this case, the head ",
 StyleBox["\[CapitalTheta]", "Output"],
 " (typed esc-Q-esc) is the notation for a generic GHP differential operator \
while \n",
 StyleBox["{1,-NP}", "Output"],
 ",..,",
 StyleBox["{4,-NP}", "Output"],
 " are, as before, the labels for each of the elements of the null tetrad. We \
can transform the GHP differentials into NP differentials as follows."
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", "GHPToNP"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Epsilon]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad\[Dagger][],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"D", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Gamma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad\[Dagger][],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"3", " ", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Beta]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad\[Dagger][],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
NP differentials can be also transformed back into GHP differentials.\
\>", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", "NPToGHP"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"3", " ", 
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Epsilon]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Epsilon]Dyad\[Dagger][],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Epsilon]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad\[Dagger][],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Thorn]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"3", " ", 
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Gamma]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Gamma]Dyad\[Dagger][],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Gamma]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad\[Dagger][],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Alpha]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad\[Dagger][],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"3", " ", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{"3", " ", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"3", " ", 
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", 
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Beta]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Beta]Dyad\[Dagger][],
        Editable->False]}], ")"}], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       OverscriptBox["\[Beta]", "_"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad\[Dagger][],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Kappa]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Kappa]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplify"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Thorn]", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Eth]", "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
     $CellContext`\[Kappa]Dyad[]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Previous operations can be performed on any weighted quantity and the \
appropriate weights will be automatically inserted. For example:\
\>", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"PsiCDe4Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"2", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"PsiCDe4Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"3", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"PsiCDe4Dyad", "[", "]"}]}], ",", 
   RowBox[{
    RowBox[{"\[CapitalTheta]", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", 
       RowBox[{"-", "NP"}]}], "}"}], "]"}], "@", 
    RowBox[{"PsiCDe4Dyad", "[", "]"}]}]}], "}"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Thorn]", 
      SubscriptBox["\[CapitalPsi]", "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Thorn]", "'"}], 
      SubscriptBox["\[CapitalPsi]", "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Eth]", 
      SubscriptBox["\[CapitalPsi]", "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad[]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Eth]", "'"}], 
      SubscriptBox["\[CapitalPsi]", "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
     $CellContext`PsiCDe4Dyad[]],
    Editable->False]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//.", "GHPToNP"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"4", " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Epsilon]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Epsilon]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPsi]", "4"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"4", " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Gamma]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Gamma]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPsi]", "4"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"4", " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Beta]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Beta]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       SubscriptBox["\[CapitalPsi]", "4"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{"4", " ", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "4"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe4Dyad[],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox["\[Alpha]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[Alpha]Dyad[],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       SubscriptBox["\[CapitalPsi]", "4"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]],
     Editable->False]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
The basic equations of each formalism are stored in lists. These are the \
names of the lists and the correspondences:\
\>", "MathCaption",
 TextJustification->1.],

Cell["Newman-Penrose commutation relations:", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["NPCommutators"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalDelta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{2, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "D", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Epsilon]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[CapitalDelta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         OverscriptBox["\[Nu]", "_"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Nu]Dyad\[Dagger][],
        Editable->False]}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Alpha]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Delta]", "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PDNP[{3, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Delta]", "\[Psi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Psi][]]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Mu]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`\[Psi][]]],
      Editable->False], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], "\[Psi]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`\[Psi][]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["Newman-Penrose Ricci identities:", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["NPRicci"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Gamma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Gamma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Beta]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Beta]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Beta]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Beta]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox["\[CapitalLambda]",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`LambdaCDe[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "11"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe11Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Gamma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Gamma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Epsilon]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Epsilon]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "10"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe10Dyad[],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Alpha]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Gamma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["Newman-Penrose Bianchi identities:", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["NPBianchi"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "01"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe01Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"4", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             OverscriptBox["\[Alpha]", "_"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Alpha]Dyad\[Dagger][],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox["\[Beta]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Beta]Dyad[],
            Editable->False]}], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Alpha]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Alpha]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Alpha]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPhi]", "22"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PhiCDe22Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "20"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe20Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[Epsilon]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Epsilon]Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             OverscriptBox["\[Epsilon]", "_"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`\[Epsilon]Dyad\[Dagger][],
            Editable->False]}], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Beta]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"D", 
       SubscriptBox["\[CapitalPsi]", "4"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{1, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "4"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Gamma]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPhi]", "00"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PhiCDe00Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Gamma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Beta]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Beta]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Beta]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[CapitalDelta]", 
       SubscriptBox["\[CapitalPhi]", "21"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{2, -$CellContext`NP}][
      $CellContext`PhiCDe21Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Alpha]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Alpha]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Beta]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Beta]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Gamma]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Gamma]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "4"}], " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        OverscriptBox["\[Delta]", "_"], 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{4, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Delta]", 
       SubscriptBox["\[CapitalPhi]", "11"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{3, -$CellContext`NP}][
      $CellContext`PhiCDe11Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Epsilon]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       OverscriptBox["\[Delta]", "_"], 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PDNP[{4, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Gamma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Gamma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Alpha]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad[],
          Editable->False]}], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Beta]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[CapitalDelta]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Delta]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    RowBox[{"2", " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         OverscriptBox["\[Delta]", "_"], 
         SubscriptBox["\[CapitalPhi]", "11"]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PDNP[{4, -$CellContext`NP}][
        $CellContext`PhiCDe11Dyad[]],
       Editable->False], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Epsilon]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Epsilon]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Gamma]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Gamma]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"9", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"6", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Epsilon]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Epsilon]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Alpha]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Alpha]Dyad\[Dagger][],
          Editable->False]}], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Beta]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Beta]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"D", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{1, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"D", 
          SubscriptBox["\[CapitalPsi]", "3"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{1, -$CellContext`NP}][
         $CellContext`PsiCDe3Dyad[]],
        Editable->False], ")"}]}], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[CapitalDelta]", 
          SubscriptBox["\[CapitalPhi]", "10"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{2, -$CellContext`NP}][
         $CellContext`PhiCDe10Dyad[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Delta]", 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PDNP[{3, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          OverscriptBox["\[Delta]", "_"], 
          SubscriptBox["\[CapitalPsi]", "2"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PDNP[{4, -$CellContext`NP}][
         $CellContext`PsiCDe2Dyad[]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "GHP commutators acting on a generic quantity ",
 Cell[BoxData[
  SubsuperscriptBox["\[Eta]", " ", 
   RowBox[{"{", 
    RowBox[{"w", ",", " ", "z"}], "}"}]]], "Output",
  CellChangeTimes->{3.458885750264591*^9}],
 "of boost weight w and spin weight ",
 "z."
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["GHPCommutators"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`LambdaCDe[],
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[CapitalLambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`LambdaCDe[],
             Editable->False]}], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "01"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe01Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "1"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe1Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "12"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe12Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "3"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe3Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "10"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe10Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Kappa]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Pi]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "1"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe1Dyad\[Dagger][],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Kappa]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Kappa]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Pi]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Pi]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Eta]", 
            StyleBox[GridBox[{
               {
                RowBox[{"{", 
                 RowBox[{"w", ",", " ", "z"}], "}"}]},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->Rational[39, 4]]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
        Editable->False]}], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "3"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe3Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Nu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Tau]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad[],
             Editable->False]}]}], ")"}]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "21"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe21Dyad[],
            Editable->False], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Nu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Nu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Tau]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Tau]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Eth]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], "\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
        $CellContext`\[Eta][
         xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
      Editable->False]}], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox["w",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPsi]", "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad[],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Mu]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Rho]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad[],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox["\[Lambda]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad[],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox["\[Sigma]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad[],
             Editable->False]}]}], ")"}]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox["z",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`z,
          Editable->False], " ", 
         RowBox[{"(", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[CapitalLambda]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`LambdaCDe[],
            Editable->False], "+", 
           InterpretationBox[
            StyleBox[
             SubscriptBox["\[CapitalPhi]", "11"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PhiCDe11Dyad[],
            Editable->False], "-", 
           InterpretationBox[
            StyleBox[
             SubscriptBox[
              OverscriptBox["\[CapitalPsi]", "_"], "2"],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`PsiCDe2Dyad\[Dagger][],
            Editable->False], "-", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Mu]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Mu]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Rho]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Rho]Dyad\[Dagger][],
             Editable->False]}], "+", 
           RowBox[{
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Lambda]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Lambda]Dyad\[Dagger][],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[
              OverscriptBox["\[Sigma]", "_"],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`\[Sigma]Dyad\[Dagger][],
             Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "-", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], GridBox[{
            {"\[Eta]", 
             StyleBox[GridBox[{
                {
                 RowBox[{"{", 
                  RowBox[{"w", ",", " ", "z"}], "}"}]},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->Rational[39, 4]]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`\[Eta][
          xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
        Editable->False], ")"}]}]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["GHP Ricci identities:", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["GHPRicci"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Tau]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Tau]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Thorn]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Nu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Nu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "0"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe0Dyad[],
      Editable->False], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Mu]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Mu]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], "2"], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Sigma]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Sigma]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Kappa]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Kappa]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "00"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe00Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Mu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Mu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "21"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe21Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "3"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe3Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Nu]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Nu]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Pi]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Pi]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "20"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe20Dyad[],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False], "2"], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[Lambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Lambda]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Sigma]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Sigma]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPhi]", "01"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PhiCDe01Dyad[],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "1"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe1Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], "\[Tau]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Tau]Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox["\[CapitalLambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`LambdaCDe[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       SubscriptBox["\[CapitalPsi]", "2"],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`PsiCDe2Dyad[],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], "\[Rho]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Rho]Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["GHP Bianchi identities:", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData["GHPBianchi"], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "02"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe02Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "0"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe0Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "01"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe01Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "1"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe1Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPsi]", "3"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe3Dyad[],
          Editable->False]}]}], ")"}], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"\[Eth]", 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"2", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "01"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe01Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Tau]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad[],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False], "-", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Tau]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Tau]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "02"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe02Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Mu]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Mu]Dyad\[Dagger][],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"2", " ", 
         InterpretationBox[
          StyleBox["\[Rho]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Rho]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "12"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe12Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "12"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe12Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Lambda]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox[
           OverscriptBox["\[Pi]", "_"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad\[Dagger][],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "2"}], " ", 
         InterpretationBox[
          StyleBox["\[Pi]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[Pi]Dyad[],
          Editable->False]}], "+", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "22"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe22Dyad[]],
      Editable->False], "+", 
     RowBox[{"3", " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          RowBox[{"\[Thorn]", "'"}], "\[CapitalLambda]"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
         $CellContext`LambdaCDe[]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPhi]", "22"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PhiCDe22Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Nu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Tau]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "3"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe3Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "0"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe0Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "2"}], " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "02"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe02Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "01"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe01Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "1"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe1Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "00"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe00Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "1"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe1Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPhi]", "21"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PhiCDe21Dyad[],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Kappa]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "12"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe12Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Kappa]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Kappa]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "0"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe0Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "00"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe00Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Mu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "1"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe1Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Pi]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Rho]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Sigma]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad[],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", "\[CapitalLambda]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`LambdaCDe[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPhi]", "11"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PhiCDe11Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "2"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe2Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPhi]", "10"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PhiCDe10Dyad[]],
      Editable->False]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "2"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe2Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     FractionBox["1", "3"], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"3", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "4"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe4Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Kappa]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Kappa]Dyad[],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "22"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe22Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Kappa]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Kappa]Dyad\[Dagger][],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "01"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe01Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Lambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Lambda]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"6", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "1"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe1Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Lambda]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Lambda]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "10"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe10Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Mu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "10"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe10Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Mu]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Mu]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "00"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe00Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Nu]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Nu]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"9", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "2"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe2Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Pi]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad[],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "20"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe20Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Pi]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Pi]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], "-", 
       RowBox[{"6", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPsi]", "3"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PsiCDe3Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Rho]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "21"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe21Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Rho]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Rho]Dyad\[Dagger][],
         Editable->False]}], "-", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "12"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe12Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Sigma]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Sigma]Dyad\[Dagger][],
         Editable->False]}], "+", 
       RowBox[{"2", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "20"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe20Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox["\[Tau]",
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad[],
         Editable->False]}], "+", 
       RowBox[{"4", " ", 
        InterpretationBox[
         StyleBox[
          SubscriptBox["\[CapitalPhi]", "11"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`PhiCDe11Dyad[],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[
          OverscriptBox["\[Tau]", "_"],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`\[Tau]Dyad\[Dagger][],
         Editable->False]}], "-", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Thorn]", 
          SubscriptBox["\[CapitalPhi]", "21"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
         $CellContext`PhiCDe21Dyad[]],
        Editable->False], "+", 
       RowBox[{"3", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{"\[Thorn]", 
            SubscriptBox["\[CapitalPsi]", "3"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
           $CellContext`PsiCDe3Dyad[]],
          Editable->False], ")"}]}], "+", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            RowBox[{"\[Thorn]", "'"}], 
            SubscriptBox["\[CapitalPhi]", "10"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
           $CellContext`PhiCDe10Dyad[]],
          Editable->False], ")"}]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{"\[Eth]", 
          SubscriptBox["\[CapitalPhi]", "20"]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
         $CellContext`PhiCDe20Dyad[]],
        Editable->False], "-", 
       RowBox[{"2", " ", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            RowBox[{"\[Eth]", "'"}], 
            SubscriptBox["\[CapitalPhi]", "11"]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
           $CellContext`PhiCDe11Dyad[]],
          Editable->False], ")"}]}]}], ")"}]}]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       RowBox[{"\[Eth]", "'"}], 
       SubscriptBox["\[CapitalPsi]", "3"]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe3Dyad[]],
     Editable->False], "\[Equal]", 
    RowBox[{
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "11"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe11Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{"3", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "2"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe2Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Lambda]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Lambda]Dyad[],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "20"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe20Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Mu]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Mu]Dyad\[Dagger][],
       Editable->False]}], "-", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Nu]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Nu]Dyad[],
       Editable->False]}], "-", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "3"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe3Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Pi]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Pi]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPsi]", "4"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PsiCDe4Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["\[Rho]",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Rho]Dyad[],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "22"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe22Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Sigma]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Sigma]Dyad\[Dagger][],
       Editable->False]}], "+", 
     RowBox[{"2", " ", 
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        OverscriptBox["\[Tau]", "_"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`\[Tau]Dyad\[Dagger][],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "20"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PhiCDe20Dyad[]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPhi]", "21"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PhiCDe21Dyad[]],
      Editable->False]}]}]}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Within the GHP formalism any expression must be weighted. The command \
GHPWeightOf will compute the weight of any expression.\
\>", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "GHPWeightOf"}]], "Input"],

Cell[BoxData[
 StyleBox["\<\"GHPWeightOf[expr] returns the GHP boost and spin weight of any \
expression. If the expression is not weighted, then an error message is \
returned\"\>", "MSG"]], "Print", "PrintUsage",
 CellTags->"Info3583346399-5671824"]
}, Open  ]],

Cell["For example: take one of the GHP Bianchi identities: ", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GHPBianchi", "[", 
  RowBox[{"[", "2", "]"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{"\[Eth]", 
     SubscriptBox["\[CapitalPsi]", "2"]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
    $CellContext`PsiCDe2Dyad[]],
   Editable->False], "\[Equal]", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "22"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe22Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[Kappa]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad[],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "10"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe10Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Lambda]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Lambda]Dyad\[Dagger][],
     Editable->False]}], "+", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "01"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe01Dyad[],
       Editable->False], "+", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPsi]", "1"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PsiCDe1Dyad[],
        Editable->False]}]}], ")"}], " ", 
    InterpretationBox[
     StyleBox["\[Mu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Mu]Dyad[],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "0"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe0Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[Nu]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Nu]Dyad[],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "02"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe02Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False]}], "-", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "11"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe11Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Pi]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad\[Dagger][],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPhi]", "12"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PhiCDe12Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Rho]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Rho]Dyad\[Dagger][],
     Editable->False]}], "-", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "21"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe21Dyad[],
       Editable->False], "+", 
      RowBox[{"2", " ", 
       InterpretationBox[
        StyleBox[
         SubscriptBox["\[CapitalPsi]", "3"],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`PsiCDe3Dyad[],
        Editable->False]}]}], ")"}], " ", 
    InterpretationBox[
     StyleBox["\[Sigma]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad[],
     Editable->False]}], "+", 
   RowBox[{"3", " ", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "2"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe2Dyad[],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["\[Tau]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Tau]Dyad[],
     Editable->False]}], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Thorn]", 
      SubscriptBox["\[CapitalPhi]", "12"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
     $CellContext`PhiCDe12Dyad[]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Thorn]", "'"}], 
      SubscriptBox["\[CapitalPsi]", "1"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
     $CellContext`PsiCDe1Dyad[]],
    Editable->False], "+", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Eth]", "\[CapitalLambda]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
     $CellContext`LambdaCDe[]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Eth]", 
      SubscriptBox["\[CapitalPhi]", "11"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
     $CellContext`PhiCDe11Dyad[]],
    Editable->False]}]}]], "Output"]
}, Open  ]],

Cell["\<\
We compute the GHP weight of the right hand side of this relation.\
\>", "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GHPWeightOf", "@", 
  RowBox[{"%", "[", 
   RowBox[{"[", "2", "]"}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1", ",", 
   RowBox[{"-", "1"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Finally we explain how to compute the action of the GHP commutators on any \
quantity. For example suppose that we wish to compute the commutation of \
\[Thorn] and \[Eth]' on the scalar ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox[
     SubscriptBox["\[CapitalPsi]", "4"], "", ""],
    AutoSpacing->False],
   $CellContext`PsiCDe4Dyad[],
   Editable->False]], "Output",
  CellChangeTimes->{3.459002960150869*^9}],
 ". First of all, we look up in the list of GHP commutators the commutation \
relation we wish to use. In our case this is"
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GHPCommutators", "[", 
  RowBox[{"[", "4", "]"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Thorn]", 
      RowBox[{"\[Eth]", "'"}], GridBox[{
        {"\[Eta]", 
         StyleBox[GridBox[{
            {
             RowBox[{"{", 
              RowBox[{"w", ",", " ", "z"}], "}"}]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Eth]", "'"}], "\[Thorn]", GridBox[{
        {"\[Eta]", 
         StyleBox[GridBox[{
            {
             RowBox[{"{", 
              RowBox[{"w", ",", " ", "z"}], "}"}]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`\[Eta][
       xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]]],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Eta]", 
         StyleBox[GridBox[{
            {
             RowBox[{"{", 
              RowBox[{"w", ",", " ", "z"}], "}"}]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->Rational[39, 4]]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Eta][
      xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox["w",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`w,
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         InterpretationBox[
          StyleBox[
           SubscriptBox["\[CapitalPhi]", "10"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PhiCDe10Dyad[],
          Editable->False], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Kappa]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Kappa]Dyad[],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["\[Lambda]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Lambda]Dyad[],
           Editable->False]}], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox["\[Pi]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Pi]Dyad[],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox["\[Rho]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Rho]Dyad[],
           Editable->False]}]}], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["z",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`z,
        Editable->False], " ", 
       RowBox[{"(", 
        RowBox[{
         InterpretationBox[
          StyleBox[
           SubscriptBox[
            OverscriptBox["\[CapitalPsi]", "_"], "1"],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`PsiCDe1Dyad\[Dagger][],
          Editable->False], "+", 
         RowBox[{
          InterpretationBox[
           StyleBox[
            OverscriptBox["\[Kappa]", "_"],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Kappa]Dyad\[Dagger][],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[
            OverscriptBox["\[Mu]", "_"],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Mu]Dyad\[Dagger][],
           Editable->False]}], "-", 
         RowBox[{
          InterpretationBox[
           StyleBox[
            OverscriptBox["\[Pi]", "_"],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Pi]Dyad\[Dagger][],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[
            OverscriptBox["\[Sigma]", "_"],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`\[Sigma]Dyad\[Dagger][],
           Editable->False]}]}], ")"}]}]}], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False], ")"}]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Kappa]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad\[Dagger][],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Sigma]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad\[Dagger][],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Rho]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Rho]Dyad[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], GridBox[{
          {"\[Eta]", 
           StyleBox[GridBox[{
              {
               RowBox[{"{", 
                RowBox[{"w", ",", " ", "z"}], "}"}]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->Rational[39, 4]]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`\[Eta][
        xAct`xTensor`LI[{$CellContext`w, $CellContext`z}]]],
      Editable->False], ")"}]}]}]}]], "Output"]
}, Open  ]],

Cell[TextData[{
 "Next we replace ",
 Cell[BoxData[
  SubsuperscriptBox["\[Eta]", " ", 
   RowBox[{"{", 
    RowBox[{"w", ",", " ", "z"}], "}"}]]], "Output",
  CellChangeTimes->{3.459002981084524*^9, 3.4590030771178837`*^9}],
 " by  ",
 Cell[BoxData[
  InterpretationBox[
   StyleBox[
    SubsuperscriptBox[
     SubscriptBox["\[CapitalPsi]", "4"], "", ""],
    AutoSpacing->False],
   $CellContext`PsiCDe4Dyad[],
   Editable->False]], "Output",
  CellChangeTimes->{3.459002960150869*^9}],
 " and the weights {w, z} by their values. This is accomplished as follows"
}], "MathCaption",
 TextJustification->1.],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"\[Eta]", "@", 
      RowBox[{"LI", "@", 
       RowBox[{"{", 
        RowBox[{"w", ",", "z"}], "}"}]}]}], "\[Rule]", 
     RowBox[{"PsiCDe4Dyad", "[", "]"}]}], ",", 
    RowBox[{"w", "\[Rule]", 
     RowBox[{"First", "@", 
      RowBox[{"GHPWeightOf", "@", 
       RowBox[{"PsiCDe4Dyad", "[", "]"}]}]}]}], ",", 
    RowBox[{"z", "\[Rule]", 
     RowBox[{"Last", "@", 
      RowBox[{"GHPWeightOf", "@", 
       RowBox[{"PsiCDe4Dyad", "[", "]"}]}]}]}]}], "}"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"\[Thorn]", 
      RowBox[{"\[Eth]", "'"}], 
      SubscriptBox["\[CapitalPsi]", "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
     $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]]],
    Editable->False], "-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      RowBox[{"\[Eth]", "'"}], "\[Thorn]", 
      SubscriptBox["\[CapitalPsi]", "4"]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
     $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
      $CellContext`PsiCDe4Dyad[]]],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "4"}], " ", 
    InterpretationBox[
     StyleBox[
      SubscriptBox["\[CapitalPsi]", "4"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`PsiCDe4Dyad[],
     Editable->False], " ", 
    RowBox[{"(", 
     RowBox[{
      InterpretationBox[
       StyleBox[
        SubscriptBox["\[CapitalPhi]", "10"],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`PhiCDe10Dyad[],
       Editable->False], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Kappa]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Kappa]Dyad[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["\[Lambda]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Lambda]Dyad[],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox["\[Pi]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Pi]Dyad[],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox["\[Rho]",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`\[Rho]Dyad[],
        Editable->False]}]}], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Pi]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Pi]Dyad[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Thorn]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{1, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], ")"}]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Kappa]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Kappa]Dyad\[Dagger][],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Thorn]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{2, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      OverscriptBox["\[Sigma]", "_"],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Sigma]Dyad\[Dagger][],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{"\[Eth]", 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{3, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], ")"}]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox["\[Rho]",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`\[Rho]Dyad[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"\[Eth]", "'"}], 
        SubscriptBox["\[CapitalPsi]", "4"]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`\[CapitalTheta][{4, -$CellContext`NP}][
       $CellContext`PsiCDe4Dyad[]],
      Editable->False], ")"}]}]}]}]], "Output"]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
CellGrouping->Manual,
WindowSize->{1117, 720},
WindowMargins->{{66, Automatic}, {Automatic, 0}},
ShowSelection->True,
FrontEndVersion->"9.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (January 25, \
2013)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info3583346399-5671824"->{
  Cell[5448863, 159706, 251, 4, 43, "Print",
   CellTags->"Info3583346399-5671824"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info3583346399-5671824", 5474517, 160531}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 100, 1, 92, "Title"],
Cell[670, 25, 300, 11, 51, "Subtitle"],
Cell[973, 38, 127, 4, 69, "Subsubtitle"],
Cell[1103, 44, 446, 8, 68, "Text"],
Cell[1552, 54, 205, 4, 30, "Text"],
Cell[CellGroupData[{
Cell[1782, 62, 66, 0, 44, "Subsection"],
Cell[1851, 64, 98, 4, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[1974, 72, 57, 1, 28, "Input"],
Cell[2034, 75, 96, 1, 70, "Print"],
Cell[2133, 78, 363, 8, 70, "Print"],
Cell[2499, 88, 117, 1, 70, "Print"],
Cell[2619, 91, 74, 0, 70, "Print"],
Cell[2696, 93, 57, 0, 70, "Print"],
Cell[2756, 95, 96, 1, 70, "Print"],
Cell[2855, 98, 367, 8, 70, "Print"],
Cell[3225, 108, 117, 1, 70, "Print"],
Cell[3345, 111, 96, 1, 70, "Print"],
Cell[3444, 114, 367, 8, 70, "Print"],
Cell[3814, 124, 155, 1, 70, "Print"],
Cell[3972, 127, 96, 1, 70, "Print"],
Cell[4071, 130, 251, 3, 70, "Print"],
Cell[4325, 135, 96, 1, 70, "Print"]
}, Open  ]],
Cell[4436, 139, 129, 5, 41, "MathCaption"],
Cell[CellGroupData[{
Cell[4590, 148, 293, 7, 28, "Input"],
Cell[4886, 157, 430, 9, 70, "Print"],
Cell[5319, 168, 440, 9, 70, "Print"]
}, Open  ]],
Cell[5774, 180, 136, 5, 41, "MathCaption"],
Cell[CellGroupData[{
Cell[5935, 189, 251, 8, 28, "Input"],
Cell[6189, 199, 588, 14, 70, "Print"],
Cell[6780, 215, 694, 17, 70, "Print"],
Cell[7477, 234, 665, 15, 70, "Print"],
Cell[8145, 251, 683, 15, 70, "Print"],
Cell[8831, 268, 516, 12, 70, "Print"],
Cell[9350, 282, 636, 15, 70, "Print"],
Cell[9989, 299, 654, 16, 70, "Print"],
Cell[10646, 317, 680, 16, 70, "Print"],
Cell[11329, 335, 598, 14, 70, "Print"],
Cell[11930, 351, 105, 1, 70, "Print"],
Cell[12038, 354, 482, 10, 70, "Print"],
Cell[12523, 366, 109, 1, 70, "Print"],
Cell[12635, 369, 614, 15, 70, "Print"],
Cell[13252, 386, 665, 15, 70, "Print"],
Cell[13920, 403, 606, 14, 70, "Print"],
Cell[14529, 419, 494, 10, 70, "Print"],
Cell[15026, 431, 240, 5, 70, "Print"],
Cell[15269, 438, 232, 5, 70, "Print"],
Cell[15504, 445, 244, 5, 70, "Print"],
Cell[15751, 452, 498, 10, 70, "Print"]
}, Open  ]],
Cell[16264, 465, 1774, 51, 130, "Text"],
Cell[CellGroupData[{
Cell[18063, 520, 428, 10, 28, "Input"],
Cell[18494, 532, 430, 9, 70, "Print"],
Cell[18927, 543, 246, 6, 70, "Message"],
Cell[19176, 551, 246, 6, 70, "Message"],
Cell[19425, 559, 604, 12, 70, "Print"],
Cell[20032, 573, 596, 13, 70, "Print"],
Cell[20631, 588, 615, 13, 70, "Print"],
Cell[21249, 603, 565, 12, 70, "Print"],
Cell[21817, 617, 621, 13, 70, "Print"],
Cell[22441, 632, 625, 13, 70, "Print"],
Cell[23069, 647, 679, 13, 70, "Print"],
Cell[23751, 662, 627, 11, 70, "Print"],
Cell[24381, 675, 622, 11, 70, "Print"],
Cell[25006, 688, 771, 14, 70, "Print"],
Cell[25780, 704, 786, 14, 70, "Print"],
Cell[26569, 720, 518, 12, 70, "Print"],
Cell[27090, 734, 628, 14, 70, "Print"],
Cell[27721, 750, 655, 16, 70, "Print"],
Cell[28379, 768, 720, 16, 70, "Print"],
Cell[29102, 786, 714, 15, 70, "Print"],
Cell[29819, 803, 800, 17, 70, "Print"],
Cell[30622, 822, 805, 17, 70, "Print"],
Cell[31430, 841, 887, 18, 70, "Print"],
Cell[32320, 861, 627, 14, 70, "Print"],
Cell[32950, 877, 745, 15, 70, "Print"],
Cell[33698, 894, 691, 15, 70, "Print"],
Cell[34392, 911, 683, 15, 70, "Print"],
Cell[35078, 928, 459, 9, 70, "Print"],
Cell[35540, 939, 627, 14, 70, "Print"],
Cell[36170, 955, 745, 15, 70, "Print"],
Cell[36918, 972, 938, 20, 70, "Print"],
Cell[37859, 994, 953, 20, 70, "Print"],
Cell[38815, 1016, 727, 15, 70, "Print"],
Cell[39545, 1033, 745, 15, 70, "Print"],
Cell[40293, 1050, 932, 20, 70, "Print"],
Cell[41228, 1072, 947, 20, 70, "Print"],
Cell[42178, 1094, 731, 15, 70, "Print"],
Cell[42912, 1111, 749, 15, 70, "Print"]
}, Open  ]],
Cell[43676, 1129, 98, 2, 40, "MathCaption"],
Cell[43777, 1133, 1133, 27, 56, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[44947, 1165, 53, 0, 36, "Subsection"],
Cell[45003, 1167, 510, 12, 69, "Text"],
Cell[45516, 1181, 83, 4, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[45624, 1189, 55, 1, 28, "Input"],
Cell[45682, 1192, 96, 1, 21, "Print"],
Cell[45781, 1195, 363, 8, 22, "Print"],
Cell[46147, 1205, 135, 1, 22, "Print"],
Cell[46285, 1208, 96, 1, 21, "Print"],
Cell[46384, 1211, 251, 3, 38, "Print"],
Cell[46638, 1216, 96, 1, 21, "Print"]
}, Open  ]],
Cell[46749, 1220, 145, 4, 42, "MathCaption"],
Cell[46897, 1226, 80, 2, 28, "Input"],
Cell[46980, 1230, 501, 14, 50, "Text"],
Cell[CellGroupData[{
Cell[47506, 1248, 34, 0, 35, "Subsubsection"],
Cell[47543, 1250, 75, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[47643, 1254, 609, 19, 65, "Input"],
Cell[48255, 1275, 462, 9, 70, "Print"],
Cell[48720, 1286, 498, 9, 70, "Print"],
Cell[49221, 1297, 474, 9, 70, "Print"],
Cell[49698, 1308, 510, 9, 70, "Print"]
}, Open  ]],
Cell[50223, 1320, 83, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[50331, 1325, 322, 8, 28, "Input"],
Cell[50656, 1335, 420, 9, 70, "Print"],
Cell[51079, 1346, 518, 11, 70, "Print"],
Cell[51600, 1359, 618, 14, 70, "Print"],
Cell[52221, 1375, 670, 16, 70, "Print"],
Cell[52894, 1393, 690, 16, 70, "Print"],
Cell[53587, 1411, 606, 14, 70, "Print"],
Cell[54196, 1427, 661, 16, 70, "Print"],
Cell[54860, 1445, 726, 16, 70, "Print"],
Cell[55589, 1463, 674, 17, 70, "Print"],
Cell[56266, 1482, 752, 17, 70, "Print"],
Cell[57021, 1501, 442, 10, 70, "Print"],
Cell[57466, 1513, 562, 12, 70, "Print"],
Cell[58031, 1527, 618, 13, 70, "Print"],
Cell[58652, 1542, 610, 14, 70, "Print"],
Cell[59265, 1558, 666, 15, 70, "Print"]
}, Open  ]],
Cell[59946, 1576, 94, 2, 40, "MathCaption"],
Cell[60043, 1580, 286, 8, 28, "Input"],
Cell[60332, 1590, 100, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[60457, 1596, 528, 18, 28, "Input"],
Cell[60988, 1616, 4311, 120, 70, "Output"]
}, Open  ]],
Cell[65314, 1739, 669, 22, 45, "MathCaption"],
Cell[CellGroupData[{
Cell[66008, 1765, 153, 4, 28, "Input"],
Cell[66164, 1771, 516, 10, 70, "Print"],
Cell[66683, 1783, 524, 10, 70, "Print"],
Cell[67210, 1795, 4693, 130, 70, "Output"]
}, Open  ]],
Cell[71918, 1928, 71, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[72014, 1932, 196, 5, 28, "Input"],
Cell[72213, 1939, 538, 11, 70, "Print"],
Cell[72754, 1952, 564, 11, 70, "Print"],
Cell[73321, 1965, 5051, 142, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[78409, 2112, 118, 3, 28, "Input"],
Cell[78530, 2117, 9933, 266, 70, "Output"]
}, Open  ]],
Cell[88478, 2386, 110, 3, 28, "Input"],
Cell[88591, 2391, 3368, 57, 48, "MathCaption"],
Cell[CellGroupData[{
Cell[91984, 2452, 387, 12, 28, "Input"],
Cell[92374, 2466, 5318, 149, 70, "Output"]
}, Open  ]],
Cell[97707, 2618, 124, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[97856, 2625, 153, 4, 28, "Input"],
Cell[98012, 2631, 5878, 162, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[103927, 2798, 196, 5, 28, "Input"],
Cell[104126, 2805, 6383, 178, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[110546, 2988, 150, 4, 28, "Input"],
Cell[110699, 2994, 11946, 309, 70, "Output"]
}, Open  ]],
Cell[122660, 3306, 110, 3, 28, "Input"],
Cell[122773, 3311, 1521, 48, 46, "MathCaption"],
Cell[CellGroupData[{
Cell[124319, 3363, 220, 6, 28, "Input"],
Cell[124542, 3371, 192, 7, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[124771, 3383, 87, 2, 28, "Input"],
Cell[124861, 3387, 192, 7, 70, "Output"]
}, Open  ]],
Cell[125068, 3397, 359, 11, 45, "MathCaption"],
Cell[125430, 3410, 322, 8, 46, "Input"],
Cell[125755, 3420, 747, 21, 44, "MathCaption"],
Cell[CellGroupData[{
Cell[126527, 3445, 630, 21, 28, "Input"],
Cell[127160, 3468, 6375, 178, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[133572, 3651, 153, 4, 28, "Input"],
Cell[133728, 3657, 7133, 196, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[140898, 3858, 196, 5, 28, "Input"],
Cell[141097, 3865, 7779, 216, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[148913, 4086, 185, 5, 28, "Input"],
Cell[149101, 4093, 25185, 672, 70, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[174323, 4770, 92, 2, 28, "Input"],
Cell[174418, 4774, 9985, 268, 70, "Output"]
}, Open  ]],
Cell[184418, 5045, 110, 3, 28, "Input"],
Cell[184531, 5050, 237, 8, 42, "MathCaption"],
Cell[184771, 5060, 233, 6, 28, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[185041, 5071, 36, 0, 29, "Subsubsection"],
Cell[185080, 5073, 112, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[185217, 5080, 171, 4, 28, "Input"],
Cell[185391, 5086, 466, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[185894, 5100, 171, 4, 28, "Input"],
Cell[186068, 5106, 466, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[186571, 5120, 222, 5, 28, "Input"],
Cell[186796, 5127, 466, 9, 22, "Print"],
Cell[187265, 5138, 484, 9, 22, "Print"]
}, Open  ]],
Cell[187764, 5150, 155, 4, 30, "Input"],
Cell[187922, 5156, 92, 2, 30, "Text"],
Cell[CellGroupData[{
Cell[188039, 5162, 289, 7, 28, "Input"],
Cell[188331, 5171, 416, 9, 22, "Print"],
Cell[188750, 5182, 514, 11, 22, "Print"],
Cell[189267, 5195, 616, 14, 22, "Print"],
Cell[189886, 5211, 666, 16, 22, "Print"],
Cell[190555, 5229, 686, 16, 22, "Print"],
Cell[191244, 5247, 602, 14, 22, "Print"],
Cell[191849, 5263, 612, 13, 22, "Print"],
Cell[192464, 5278, 704, 17, 22, "Print"]
}, Open  ]],
Cell[193183, 5298, 201, 5, 42, "MathCaption"],
Cell[193387, 5305, 456, 13, 28, "Input"],
Cell[193846, 5320, 316, 8, 49, "Text"],
Cell[194165, 5330, 46, 0, 40, "MathCaption"],
Cell[194214, 5332, 81, 2, 28, "Input"],
Cell[194298, 5336, 102, 2, 40, "MathCaption"],
Cell[194403, 5340, 981, 33, 87, "Input"],
Cell[195387, 5375, 527, 15, 57, "MathCaption"],
Cell[195917, 5392, 977, 33, 87, "Input"],
Cell[196897, 5427, 50, 0, 40, "MathCaption"],
Cell[196950, 5429, 82, 2, 28, "Input"],
Cell[197035, 5433, 168, 5, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[197228, 5442, 913, 30, 28, "Input"],
Cell[198144, 5474, 8584, 239, 37, "Output"]
}, Open  ]],
Cell[206743, 5716, 3182, 69, 44, "MathCaption"],
Cell[CellGroupData[{
Cell[209950, 5789, 149, 4, 28, "Input"],
Cell[210102, 5795, 502, 10, 22, "Print"],
Cell[210607, 5807, 9332, 259, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[219976, 6071, 91, 2, 28, "Input"],
Cell[220070, 6075, 17905, 502, 37, "Output"]
}, Open  ]],
Cell[237990, 6580, 115, 3, 28, "Input"],
Cell[238108, 6585, 93, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[238226, 6591, 552, 18, 28, "Input"],
Cell[238781, 6611, 9193, 262, 37, "Output"]
}, Open  ]],
Cell[247989, 6876, 469, 14, 45, "MathCaption"],
Cell[CellGroupData[{
Cell[248483, 6894, 85, 2, 28, "Input"],
Cell[248571, 6898, 10089, 283, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[258697, 7186, 92, 2, 28, "Input"],
Cell[258792, 7190, 20621, 540, 61, "Output"]
}, Open  ]],
Cell[279428, 7733, 67, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[279520, 7737, 91, 2, 28, "Input"],
Cell[279614, 7741, 20605, 540, 61, "Output"]
}, Open  ]],
Cell[300234, 8284, 342, 11, 44, "MathCaption"],
Cell[CellGroupData[{
Cell[300601, 8299, 217, 6, 28, "Input"],
Cell[300821, 8307, 703, 21, 92, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[301561, 8333, 87, 2, 28, "Input"],
Cell[301651, 8337, 703, 21, 92, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[302391, 8363, 119, 3, 28, "Input"],
Cell[302513, 8368, 20637, 540, 61, "Output"]
}, Open  ]],
Cell[323165, 8911, 278, 9, 42, "MathCaption"],
Cell[323446, 8922, 57, 1, 28, "Input"],
Cell[323506, 8925, 182, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[323713, 8932, 1074, 38, 28, "Input"],
Cell[324790, 8972, 12672, 356, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[337499, 9333, 149, 4, 28, "Input"],
Cell[337651, 9339, 14156, 392, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[351844, 9736, 127, 3, 28, "Input"],
Cell[351974, 9741, 43416, 1159, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[395427, 10905, 122, 3, 28, "Input"],
Cell[395552, 10910, 17961, 504, 37, "Output"]
}, Open  ]],
Cell[413528, 11417, 115, 3, 28, "Input"],
Cell[413646, 11422, 122, 3, 40, "MathCaption"],
Cell[413771, 11427, 165, 5, 28, "Input"],
Cell[413939, 11434, 3226, 69, 44, "MathCaption"],
Cell[CellGroupData[{
Cell[417190, 11507, 770, 22, 46, "Input"],
Cell[417963, 11531, 13448, 376, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[431448, 11912, 127, 3, 28, "Input"],
Cell[431578, 11917, 8584, 239, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[440199, 12161, 290, 9, 28, "Input"],
Cell[440492, 12172, 8580, 239, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[449109, 12416, 344, 11, 28, "Input"],
Cell[449456, 12429, 10074, 279, 70, "Output"]
}, Open  ]],
Cell[459545, 12711, 80, 2, 28, "Input"],
Cell[459628, 12715, 73, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[459726, 12719, 942, 30, 46, "Input"],
Cell[460671, 12751, 13428, 372, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[474136, 13128, 127, 3, 28, "Input"],
Cell[474266, 13133, 8710, 243, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[483013, 13381, 229, 7, 28, "Input"],
Cell[483245, 13390, 8703, 243, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[491985, 13638, 344, 11, 28, "Input"],
Cell[492332, 13651, 10193, 283, 70, "Output"]
}, Open  ]],
Cell[502540, 13937, 116, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[502681, 13944, 97, 2, 28, "Input"],
Cell[502781, 13948, 20816, 563, 99, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[523646, 14517, 75, 0, 29, "Subsubsection"],
Cell[523724, 14519, 135, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[523884, 14526, 1517, 45, 46, "Input"],
Cell[525404, 14573, 17452, 487, 37, "Output"]
}, Open  ]],
Cell[542871, 15063, 926, 24, 44, "MathCaption"],
Cell[CellGroupData[{
Cell[543822, 15091, 151, 4, 28, "Input"],
Cell[543976, 15097, 18208, 503, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[562221, 15605, 160, 4, 28, "Input"],
Cell[562384, 15611, 19080, 527, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[581501, 16143, 149, 4, 28, "Input"],
Cell[581653, 16149, 19788, 543, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[601478, 16697, 127, 3, 28, "Input"],
Cell[601608, 16702, 9960, 271, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[611605, 16978, 62, 1, 28, "Input"],
Cell[611670, 16981, 23421, 603, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[635128, 17589, 96, 2, 28, "Input"],
Cell[635227, 17593, 23393, 602, 61, "Output"]
}, Open  ]],
Cell[658635, 18198, 149, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[658809, 18205, 315, 8, 28, "Input"],
Cell[659127, 18215, 17665, 495, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[676829, 18715, 151, 4, 28, "Input"],
Cell[676983, 18721, 18417, 511, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[695437, 19237, 160, 4, 28, "Input"],
Cell[695600, 19243, 19257, 531, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[714894, 19779, 149, 4, 28, "Input"],
Cell[715046, 19785, 20005, 551, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[735088, 20341, 127, 3, 28, "Input"],
Cell[735218, 20346, 15169, 418, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[750424, 20769, 91, 2, 28, "Input"],
Cell[750518, 20773, 22053, 594, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[772608, 21372, 164, 4, 28, "Input"],
Cell[772775, 21378, 22829, 611, 61, "Output"]
}, Open  ]],
Cell[795619, 21992, 45, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[795689, 21996, 95, 2, 28, "Input"],
Cell[795787, 22000, 22829, 611, 61, "Output"]
}, Open  ]],
Cell[818631, 22614, 169, 6, 42, "MathCaption"],
Cell[818803, 22622, 64, 1, 28, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[818904, 22628, 57, 0, 29, "Subsubsection"],
Cell[818964, 22630, 115, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[819104, 22637, 456, 13, 63, "Input"],
Cell[819563, 22652, 465, 9, 22, "Print"],
Cell[820031, 22663, 483, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[820551, 22677, 467, 13, 63, "Input"],
Cell[821021, 22692, 469, 9, 22, "Print"],
Cell[821493, 22703, 487, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[822017, 22717, 441, 13, 63, "Input"],
Cell[822461, 22732, 459, 9, 22, "Print"],
Cell[822923, 22743, 477, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[823437, 22757, 446, 13, 63, "Input"],
Cell[823886, 22772, 461, 9, 22, "Print"],
Cell[824350, 22783, 479, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[824866, 22797, 456, 13, 63, "Input"],
Cell[825325, 22812, 465, 9, 22, "Print"],
Cell[825793, 22823, 483, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[826313, 22837, 441, 13, 63, "Input"],
Cell[826757, 22852, 459, 9, 22, "Print"],
Cell[827219, 22863, 477, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[827733, 22877, 456, 13, 63, "Input"],
Cell[828192, 22892, 465, 9, 22, "Print"],
Cell[828660, 22903, 483, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[829180, 22917, 451, 13, 66, "Input"],
Cell[829634, 22932, 463, 9, 22, "Print"],
Cell[830100, 22943, 481, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[830618, 22957, 441, 13, 63, "Input"],
Cell[831062, 22972, 459, 9, 22, "Print"],
Cell[831524, 22983, 477, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[832038, 22997, 446, 13, 63, "Input"],
Cell[832487, 23012, 461, 9, 22, "Print"],
Cell[832951, 23023, 479, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[833467, 23037, 456, 13, 63, "Input"],
Cell[833926, 23052, 465, 9, 22, "Print"],
Cell[834394, 23063, 483, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[834914, 23077, 462, 13, 66, "Input"],
Cell[835379, 23092, 467, 9, 22, "Print"],
Cell[835849, 23103, 485, 9, 22, "Print"]
}, Open  ]],
Cell[836349, 23115, 83, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[836457, 23120, 634, 15, 28, "Input"],
Cell[837094, 23137, 1975, 74, 28, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[839106, 23216, 55, 1, 28, "Input"],
Cell[839164, 23219, 29, 0, 28, "Output"]
}, Open  ]],
Cell[839208, 23222, 162, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[839395, 23229, 4382, 144, 114, "Input"],
Cell[843780, 23375, 29473, 832, 61, "Output"]
}, Open  ]],
Cell[873268, 24210, 284, 6, 55, "MathCaption"],
Cell[CellGroupData[{
Cell[873577, 24220, 151, 4, 28, "Input"],
Cell[873731, 24226, 41141, 1136, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[914909, 25367, 151, 4, 28, "Input"],
Cell[915063, 25373, 604, 13, 22, "Print"],
Cell[915670, 25388, 669, 13, 22, "Print"],
Cell[916342, 25403, 54723, 1464, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[971102, 26872, 127, 3, 28, "Input"],
Cell[971232, 26877, 44623, 1200, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1015892, 28082, 91, 2, 28, "Input"],
Cell[1015986, 28086, 27297, 754, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1043320, 28845, 214, 6, 28, "Input"],
Cell[1043537, 28853, 26929, 746, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1070503, 29604, 90, 2, 28, "Input"],
Cell[1070596, 29608, 26929, 746, 61, "Output"]
}, Open  ]],
Cell[1097540, 30357, 76, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[1097641, 30361, 118, 3, 28, "Input"],
Cell[1097762, 30366, 29609, 826, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1127408, 31197, 90, 2, 28, "Input"],
Cell[1127501, 31201, 29609, 826, 85, "Output"]
}, Open  ]],
Cell[1157125, 32030, 1181, 36, 46, "MathCaption"],
Cell[1158309, 32068, 95, 2, 28, "Input"],
Cell[1158407, 32072, 107, 3, 28, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[1158551, 32080, 86, 1, 29, "Subsubsection"],
Cell[1158640, 32083, 411, 7, 68, "Text"],
Cell[1159054, 32092, 102, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[1159181, 32098, 46, 0, 28, "Input"],
Cell[1159230, 32100, 17452, 487, 37, "Output"]
}, Open  ]],
Cell[1176697, 32590, 117, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[1176839, 32597, 164, 5, 28, "Input"],
Cell[1177006, 32604, 33489, 891, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1210532, 33500, 149, 4, 28, "Input"],
Cell[1210684, 33506, 40323, 1063, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1251044, 34574, 149, 4, 28, "Input"],
Cell[1251196, 34580, 609, 13, 22, "Print"],
Cell[1251808, 34595, 46879, 1220, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1298724, 35820, 151, 4, 28, "Input"],
Cell[1298878, 35826, 69467, 1792, 109, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1368382, 37623, 160, 4, 28, "Input"],
Cell[1368545, 37629, 94575, 2440, 109, "Output"]
}, Open  ]],
Cell[1463135, 40072, 244, 6, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[1463404, 40082, 84, 2, 28, "Input"],
Cell[1463491, 40086, 95339, 2464, 109, "Output"]
}, Open  ]],
Cell[1558845, 42553, 147, 4, 28, "Input"],
Cell[CellGroupData[{
Cell[1559017, 42561, 91, 2, 28, "Input"],
Cell[1559111, 42565, 109639, 3018, 229, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1668787, 45588, 279, 8, 28, "Input"],
Cell[1669069, 45598, 109575, 3018, 229, "Output"]
}, Open  ]],
Cell[1778659, 48619, 164, 3, 40, "MathCaption"],
Cell[1778826, 48624, 114, 3, 28, "Input"],
Cell[1778943, 48629, 71, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[1779039, 48633, 77, 1, 28, "Input"],
Cell[1779119, 48636, 115689, 3021, 229, "Output"]
}, Open  ]],
Cell[1894823, 51660, 171, 3, 40, "MathCaption"],
Cell[1894997, 51665, 92, 2, 28, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[1895126, 51672, 46, 0, 29, "Subsubsection"],
Cell[1895175, 51674, 488, 10, 69, "Text"],
Cell[1895666, 51686, 190, 5, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[1895881, 51695, 241, 7, 28, "Input"],
Cell[1896125, 51704, 164102, 4261, 181, "Output"]
}, Open  ]],
Cell[2060242, 55968, 133, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2060400, 55975, 1025, 35, 46, "Input"],
Cell[2061428, 56012, 4135, 112, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2065600, 56129, 85, 2, 28, "Input"],
Cell[2065688, 56133, 4399, 118, 37, "Output"]
}, Open  ]],
Cell[2070102, 56254, 193, 6, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[2070320, 56264, 91, 2, 28, "Input"],
Cell[2070414, 56268, 112365, 3242, 181, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2182816, 59515, 83, 2, 28, "Input"],
Cell[2182902, 59519, 87386, 2554, 157, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2270325, 62078, 234, 7, 28, "Input"],
Cell[2270562, 62087, 42901, 1255, 85, "Output"]
}, Open  ]],
Cell[2313478, 63345, 45, 0, 40, "MathCaption"],
Cell[2313526, 63347, 114, 3, 28, "Input"],
Cell[2313643, 63352, 85, 2, 28, "Input"],
Cell[2313731, 63356, 192, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2313948, 63364, 115, 3, 28, "Input"],
Cell[2314066, 63369, 453, 9, 22, "Print"]
}, Open  ]],
Cell[2314534, 63381, 215, 8, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[2314774, 63393, 203, 7, 28, "Input"],
Cell[2314980, 63402, 1941, 59, 37, "Output"]
}, Open  ]],
Cell[2316936, 63464, 241, 7, 42, "MathCaption"],
Cell[2317180, 63473, 92, 2, 28, "Input"],
Cell[2317275, 63477, 152, 6, 42, "MathCaption"],
Cell[CellGroupData[{
Cell[2317452, 63487, 907, 32, 28, "Input"],
Cell[2318362, 63521, 2443, 75, 37, "Output"]
}, Open  ]],
Cell[2320820, 63599, 131, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2320976, 63606, 85, 2, 28, "Input"],
Cell[2321064, 63610, 5332, 152, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2326433, 63767, 85, 2, 28, "Input"],
Cell[2326521, 63771, 3108, 92, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2329666, 63868, 64, 1, 28, "Input"],
Cell[2329733, 63871, 8046, 234, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2337816, 64110, 119, 3, 28, "Input"],
Cell[2337938, 64115, 45829, 1476, 189, "Output"]
}, Open  ]],
Cell[2383782, 65594, 74, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2383881, 65598, 391, 12, 28, "Input"],
Cell[2384275, 65612, 400, 8, 24, "Message"],
Cell[2384678, 65622, 26749, 869, 123, "Output"]
}, Open  ]],
Cell[2411442, 66494, 135, 3, 40, "MathCaption"],
Cell[2411580, 66499, 74, 2, 28, "Input"],
Cell[2411657, 66503, 257, 9, 28, "Input"],
Cell[CellGroupData[{
Cell[2411939, 66516, 62, 1, 28, "Input"],
Cell[2412004, 66519, 15056, 485, 57, "Output"]
}, Open  ]],
Cell[2427075, 67007, 94, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2427194, 67013, 117, 3, 28, "Input"],
Cell[2427314, 67018, 15060, 485, 57, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[2442423, 67509, 119, 3, 29, "Subsubsection"],
Cell[2442545, 67514, 292, 5, 53, "MathCaption"],
Cell[CellGroupData[{
Cell[2442862, 67523, 290, 9, 28, "Input"],
Cell[2443155, 67534, 50033, 1238, 85, "Output"]
}, Open  ]],
Cell[2493203, 68775, 62, 1, 28, "Input"],
Cell[CellGroupData[{
Cell[2493290, 68780, 344, 9, 28, "Input"],
Cell[2493637, 68791, 57425, 1462, 133, "Output"]
}, Open  ]],
Cell[2551077, 70256, 71, 1, 28, "Input"],
Cell[CellGroupData[{
Cell[2551173, 70261, 386, 11, 28, "Input"],
Cell[2551562, 70274, 52657, 1302, 109, "Output"]
}, Open  ]],
Cell[2604234, 71579, 62, 1, 28, "Input"],
Cell[2604299, 71582, 112, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2604436, 71589, 465, 13, 28, "Input"],
Cell[2604904, 71604, 463, 9, 22, "Print"],
Cell[2605370, 71615, 481, 9, 22, "Print"],
Cell[2605854, 71626, 463, 9, 22, "Print"],
Cell[2606320, 71637, 481, 9, 22, "Print"],
Cell[2606804, 71648, 463, 9, 22, "Print"],
Cell[2607270, 71659, 481, 9, 22, "Print"],
Cell[2607754, 71670, 463, 9, 22, "Print"],
Cell[2608220, 71681, 481, 9, 22, "Print"],
Cell[2608704, 71692, 463, 9, 22, "Print"],
Cell[2609170, 71703, 481, 9, 22, "Print"]
}, Open  ]],
Cell[2609666, 71715, 121, 3, 40, "MathCaption"],
Cell[2609790, 71720, 1841, 50, 192, "Input"],
Cell[2611634, 71772, 62, 1, 28, "Input"],
Cell[2611699, 71775, 71, 1, 28, "Input"],
Cell[2611773, 71778, 82, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2611880, 71783, 2447, 82, 97, "Input"],
Cell[2614330, 71867, 8503, 232, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2622870, 72104, 95, 2, 28, "Input"],
Cell[2622968, 72108, 8503, 232, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2631508, 72345, 118, 3, 28, "Input"],
Cell[2631629, 72350, 9773, 272, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2641439, 72627, 95, 2, 28, "Input"],
Cell[2641537, 72631, 9773, 272, 37, "Output"]
}, Open  ]],
Cell[2651325, 72906, 62, 1, 28, "Input"],
Cell[2651390, 72909, 71, 1, 28, "Input"],
Cell[2651464, 72912, 114, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2651603, 72919, 658, 17, 46, "Input"],
Cell[2652264, 72938, 465, 9, 22, "Print"],
Cell[2652732, 72949, 483, 9, 22, "Print"],
Cell[2653218, 72960, 465, 9, 22, "Print"],
Cell[2653686, 72971, 483, 9, 22, "Print"],
Cell[2654172, 72982, 465, 9, 22, "Print"],
Cell[2654640, 72993, 483, 9, 22, "Print"],
Cell[2655126, 73004, 465, 9, 22, "Print"],
Cell[2655594, 73015, 483, 9, 22, "Print"],
Cell[2656080, 73026, 465, 9, 22, "Print"],
Cell[2656548, 73037, 483, 9, 22, "Print"],
Cell[2657034, 73048, 465, 9, 22, "Print"],
Cell[2657502, 73059, 483, 9, 22, "Print"],
Cell[2657988, 73070, 465, 9, 22, "Print"],
Cell[2658456, 73081, 483, 9, 22, "Print"],
Cell[2658942, 73092, 465, 9, 22, "Print"],
Cell[2659410, 73103, 483, 9, 22, "Print"],
Cell[2659896, 73114, 465, 9, 22, "Print"],
Cell[2660364, 73125, 483, 9, 22, "Print"]
}, Open  ]],
Cell[2660862, 73137, 112, 3, 40, "MathCaption"],
Cell[2660977, 73142, 1585, 45, 165, "Input"],
Cell[2662565, 73189, 62, 1, 28, "Input"],
Cell[2662630, 73192, 83, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2662738, 73197, 4579, 146, 165, "Input"],
Cell[2667320, 73345, 15983, 434, 61, "Output"]
}, Open  ]],
Cell[2683318, 73782, 114, 3, 28, "Input"],
Cell[CellGroupData[{
Cell[2683457, 73789, 150, 4, 28, "Input"],
Cell[2683610, 73795, 3743, 128, 46, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2687390, 73928, 86, 2, 28, "Input"],
Cell[2687479, 73932, 3752, 128, 46, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2691268, 74065, 659, 15, 46, "Input"],
Cell[2691930, 74082, 3743, 128, 46, "Output"]
}, Open  ]],
Cell[2695688, 74213, 103, 3, 28, "Input"],
Cell[2695794, 74218, 62, 1, 28, "Input"],
Cell[2695859, 74221, 138, 4, 28, "Input"],
Cell[2696000, 74227, 77, 0, 40, "MathCaption"],
Cell[2696080, 74229, 317, 10, 28, "Input"],
Cell[2696400, 74241, 65, 1, 28, "Input"],
Cell[2696468, 74244, 132, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2696625, 74251, 181, 6, 28, "Input"],
Cell[2696809, 74259, 1073, 30, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2697919, 74294, 234, 7, 28, "Input"],
Cell[2698156, 74303, 50367, 1369, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2748560, 75677, 220, 6, 28, "Input"],
Cell[2748783, 75685, 62490, 1677, 85, "Output"]
}, Open  ]],
Cell[2811288, 77365, 82, 2, 28, "Input"],
Cell[2811373, 77369, 110, 3, 28, "Input"],
Cell[CellGroupData[{
Cell[2811508, 77376, 53, 1, 28, "Input"],
Cell[2811564, 77379, 80915, 2285, 181, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2892516, 79669, 232, 7, 28, "Input"],
Cell[2892751, 79678, 39735, 1123, 109, "Output"]
}, Open  ]],
Cell[2932501, 80804, 128, 3, 40, "MathCaption"],
Cell[2932632, 80809, 65, 1, 28, "Input"],
Cell[2932700, 80812, 110, 3, 28, "Input"],
Cell[2932813, 80817, 65, 1, 28, "Input"],
Cell[2932881, 80820, 134, 3, 40, "MathCaption"],
Cell[2933018, 80825, 251, 8, 28, "Input"],
Cell[CellGroupData[{
Cell[2933294, 80837, 392, 14, 28, "Input"],
Cell[2933689, 80853, 3183, 89, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2936909, 80947, 85, 2, 28, "Input"],
Cell[2936997, 80951, 3895, 108, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2940929, 81064, 64, 1, 28, "Input"],
Cell[2940996, 81067, 7056, 192, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2948089, 81264, 91, 2, 28, "Input"],
Cell[2948183, 81268, 25141, 720, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2973361, 81993, 53, 1, 28, "Input"],
Cell[2973417, 81996, 25141, 720, 85, "Output"]
}, Open  ]],
Cell[2998573, 82719, 115, 3, 28, "Input"],
Cell[2998691, 82724, 63, 1, 28, "Input"],
Cell[2998757, 82727, 79, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[2998861, 82731, 328, 11, 28, "Input"],
Cell[2999192, 82744, 2476, 71, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3001705, 82820, 85, 2, 28, "Input"],
Cell[3001793, 82824, 2656, 76, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3004486, 82905, 93, 2, 28, "Input"],
Cell[3004582, 82909, 427, 15, 28, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3005046, 82929, 96, 2, 28, "Input"],
Cell[3005145, 82933, 426, 15, 28, "Output"]
}, Open  ]],
Cell[3005586, 82951, 69, 1, 28, "Input"],
Cell[3005658, 82954, 105, 2, 40, "MathCaption"],
Cell[3005766, 82958, 253, 8, 28, "Input"],
Cell[CellGroupData[{
Cell[3006044, 82970, 319, 11, 28, "Input"],
Cell[3006366, 82983, 3374, 96, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3009777, 83084, 85, 2, 28, "Input"],
Cell[3009865, 83088, 3914, 111, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3013816, 83204, 122, 3, 28, "Input"],
Cell[3013941, 83209, 25285, 722, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3039263, 83936, 53, 1, 28, "Input"],
Cell[3039319, 83939, 15697, 448, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3055053, 84392, 91, 2, 28, "Input"],
Cell[3055147, 84396, 15183, 432, 61, "Output"]
}, Open  ]],
Cell[3070345, 84831, 65, 1, 28, "Input"],
Cell[3070413, 84834, 98, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3070536, 84840, 771, 18, 63, "Input"],
Cell[3071310, 84860, 2912, 106, 28, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3074271, 84972, 97, 1, 29, "Subsubsection"],
Cell[3074371, 84975, 99, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3074495, 84981, 468, 15, 28, "Input"],
Cell[3074966, 84998, 607, 13, 22, "Print"],
Cell[3075576, 85013, 12936, 337, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3088549, 85355, 107, 2, 28, "Input"],
Cell[3088659, 85359, 12943, 337, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3101639, 85701, 85, 2, 28, "Input"],
Cell[3101727, 85705, 24448, 649, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3126212, 86359, 85, 2, 28, "Input"],
Cell[3126300, 86363, 20356, 537, 37, "Output"]
}, Open  ]],
Cell[3146671, 86903, 82, 2, 28, "Input"],
Cell[3146756, 86907, 110, 3, 28, "Input"],
Cell[3146869, 86912, 82, 1, 40, "MathCaption"],
Cell[3146954, 86915, 143, 4, 28, "Input"],
Cell[CellGroupData[{
Cell[3147122, 86923, 55, 1, 28, "Input"],
Cell[3147180, 86926, 206474, 6973, 559, "Output"]
}, Open  ]],
Cell[3353669, 93902, 228, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3353922, 93910, 676, 22, 28, "Input"],
Cell[3354601, 93934, 13441, 446, 57, "Output"]
}, Open  ]],
Cell[3368057, 94383, 143, 4, 28, "Input"],
Cell[3368203, 94389, 71, 2, 28, "Input"],
Cell[CellGroupData[{
Cell[3368299, 94395, 331, 10, 28, "Input"],
Cell[3368633, 94407, 400, 8, 24, "Message"],
Cell[3369036, 94417, 73098, 2508, 189, "Output"]
}, Open  ]],
Cell[3442149, 96928, 157, 5, 28, "Input"],
Cell[CellGroupData[{
Cell[3442331, 96937, 130, 3, 28, "Input"],
Cell[3442464, 96942, 65446, 2197, 198, "Output"]
}, Open  ]],
Cell[3507925, 99142, 92, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3508042, 99148, 164, 4, 28, "Input"],
Cell[3508209, 99154, 80551, 2704, 191, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3588809, 101864, 43, 0, 29, "Subsubsection"],
Cell[3588855, 101866, 109, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3588989, 101872, 544, 18, 28, "Input"],
Cell[3589536, 101892, 7967, 203, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3597540, 102100, 60, 1, 28, "Input"],
Cell[3597603, 102103, 3786, 101, 37, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3601426, 102209, 145, 4, 28, "Input"],
Cell[3601574, 102215, 32895, 877, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3634506, 103097, 60, 1, 28, "Input"],
Cell[3634569, 103100, 32895, 877, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3667501, 103982, 107, 2, 28, "Input"],
Cell[3667611, 103986, 32907, 877, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3700555, 104868, 145, 4, 28, "Input"],
Cell[3700703, 104874, 41904, 1090, 85, "Output"]
}, Open  ]],
Cell[3742622, 105967, 82, 2, 28, "Input"],
Cell[3742707, 105971, 110, 3, 28, "Input"],
Cell[3742820, 105976, 142, 4, 28, "Input"],
Cell[3742965, 105982, 66, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3743056, 105986, 795, 26, 28, "Input"],
Cell[3743854, 106014, 6288, 207, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3750179, 106226, 277, 9, 28, "Input"],
Cell[3750459, 106237, 6558, 216, 35, "Output"]
}, Open  ]],
Cell[3757032, 106456, 141, 4, 28, "Input"],
Cell[3757176, 106462, 69, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3757270, 106466, 157, 4, 28, "Input"],
Cell[3757430, 106472, 75743, 2508, 258, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[3833234, 108987, 42, 0, 36, "Subsection"],
Cell[3833279, 108989, 1127, 29, 107, "Text"],
Cell[CellGroupData[{
Cell[3834431, 109022, 47, 0, 35, "Subsubsection"],
Cell[3834481, 109024, 498, 12, 70, "Text"],
Cell[3834982, 109038, 264, 6, 63, "Input"],
Cell[3835249, 109046, 74, 0, 40, "MathCaption"],
Cell[3835326, 109048, 527, 16, 63, "Input"],
Cell[3835856, 109066, 43, 0, 40, "MathCaption"],
Cell[3835902, 109068, 310, 9, 28, "Input"],
Cell[3836215, 109079, 38, 0, 40, "MathCaption"],
Cell[3836256, 109081, 832, 24, 63, "Input"],
Cell[3837091, 109107, 40, 0, 40, "MathCaption"],
Cell[3837134, 109109, 259, 8, 28, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[3837430, 109122, 57, 0, 29, "Subsubsection"],
Cell[3837490, 109124, 144, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3837659, 109132, 242, 7, 28, "Input"],
Cell[3837904, 109141, 1325, 50, 28, "Output"]
}, Open  ]],
Cell[3839244, 109194, 116, 3, 40, "MathCaption"],
Cell[3839363, 109199, 1547, 51, 63, "Input"],
Cell[3840913, 109252, 134, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3841072, 109260, 73, 1, 28, "Input"],
Cell[3841148, 109263, 1605, 58, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3842790, 109326, 264, 8, 28, "Input"],
Cell[3843057, 109336, 693, 26, 28, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3843799, 109368, 75, 0, 29, "Subsubsection"],
Cell[3843877, 109370, 133, 3, 40, "MathCaption"],
Cell[3844013, 109375, 927, 30, 46, "Input"],
Cell[3844943, 109407, 1144, 35, 63, "Input"],
Cell[3846090, 109444, 1676, 54, 63, "Input"],
Cell[3847769, 109500, 184, 6, 28, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[3847990, 109511, 51, 0, 29, "Subsubsection"],
Cell[3848044, 109513, 336, 8, 50, "Text"],
Cell[3848383, 109523, 991, 20, 55, "MathCaption"],
Cell[3849377, 109545, 1486, 43, 131, "Input"],
Cell[3850866, 109590, 85, 1, 40, "MathCaption"],
Cell[3850954, 109593, 1128, 34, 114, "Input"],
Cell[3852085, 109629, 109, 2, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3852219, 109635, 70, 1, 28, "Input"],
Cell[3852292, 109638, 454, 9, 22, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3852783, 109652, 70, 1, 28, "Input"],
Cell[3852856, 109655, 454, 9, 22, "Print"]
}, Open  ]],
Cell[3853325, 109667, 69, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3853419, 109671, 199, 6, 28, "Input"],
Cell[3853621, 109679, 596, 13, 22, "Print"],
Cell[3854220, 109694, 614, 13, 22, "Print"]
}, Open  ]],
Cell[3854849, 109710, 1470, 25, 45, "MathCaption"],
Cell[3856322, 109737, 522, 18, 46, "Input"],
Cell[3856847, 109757, 1419, 35, 45, "MathCaption"],
Cell[CellGroupData[{
Cell[3858291, 109796, 4414, 138, 80, "Input"],
Cell[3862708, 109936, 18716, 529, 61, "Output"]
}, Open  ]],
Cell[3881439, 110468, 87, 2, 28, "Input"],
Cell[3881529, 110472, 105, 3, 28, "Input"],
Cell[CellGroupData[{
Cell[3881659, 110479, 258, 8, 28, "Input"],
Cell[3881920, 110489, 22761, 637, 61, "Output"]
}, Open  ]],
Cell[3904696, 111129, 87, 2, 28, "Input"],
Cell[3904786, 111133, 135, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3904946, 111140, 274, 8, 28, "Input"],
Cell[3905223, 111150, 5976, 197, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3911236, 111352, 234, 7, 28, "Input"],
Cell[3911473, 111361, 6088, 201, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3917598, 111567, 327, 10, 28, "Input"],
Cell[3917928, 111579, 5444, 169, 57, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3923409, 111753, 274, 8, 28, "Input"],
Cell[3923686, 111763, 5321, 177, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3929044, 111945, 207, 6, 28, "Input"],
Cell[3929254, 111953, 5457, 185, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3934748, 112143, 327, 10, 28, "Input"],
Cell[3935078, 112155, 4925, 153, 57, "Output"]
}, Open  ]],
Cell[3940018, 112311, 3183, 66, 45, "MathCaption"],
Cell[3943204, 112379, 1126, 40, 80, "Input"],
Cell[3944333, 112421, 145, 4, 40, "MathCaption"],
Cell[3944481, 112427, 845, 27, 46, "Input"]
}, Closed]],
Cell[CellGroupData[{
Cell[3945363, 112459, 70, 0, 29, "Subsubsection"],
Cell[3945436, 112461, 1594, 28, 51, "Text"],
Cell[3947033, 112491, 83, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3947141, 112496, 853, 29, 28, "Input"],
Cell[3947997, 112527, 2926, 81, 37, "Output"]
}, Open  ]],
Cell[3950938, 112611, 117, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[3951080, 112618, 180, 5, 28, "Input"],
Cell[3951263, 112625, 33125, 890, 85, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3984425, 113520, 218, 7, 28, "Input"],
Cell[3984646, 113529, 16241, 440, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4000924, 113974, 231, 8, 28, "Input"],
Cell[4001158, 113984, 16241, 440, 61, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4017436, 114429, 160, 5, 28, "Input"],
Cell[4017599, 114436, 146779, 4094, 301, "Output"]
}, Open  ]],
Cell[4164393, 118533, 381, 12, 28, "Input"],
Cell[4164777, 118547, 487, 15, 28, "Input"],
Cell[CellGroupData[{
Cell[4165289, 118566, 124, 3, 28, "Input"],
Cell[4165416, 118571, 87869, 2473, 229, "Output"]
}, Open  ]],
Cell[4253300, 121047, 76, 0, 40, "MathCaption"],
Cell[4253379, 121049, 163, 5, 28, "Input"],
Cell[4253545, 121056, 172, 5, 28, "Input"],
Cell[CellGroupData[{
Cell[4253742, 121065, 120, 3, 28, "Input"],
Cell[4253865, 121070, 100669, 2795, 277, "Output"]
}, Open  ]],
Cell[4354549, 123868, 4364, 94, 46, "MathCaption"],
Cell[4358916, 123964, 74, 2, 28, "Input"],
Cell[CellGroupData[{
Cell[4359015, 123970, 124, 3, 28, "Input"],
Cell[4359142, 123975, 77215, 2128, 157, "Output"]
}, Open  ]],
Cell[4436372, 126106, 80, 2, 28, "Input"],
Cell[4436455, 126110, 46, 0, 30, "Text"],
Cell[CellGroupData[{
Cell[4436526, 126114, 136, 3, 28, "Input"],
Cell[4436665, 126119, 137, 4, 28, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4436851, 126129, 62, 0, 29, "Subsubsection"],
Cell[4436916, 126131, 360, 7, 49, "Text"],
Cell[4437279, 126140, 64, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4437368, 126144, 45, 0, 28, "Input"],
Cell[4437416, 126146, 1325, 50, 28, "Output"]
}, Open  ]],
Cell[4438756, 126199, 83, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4438864, 126204, 633, 21, 28, "Input"],
Cell[4439500, 126227, 7789, 266, 57, "Output"]
}, Open  ]],
Cell[4447304, 126496, 115, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4447444, 126503, 158, 5, 28, "Input"],
Cell[4447605, 126510, 43605, 1482, 127, "Output"]
}, Open  ]],
Cell[4491225, 127995, 141, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4491391, 128002, 56, 1, 28, "Input"],
Cell[4491450, 128005, 73820, 2532, 189, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4565307, 130542, 58, 1, 28, "Input"],
Cell[4565368, 130545, 53998, 1847, 153, "Output"]
}, Open  ]],
Cell[4619381, 132395, 144, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4619550, 132402, 289, 9, 28, "Input"],
Cell[4619842, 132413, 28421, 972, 96, "Output"]
}, Open  ]],
Cell[4648278, 133388, 74, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4648377, 133392, 305, 9, 28, "Input"],
Cell[4648685, 133403, 400, 8, 24, "Message"],
Cell[4649088, 133413, 28729, 984, 96, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4677854, 134402, 62, 1, 28, "Input"],
Cell[4677919, 134405, 27802, 946, 96, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4705758, 135356, 81, 2, 28, "Input"],
Cell[4705842, 135360, 27814, 946, 96, "Output"]
}, Open  ]],
Cell[4733671, 136309, 125, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4733821, 136316, 89, 2, 28, "Input"],
Cell[4733913, 136320, 29, 0, 28, "Output"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[4733991, 136326, 64, 0, 29, "Subsubsection"],
Cell[4734058, 136328, 296, 6, 49, "Text"],
Cell[4734357, 136336, 195, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4734577, 136344, 64, 1, 28, "Input"],
Cell[4734644, 136347, 86873, 2864, 276, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4821554, 139216, 58, 1, 28, "Input"],
Cell[4821615, 139219, 62841, 2084, 233, "Output"]
}, Open  ]],
Cell[4884471, 141306, 74, 0, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[4884570, 141310, 305, 9, 28, "Input"],
Cell[4884878, 141321, 400, 8, 24, "Message"],
Cell[4885281, 141331, 66574, 2185, 259, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4951892, 143521, 62, 1, 28, "Input"],
Cell[4951957, 143524, 64502, 2112, 250, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5016496, 145641, 81, 2, 28, "Input"],
Cell[5016580, 145645, 64513, 2112, 250, "Output"]
}, Open  ]],
Cell[5081108, 147760, 126, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5081259, 147767, 91, 2, 28, "Input"],
Cell[5081353, 147771, 29, 0, 28, "Output"]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[5081443, 147778, 36, 0, 36, "Subsection"],
Cell[5081482, 147780, 612, 11, 88, "Text"],
Cell[5082097, 147793, 178, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5082300, 147800, 44, 0, 28, "Input"],
Cell[5082347, 147802, 1975, 74, 28, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5084359, 147881, 40, 0, 28, "Input"],
Cell[5084402, 147883, 2912, 106, 28, "Output"]
}, Open  ]],
Cell[5087329, 147992, 565, 18, 61, "MathCaption"],
Cell[CellGroupData[{
Cell[5087919, 148014, 127, 3, 28, "Input"],
Cell[5088049, 148019, 32, 0, 40, "Output"]
}, Open  ]],
Cell[5088096, 148022, 162, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5088283, 148029, 60, 1, 28, "Input"],
Cell[5088346, 148032, 168, 6, 28, "Output"]
}, Open  ]],
Cell[5088529, 148041, 620, 17, 57, "MathCaption"],
Cell[CellGroupData[{
Cell[5089174, 148062, 772, 26, 28, "Input"],
Cell[5089949, 148090, 1032, 35, 35, "Output"]
}, Open  ]],
Cell[5090996, 148128, 892, 20, 83, "MathCaption"],
Cell[CellGroupData[{
Cell[5091913, 148152, 45, 0, 28, "Input"],
Cell[5091961, 148154, 1325, 50, 28, "Output"]
}, Open  ]],
Cell[5093301, 148207, 352, 12, 44, "MathCaption"],
Cell[CellGroupData[{
Cell[5093678, 148223, 816, 26, 28, "Input"],
Cell[5094497, 148251, 1089, 36, 28, "Output"]
}, Open  ]],
Cell[5095601, 148290, 440, 11, 65, "MathCaption"],
Cell[CellGroupData[{
Cell[5096066, 148305, 56, 1, 28, "Input"],
Cell[5096125, 148308, 3935, 135, 35, "Output"]
}, Open  ]],
Cell[5100075, 148446, 124, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5100224, 148453, 56, 1, 28, "Input"],
Cell[5100283, 148456, 7089, 244, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5107409, 148705, 57, 1, 28, "Input"],
Cell[5107469, 148708, 1089, 36, 28, "Output"]
}, Open  ]],
Cell[5108573, 148747, 191, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5108789, 148755, 812, 26, 28, "Input"],
Cell[5109604, 148783, 1209, 40, 28, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5110850, 148828, 57, 1, 28, "Input"],
Cell[5110910, 148831, 2863, 99, 35, "Output"]
}, Open  ]],
Cell[5113788, 148933, 173, 4, 40, "MathCaption"],
Cell[5113964, 148939, 84, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5114073, 148944, 39, 0, 28, "Input"],
Cell[5114115, 148946, 15060, 485, 57, "Output"]
}, Open  ]],
Cell[5129190, 149434, 79, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5129294, 149439, 33, 0, 28, "Input"],
Cell[5129330, 149441, 73116, 2508, 189, "Output"]
}, Open  ]],
Cell[5202461, 151952, 81, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5202567, 151957, 35, 0, 28, "Input"],
Cell[5202605, 151959, 75743, 2508, 258, "Output"]
}, Open  ]],
Cell[5278363, 154470, 318, 10, 45, "MathCaption"],
Cell[CellGroupData[{
Cell[5278706, 154484, 40, 0, 28, "Input"],
Cell[5278749, 154486, 77215, 2128, 157, "Output"]
}, Open  ]],
Cell[5355979, 156617, 68, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5356072, 156622, 34, 0, 28, "Input"],
Cell[5356109, 156624, 27814, 946, 96, "Output"]
}, Open  ]],
Cell[5383938, 157573, 70, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5384033, 157578, 36, 0, 28, "Input"],
Cell[5384072, 157580, 64513, 2112, 250, "Output"]
}, Open  ]],
Cell[5448600, 159695, 181, 4, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5448806, 159703, 54, 1, 28, "Input"],
Cell[5448863, 159706, 251, 4, 43, "Print",
 CellTags->"Info3583346399-5671824"]
}, Open  ]],
Cell[5449129, 159713, 100, 1, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5449254, 159718, 86, 2, 28, "Input"],
Cell[5449343, 159722, 5565, 196, 32, "Output"]
}, Open  ]],
Cell[5454923, 159921, 121, 3, 40, "MathCaption"],
Cell[CellGroupData[{
Cell[5455069, 159928, 111, 3, 28, "Input"],
Cell[5455183, 159933, 92, 3, 28, "Output"]
}, Open  ]],
Cell[5455290, 159939, 628, 16, 57, "MathCaption"],
Cell[CellGroupData[{
Cell[5455943, 159959, 90, 2, 28, "Input"],
Cell[5456036, 159963, 12672, 364, 37, "Output"]
}, Open  ]],
Cell[5468723, 160330, 608, 19, 46, "MathCaption"],
Cell[CellGroupData[{
Cell[5469356, 160353, 561, 17, 28, "Input"],
Cell[5469920, 160372, 4194, 141, 28, "Output"]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
